#include "setplay_message_parser.h"

#include <rcsc/common/audio_codec.h>
#include <rcsc/common/logger.h>

#include <setplay/model/setplay.h>

#include <sstream>

using namespace setplay;
using namespace rcsc;

unsigned int SetplayMessageParser::parseSetplayBase(const char *msg,
                                                    unsigned int &setplay_id,
                                                    bool &inverted)
{
    /// MessageBitSchema: SetplayId Inverted Arguments
    /// FirstByte: 00SSSSSI
    /// Remaining bytes: tuple of punum
    /// Septlay
    /// Inverted (1bits): SSSS
    /// PlayerUnum (4bits): PPPP

    // 1) Encode set-play id => This should be further abstracted
    // Setplay-Id => Must use one byte or agree on the number of used set-plays => Must do renunmbering (no way to infer which plan id is used if mixed with other info)

    // Extract set-play
    unsigned int value = AudioCodec::i().charToIntMap().at( *msg );

    setplay_id = value >> 1;

    inverted = value & 0x01;

    Setplay* setplay = M_setplay_database->getSetplay( setplay_id );
    if ( setplay == NULL )
    {
        dlog.addText( Logger::SENSOR,
                      __FILE__ " SetplayMessage: No setplay found with id %d. Abort parsing message ! Should assert",
                      setplay_id);
        assert( false );
    }

    return 1;
}

unsigned int SetplayMessageParser::parseSetplayArguments(const char *msg,
                                                         Setplay& setplay,
                                                         IMessage& smsg)
{
    if ( !setplay.hasParameters() )
    {
        return 0;
    }

    setplay::Parameters::const_iterator itp = setplay.parameters()->begin();
    unsigned int params_read = 0;
    while ( itp != setplay.parameters()->end() )
    {
        if ( itp->second->type() == TT_VAR_PLAYER )
        {
            unsigned int number = AudioCodec::i().hex2unum( *(msg + params_read) );
            setplay::PlayerAffiliation* player = new setplay::PlayerAffiliation( number, setplay::TA_OUR );
            smsg.push( itp->first, player );
            params_read++;
        }
        else
        {
            // Unexpected setplay object type
            assert( false );
        }
        itp++;
    }

    return params_read;
}

/*
 * StepTransitionMessageParser
 */

StepTransitionMessageParser::StepTransitionMessageParser( SetplayCommunication *setplay_communication,
                                                          SetplayDatabase *setplay_database )
    : SetplayMessageParser( setplay_communication, setplay_database )
{

}

StepTransitionMessageParser::~StepTransitionMessageParser()
{

}

int StepTransitionMessageParser::parse( const int sender,
                              const double & ,
                              const char * msg,
                              const GameTime & current )
{
    int msg_size = 0;
    if ( *msg != sheader() )
    {
        return msg_size;
    }

    dlog.addText(Logger::SENSOR, "Started the decoding of StepTransitionMessage %s", msg);

    msg_size++;

    // 1) Try to parse the setplay_id
    unsigned int setplay_id = 0;
    bool inverted = false;
    unsigned int bytes_read = parseSetplayBase( msg + msg_size, setplay_id, inverted );
    if ( bytes_read == 0 )
    {
        dlog.addText( Logger::PLAN, __FILE__" StepTransitionMessageParser: Could not read the set-play id");
        assert( false );
        return msg_size;
    }

    Setplay* setplay = M_setplay_database->getSetplay( setplay_id );
    if ( setplay == NULL )
    {
        dlog.addText( Logger::PLAN,
                      __FILE__ " StepTransitionMessageParser: Setplay %d not found",
                      setplay_id);
        assert( false );
        return msg_size;
    }

    msg_size += bytes_read;

    StepTransitionMessage* smsg = new StepTransitionMessage( setplay::PlayerAffiliation(sender, TA_OUR),
                                                             current.cycle(),
                                                             setplay_id);
    smsg->setInverted( inverted );


    if ( setplay->hasParameters() )
    {
        bytes_read = parseSetplayArguments( msg + msg_size, *setplay, *smsg );
        if ( bytes_read == 0 )
        {
            assert( false );
        }
        msg_size += bytes_read;
    }

    // 3) Try to parse the current and next step ids
    bytes_read = parseSteps( msg + msg_size, *smsg );
//     assert ( bytes_read == 2 );

    msg_size += bytes_read;

    stringstream ss;
    smsg->print(ss);
    dlog.addText( Logger::SENSOR,
                  __FILE__ " StepTransitionMessageParser: Received message %s",
                  ss.str().c_str());

    // Forward the message to an adequate handler
    M_setplay_communication->receive( smsg );

    return msg_size;
}


unsigned int StepTransitionMessageParser::parseSteps( const char* msg,
                                                      StepTransitionMessage& smsg ) const
{
    /*unsigned int res = AudioCodec::i().charToIntMap().at( *msg );

    /// TODO This will not for steps
    smsg.setCurrentStepId( res & 0xF0 );
    smsg.setNextStepId( res & 0x0F );
    return 1;
    */

    smsg.setCurrentStepId( AudioCodec::i().charToIntMap().at( *msg ) );
    smsg.setNextStepId( AudioCodec::i().charToIntMap().at( *(msg + 1) ) );

    return 2;
}

/*
 * StepEnterMessageParser
 */

StepEnterMessageParser::StepEnterMessageParser( SetplayCommunication *setplay_communication,
                                                          SetplayDatabase *setplay_database )
    : SetplayMessageParser( setplay_communication, setplay_database )
{

}

StepEnterMessageParser::~StepEnterMessageParser()
{

}

int StepEnterMessageParser::parse( const int sender,
                              const double & ,
                              const char * msg,
                              const GameTime & current )
{
    int msg_size = 0;
    if ( *msg != sheader() )
    {
        return msg_size;
    }

    dlog.addText(Logger::SENSOR, "Started the decoding of StepEnterMessage %s", msg);

    msg_size++;

    // 1) Try to parse setplay_id and invertible flag
    unsigned int setplay_id = 0;
    bool inverted = false;
    unsigned int bytes_read = parseSetplayBase( msg + msg_size, setplay_id, inverted );
    if ( bytes_read == 0 )
    {
        dlog.addText( Logger::PLAN, __FILE__" StepTransitionMessageParser: Could not read the set-play id");
        assert( false );
        return msg_size;
    }
    msg_size += bytes_read;

    Setplay* setplay = M_setplay_database->getSetplay( setplay_id );
    if ( setplay == NULL )
    {
        dlog.addText( Logger::PLAN,
                      __FILE__ " StepTransitionMessageParser: Setplay %d not found",
                      setplay_id);
        assert( false );
        return msg_size;
    }


    StepEnterMessage* smsg = new StepEnterMessage( setplay::PlayerAffiliation(sender, TA_OUR),
                                                   current.cycle(),
                                                   setplay_id);
    smsg->setInverted( inverted );

    if ( setplay->hasParameters() )
    {
        bytes_read = parseSetplayArguments( msg + msg_size, *setplay, *smsg );
        if ( bytes_read == 0 )
        {
            assert( false );
        }
        msg_size += bytes_read;
    }

    // 3) Try to parse the current step id
    unsigned int current_step_id = AudioCodec::i().charToIntMap().at( *(msg + msg_size) );
    smsg->setStepId( current_step_id );
    msg_size++;

    stringstream ss;
    smsg->print(ss);
    dlog.addText( Logger::SENSOR,
                  __FILE__ " StepEnterMessageParser: Received message %s",
                  ss.str().c_str());


    // Forward the message to an adequate handler
    M_setplay_communication->receive( smsg );

    return msg_size;
}


/*
 * SetplayStartMessageParser
 */

SetplayStartMessageParser::SetplayStartMessageParser( setplay::SetplayCommunication* setplay_communication,
                                                      setplay::SetplayDatabase *setplay_database)
    : SetplayMessageParser(setplay_communication, setplay_database)
{

}

SetplayStartMessageParser::~SetplayStartMessageParser()
{

}

int SetplayStartMessageParser::parse( const int sender,
                                      const double & ,
                                      const char * msg,
                                      const GameTime & current )
{
    int msg_size = 0;

    if ( *msg != sheader() )
    {
        return msg_size;
    }

    dlog.addText(Logger::SENSOR, __FILE__ " SetplayStartMessage decoding start");

    msg_size++;

    // 1) Try to parse the setplay_id
    unsigned int setplay_id = 0;
    bool inverted = false;
    unsigned int bytes_read = parseSetplayBase( msg + msg_size, setplay_id, inverted );
    if ( bytes_read == 0 )
    {
        dlog.addText( Logger::PLAN, __FILE__" Could not read the set-play id");
        assert( false );
        return msg_size;
    }
    msg_size += bytes_read;

    Setplay* setplay = M_setplay_database->getSetplay( setplay_id );
    if ( setplay == NULL )
    {
        dlog.addText( Logger::SENSOR,
                      __FILE__ " SetplayStartMessageParser: No setplay found with id %d. Abort parsing message ! Should assert",
                      setplay_id);
        assert( false );
        return msg_size;
    }

    /// Create the set-play
    SetplayStartMessage* smsg = new SetplayStartMessage( setplay::PlayerAffiliation(sender, TA_OUR),
                                                         current.cycle(),
                                                         setplay_id,
                                                         inverted);
    // 3) Check for setplay parameters if any
    if ( setplay->hasParameters() )
    {
        bytes_read = parseSetplayArguments( msg + msg_size, *setplay, *smsg );
        if ( bytes_read == 0 )
        {
            assert( false );
        }
        msg_size += bytes_read;
    }

    // Register the message in the audio memory
    // TODO Get a handle to the setplay manager (make this class register as an observer? or do polling everytime)
    M_setplay_communication->receive( smsg );

    // Message received
    stringstream ss;
    smsg->print( ss );
    dlog.addText(Logger::SENSOR,
                 __FILE__ " SetplayStartMessageParser: Decoded %s",
                 ss.str().c_str());

    return msg_size;
}


/*
 * AbortMessageParser
 */
int SetplayAbortMessageParser::parse( const int sender,
                              const double &,
                              const char * msg,
                              const GameTime & current )
{
    unsigned int msg_size = 0;
    if ( *msg != sheader() )
    {
        return msg_size;
    }

    msg_size++;

    // 1) Try to parse setplay_id and invertible flag
    unsigned int setplay_id = 0;
    bool inverted = false;
    unsigned int bytes_read = parseSetplayBase( msg + msg_size, setplay_id, inverted );
    if ( bytes_read == 0 )
    {
        dlog.addText( Logger::PLAN, __FILE__" Could not read the set-play id");
        assert( false );
        return msg_size;
    }
    msg_size += bytes_read;

    Setplay* setplay = M_setplay_database->getSetplay( setplay_id );
    if ( setplay == NULL )
    {
        dlog.addText( Logger::SENSOR,
                      __FILE__ " SetplayAbortMessageParser: No setplay found with id %d. Abort parsing message ! Should assert",
                      setplay_id);
        assert( false );
        return msg_size;
    }

    SetplayAbortMessage* smsg = new SetplayAbortMessage( setplay::PlayerAffiliation(sender, TA_OUR),
                                                         current.cycle(),
                                                         setplay_id,
                                                         inverted );

    // Forward the message to an adequate handler
    M_setplay_communication->receive( smsg );

    stringstream ss;
    smsg->print(ss);
    dlog.addText( Logger::SENSOR,
                  __FILE__ " Successfully parsed the SetplayAbortMessage %s ",
                  ss.str().c_str());


    return msg_size;
}


/*
 * FinishMessageParser
 */
int SetplayFinishMessageParser::parse( const int sender,
                                       const double & ,
                                       const char * msg,
                                       const GameTime & current )
{
    unsigned int msg_size = 0;
    if ( *msg != sheader() )
    {
        return msg_size;
    }

    msg_size++;

    // 1) Try to parse the setplay_id
    // 1) Try to parse setplay_id and invertible flag
    unsigned int setplay_id = 0;
    bool inverted = false;
    unsigned int bytes_read = parseSetplayBase( msg + msg_size, setplay_id, inverted );
    if ( bytes_read == 0 )
    {
        dlog.addText( Logger::PLAN, __FILE__" Could not read the set-play id");
        assert( false );
        return msg_size;
    }
    msg_size += bytes_read;


    Setplay* setplay = M_setplay_database->getSetplay( setplay_id );
    if ( setplay == NULL )
    {
        dlog.addText( Logger::SENSOR,
                      __FILE__ " SetplayFinishMessageParser: No setplay found with id %d. Abort parsing message ! Should assert",
                      setplay_id);
        assert( false );
        return msg_size;
    }

    SetplayFinishMessage* smsg = new SetplayFinishMessage( setplay::PlayerAffiliation(sender, TA_OUR),
                                                           current.cycle(),
                                                           setplay_id,
                                                           inverted );

    // Forward the message to an adequate handler
    M_setplay_communication->receive( smsg );

    stringstream ss;
    smsg->print(ss);
    dlog.addText( Logger::SENSOR,
                  __FILE__ " Successfully parsed the SetplayFinishMessage %s ",
                  ss.str().c_str());

    return msg_size;
}
