#ifndef FCPORTUGAL_SETPLAY_WORLD_MODEL_H
#define FCPORTUGAL_SETPLAY_WORLD_MODEL_H

#include <map>
#include <vector>
#include <string>

#include <boost/shared_ptr.hpp>

#include <rcsc/player/world_model.h>

#include <setplay/runtime/iworld_model.h>
#include <setplay/runtime/memory_space.h>

// #include <strategy/types.h>

#define BALL_OWNER_DISTANCE_THRESHOLD 1.0
#define PASS_MAX_DIST 50.0

namespace fcportugal
{

    // This should be our global environment of execution where variables reside
    class SetplayWorldModel
            : public setplay::IWorldModel {
    public:

          SetplayWorldModel(const rcsc::WorldModel* world_model );
          virtual ~SetplayWorldModel() ;

          const rcsc::WorldModel* world() const
          {
              return M_world;
          }

          void update ( );

          /// Override Soccer::IWorldModel
          unsigned int currentTime() const;
          unsigned int remainingTimeInPeriod() const;

          setplay::GameSituation gameSituation() const;

          setplay::TimePeriod timePeriod() const;

          unsigned int ballInterceptionTime( const setplay::PlayerAffiliation& player ) const;
          unsigned int playerInterceptionTime(const setplay::PlayerAffiliation& player1,
                                              const setplay::PlayerAffiliation& player2 ) const;


          /// COMMON REGIONS
          rcsc::Rect2D field( ) const;
          double defenseLine(setplay::TeamAffiliation team_affiliation) const;
          double offenseLine(setplay::TeamAffiliation team_affiliation) const;
          double goalLine(setplay::TeamAffiliation team_affiliation) const;
          double endLine(setplay::TeamAffiliation team_affiliation) const;
          rcsc::Rect2D goalieArea(setplay::TeamAffiliation team_affiliation) const;
          rcsc::Rect2D penaltyArea(setplay::TeamAffiliation team_affiliation) const;
          rcsc::Rect2D midfieldArea(setplay::TeamAffiliation team_affiliation) const;

          /// BALL INFORMATION
          rcsc::Vector2D ballPosition( ) const;
          double ballSpeed() const;
          rcsc::AngleDeg ballDirection() const;

          /// PLAYER INFORMATION
          unsigned int self() const;
          unsigned int goalie( setplay::TeamAffiliation team_affiliation ) const;
          rcsc::Vector2D playerPosition( const setplay::PlayerAffiliation& player ) const;
          double playerSpeed(const setplay::PlayerAffiliation &player) const;
          rcsc::AngleDeg playerBodyDirection(const setplay::PlayerAffiliation& player ) const;
          int playerTimeToReachPosition( const setplay::PlayerAffiliation& player, const rcsc::Vector2D& position ) const;
          int playerTimeToReachMaxSpeed( const setplay::PlayerAffiliation &player) const;
          double playerMaxSpeed( const setplay::PlayerAffiliation& player ) const;

          std::vector<unsigned int> activePlayers( setplay::TeamAffiliation team_affiliation ) const;
          /// Opps to ball owner (if ball owner is self)

          setplay::PlayerAffiliation playerFastestTo( const rcsc::Vector2D& position );
          setplay::PlayerAffiliation playerFastestTo( const rcsc::Region2D& region );
          setplay::PlayerAffiliation playerFastestTo( const setplay::PlayerAffiliation& player );
          setplay::PlayerAffiliation playerFastestToBall( );

          unsigned int teammateFastestTo( const rcsc::Vector2D& position );
          unsigned int teammateFastestTo( const rcsc::Region2D& region );
          unsigned int teammateFastestTo( const setplay::PlayerAffiliation& player );
          unsigned int teammateFastestToBall( );

          unsigned int opponentFastestTo( const rcsc::Vector2D& position );
          unsigned int opponentFastestTo( const rcsc::Region2D& region );
          unsigned int opponentFastestTo( const setplay::PlayerAffiliation& player );
          unsigned int opponentFastestToBall( );

          /// Teammates to ball owner (
          std::vector<unsigned int> teammatesFromPlayer( const setplay::PlayerAffiliation& player ) const;
          std::vector<unsigned int> opponentsFromPlayer( const setplay::PlayerAffiliation& player ) const;

          /*!
            \brief Get players closest to ball
            */
          std::vector<unsigned int> playersFromBall( setplay::TeamAffiliation team_affiliation ) const;
          std::vector<unsigned int> playersFromGoal( setplay::TeamAffiliation team_affiliation,
                                                       setplay::TeamAffiliation goal_affiliation) const;
          std::vector<unsigned int> playersBehindBall( setplay::TeamAffiliation team_affiliation) const;
          std::vector<unsigned int> playersInFrontOfBall( setplay::TeamAffiliation team_affiliation) const;

          setplay::PlayerAffiliation lastKicker() const;

          setplay::TeamAffiliation teamInPossession() const;

          unsigned int goalsScored(setplay::TeamAffiliation affiliation) const;

          std::vector<unsigned int> playersOffside(setplay::TeamAffiliation team_affiliation) const;
          std::vector<unsigned int> playersNotOffside(setplay::TeamAffiliation team_affiliation) const;


          // LMOTA For Setplays
          bool commAllowed() const;
          bool singleChannelComm() const;
          bool limitedComm() const;
          // TODO Should abstract to shouldRepeat(MsgType msg)
          unsigned int messageRepeatTime() const;

          std::vector<rcsc::Vector2D> teammateStrategicPos() const;


          // Helper methods
          void initPlayModes( );

          // Implementation of strategy::WorldModel
          int gamePace( ) const;
          int relativeScore( ) const;

          /// Helper methods
          rcsc::SideID side( const setplay::TeamAffiliation& team ) const;
          const rcsc::AbstractPlayerObject* getAbstractPlayerObject( const setplay::PlayerAffiliation& player ) const;


          const setplay::MemorySpace& globalSpace() const { return M_global_space; }
    protected:
          // Helper method for update
          void updatePlayersRelativeToBall( );
          void upatePlayerPositions( );
          void updateGameSituation( );
          void updateScoreAssessment( );
          void updateGameMode( );
          void updatePlayersCalculus( );
          void updateGamePace( );
          void updateScore( );
          void updateSoccerMoment();

          /// Updater functions
          void setSoccerMoment( const setplay::SoccerMomentAffiliation& soccer_moment );

          const rcsc::AbstractPlayerCont& playersBehindBall( rcsc::SideID team_side);
          const rcsc::AbstractPlayerCont& oursBehindBall( );
          const rcsc::AbstractPlayerCont& theirsBehindBall( );
    private:
          void updatePlayerOffside( std::vector<unsigned int>& players_offside,
                                    std::vector<unsigned int>& players_not_offside,
                                    double offside_line,
                                    const rcsc::AbstractPlayerObject& getAbstractPlayerObject  );
          void updatePlayerRelativeToBallX( std::vector<unsigned int>& players_behind_ball,
                                             std::vector<unsigned int>& players_in_front_ball,
                                             const rcsc::AbstractPlayerObject& getAbstractPlayerObject  );
          void updatePlayersFromBall( std::vector<unsigned int>& player_ids_from_ball,
                                      const rcsc::PlayerPtrCont& players_from_ball);

          void updatePlayers( const rcsc::AbstractPlayerObject* getAbstractPlayerObject  );


          /*!
            \class PlayerToPointDistCmp
            \brief predicate functor to compare player's distance to their goal
          */
          class PlayerToPointDistCmp
              : public std::binary_function< const unsigned int,
                                             const unsigned int,
                                             bool > {
          public:
              PlayerToPointDistCmp( const rcsc::AbstractPlayerCont& activePlayers,
                                    rcsc::Vector2D point);

              /*!
                \brief operation function
                \param lhs left hand side variable
                \param rhs right hand side variable
                \return compared result
               */
              bool operator()( unsigned int lhs,
                               unsigned int rhs ) const;

              void updateCache();
              void setPoint(rcsc::Vector2D point);
          private:
              const rcsc::AbstractPlayerCont& M_players;
              rcsc::Vector2D M_point;
              std::map<unsigned int, double> M_cached_distances;
          };

          unsigned int M_current_time;
          unsigned int M_remaining_time;
          setplay::TimePeriod M_time_period;
          setplay::SoccerMomentAffiliation M_soccer_moment;


          const rcsc::WorldModel* M_world;

          /// COMMON REGIONS
          /// Static regions
          rcsc::Rect2D M_field;
          rcsc::Rect2D M_left_goalie_area;
          rcsc::Rect2D M_right_goalie_area;
          rcsc::Rect2D M_left_penalty_area;
          rcsc::Rect2D M_right_penalty_area;
          rcsc::Rect2D M_left_midfield_area;
          rcsc::Rect2D M_right_midfield_area;

          double M_left_end_line;
          double M_right_end_line;
          double M_left_goal_line;
          double M_right_goal_line;

          /// Variable lines/regions
          double M_our_defense_line;
          double M_opp_defense_line;
          double M_our_offense_line;
          double M_opp_offense_line;


          /// Players calculations => Sorted by distances
          std::vector<unsigned int> M_our_players_from_ball;
          std::vector<unsigned int> M_opp_players_from_ball;

          std::vector<unsigned int> M_our_players_from_own_goal;
          std::vector<unsigned int> M_opp_players_from_own_goal;

          std::vector<unsigned int> M_our_players_to_opp_goal;
          std::vector<unsigned int> M_opp_players_to_our_goal;

          std::vector<unsigned int> M_our_players_behind_ball;
          std::vector<unsigned int> M_opp_players_behind_ball;

          std::vector<unsigned int> M_our_players_in_front_of_ball;
          std::vector<unsigned int> M_opp_players_in_front_of_ball;

          std::vector<unsigned int> M_our_players_offside;
          std::vector<unsigned int> M_opp_players_offside;

          std::vector<unsigned int> M_our_players_non_offside;
          std::vector<unsigned int> M_opp_players_non_offside;

          // soccer::strategy::ScoreAssessment M_score_assessment;

          setplay::GameSituation M_current_game_situation;
          setplay::GameSituation M_previous_game_situation;

          unsigned int M_our_score;
          unsigned int M_opp_score;
          int M_relative_score;
          int M_game_pace;

          PlayerToPointDistCmp M_teammates_from_own_goal_cmp;
          PlayerToPointDistCmp M_opponents_from_own_goal_cmp;

          //
          typedef std::map<rcsc::PlayMode, setplay::GameSituation> PlayModeToGameSituationMap;
          PlayModeToGameSituationMap M_play_modes_map;

          setplay::MemorySpace M_global_space;
    };

}
#endif // FCPORTUGAL_WORLD_MODEL_H
