#ifndef SP_SETPLAY_GENERATOR_SEXPR_H
#define SP_SETPLAY_GENERATOR_SEXPR_H

#include <vector>
#include <string>
#include <iostream>
#include <fstream>
#include <map>

#include <setplay/model/ivisitor.h>
#include <setplay/model/variable.h>
#include <setplay/model/operation.h>
#include <setplay/model/types.h>
#include <setplay/model/game_situation.h>

namespace setplay {
    class IPlayer;
    class IDirective;
    class IAction;
    class ICondition;
    class ITransition;

    class SetplayGeneratorSexpr
            : public IVisitor
    {
    private:
        std::ostream& M_os;

    public:
        SetplayGeneratorSexpr(std::ostream& os );

        void visit( Setplay& setplay );
        void visit( Step& step );
        void visit( IParticipation& participation);

        /// Transitions
        void visit( NextStep& transition );
        void visit( Finish& transition );
        void visit( Abort& transition );

        /// Directives
        void visit( Do& directive );
        void visit( DoWhile& directive);
        void visit( Dont& directive );

        /// Actions
        void visit( ActDribble& action );
        void visit( ActFollow& action );
        void visit( ActHold& );
        void visit( ActIntercept& action );
        void visit( ActPass& action );
        void visit( ActPos& action );
        void visit( ActShoot& action );
        void visit( ActLookAt& action);

        /// Conditions
        void visit( CondAnd& condition );
        void visit( CondOr& condition );
        void visit( CondNot& condition );
        void visit( CondBallOwner& condition );
        void visit( CondTeamInPossession& condition );
        void visit( CondBool& condition );
        void visit( CondGameSituation& condition );
        void visit( CondAtLocation& condition );
        void visit( CondTeamFormation& condition );
        void visit( CondOffside& condition );
        void visit( CondCompDecimal& condition );
        void visit( CondKickable& condition );
        void visit( CondCanIntercept& condition);
        void visit( CondCanShoot& condition);
        void visit( CondCanPass& condition);

        /// Locations
        void visit( Rect2D& location );
        void visit( Triangle2D& location );
        void visit( Sector2D& location );
        void visit( Circle2D& location );
        void visit( Ray2D& location );
        void visit( Line2D& location );

        /// Participant
        void visit( PlayerAffiliation& participant );

        /// Player functors
        void visit( PlayerSelf& player_functor);
        void visit( NearestPlayer& player_functor);
        void visit( NearestTeammate& player_functor);
        void visit( NearestOpponent& player_functor);
        void visit( Goalie& player_functor);

        /// Points
        void visit( Point2D& point );

        /// Variables => Is this really necessary??
        void visit( Variable& variable );
        void visit( VarLocation& variable );
        void visit( VarRegion& variable);
        void visit( VarPoint& variable);
        void visit( VarCondition& variable );
        void visit( VarDecimal& variable );
        void visit( VarInteger& variable );
        void visit( VarPlayer& variable );

        /// Point Functors
        void visit( MobileObjectLocation& point_functor);
        void visit( PlayerPosition& point_functor );
        void visit( BallPosition&  );
        // Vector unary operations
        void visit( VectorNormalize& point_functor );
        void visit( VectorInvertX& point_functor );
        void visit( VectorInvertY& point_functor );
        void visit( VectorInvertXY& point_functor );
        void visit( VectorSubtraction& point_functor );
        void visit( VectorAddition& point_functor );
        // Vector scalar operations
        void visit( VectorScale& point_functor );
        void visit( VectorRotation& point_functor );
        void visit( VectorDivision& point_functor );
        void visit( NearestIntersectionPoint& point_functor);
        void visit( RegionCentroid& point_functor);

        /// Basic Data Types
        void visit( Decimal& value );
        void visit( Integer& value );
        void visit( Bool& value );

        /// Decimal Functors
        // Number operations
        void visit( NumberAbs& decimal_functor );
        void visit( NumberSignChange& decimal_functor );
        void visit( NumberMultiplication& decimal_functor );
        void visit( NumberSubtraction& decimal_functor );
        void visit( NumberDivision& decimal_functor );
        void visit( NumberAddition& decimal_functor );
        // Vector to scalar operations
        void visit( VectorDirection& decimal_functor );
        void visit( VectorLength& decimal_functor );
        // Two vector to scalar operations
        void visit( VectorDistance& decimal_functor );
        void visit( VectorAngularDiff& decimal_functor );
        void visit( VectorInnerProduct& decimal_functor );
        void visit( VectorOuterProduct& decimal_functor );
        // In game queries
        void visit( DefenseLine& decimal_functor );
        void visit( OffenseLine& decimal_functor );
        void visit( OffsideLine& decimal_functor );
        void visit( MobileObjectSpeed& decimal_functor );

        void visit( PointX& decimal_functor );
        void visit( PointY& decimal_functor );

        /// Integer Functors
        void visit( CurrentTime&  ); // TODO Change this
        void visit( RemainingTime&);
        void visit( GoalsScored& integer_functor);
        void visit( GoalDiff& integer_functor);
        void visit( InterceptionTime& integer_functor );

        /// Assignment operation
        void visit( Assignment& assign);

        void visit( Ball& );
    protected:
        /// Helper methods
        void visitPlayers( const std::string &label, const vector<IPlayer *> *participants );
        void visitCondition( std::string label, ICondition* condition );
        void visitDirectives( std::vector<IDirective*>* directives  );
        void visitDirective( const std::string& label, IDirective& directive);
        void visitActions( std::vector<IAction*>* actions  );
        void visitConditions( std::vector<ICondition*>* conditions  );
        void visitTeamAffiliation( TeamAffiliation team_affiliation );
        void visitBool( bool value );
        void visitGameSituation( GameSituation game_situation );
        void visitVariable(const Variable& variable);

        template <typename ConcreteOperation, class Operand>
        void visit( std::string label, const UnaryOperation<ConcreteOperation, Operand>& unary_operation)
        {
            M_os << "(" << label << " ";
            unary_operation.operand()->visit( *this );
            M_os << ")";
        }

        template <typename ConcreteOperation, class OperandLeft, class OperandRight>
        void visit( std::string label, const BinaryOperation<ConcreteOperation, OperandLeft, OperandRight>& binary_operation)
        {
            M_os << "(" << label << " ";
            binary_operation.operandLeft()->visit( *this );
            M_os << " ";
            binary_operation.operandRight()->visit( *this );
            M_os << ")";
        }
    };

    static SetplayGeneratorSexpr SetplayGeneratorSexprLog(std::clog);
}

#endif // SP_SETPLAY_GENERATOR_SEXPR_H
