#include "iparticipation.h"

#include <algorithm>
#include <cassert>

#include <setplay/model/player_vars_visitor.h>
#include <setplay/model/participants/player_var.h>
#include <setplay/model/participants/player_affiliation.h>

#include <setplay/generators/sexpr/setplay_generator_sexpr.h>

namespace setplay {

    IParticipation::IParticipation(IPlayer *player,
                                   ILocation2D* location)
        : M_player(player)
        , M_location(location)
    {

    }

    IParticipation::IParticipation(const IParticipation& other)
    {
        M_player = other.M_player != NULL ? other.M_player->clone() : NULL;
        M_location = other.M_location != NULL ? other.M_location->clone() : NULL;
    }

    IParticipation::~IParticipation()
    {
        if ( M_player != NULL )
        {
            delete M_player;
        }
        if ( M_location != NULL )
        {
            delete M_location;
        }
    }

    //Inspectors
    IPlayer *IParticipation::player() const
    {
        return M_player;
    }

    ILocation2D* IParticipation::location() const
    {
        return M_location;
    }

    bool IParticipation::hasLocation() const
    {
        return M_location != NULL;
    }

    /// Modifiers

    IParticipation& IParticipation::setPlayer(IPlayer* player)
    {
        if ( M_player != NULL )
        {
            delete M_player;
        }
        M_player = player;
        return *this;
    }

    IParticipation& IParticipation::setLocation(ILocation2D* location)
    {
        if ( M_location != NULL )
        {
            delete M_location;
        }
        M_location = location;
        return *this;
    }


    IParticipation& IParticipation::operator=(const IParticipation& other)
    {
        if ( this != &other)
        {
            setPlayer( other.M_player !=NULL ? other.M_player->clone() : NULL);
            setLocation( other.M_location !=NULL ? other.M_location->clone() : NULL);
        }

        return *this;
    }

    /// Overriden methods
    TokenType IParticipation::type() const
    {
        return TT_PARTICIPATION;
    }

    IParticipation* IParticipation::clone() const
    {
        return new IParticipation( *this );
    }

    void IParticipation::visit(IVisitor& visitor)
    {
        visitor.visit( *this );
    }


    /// ParticipationCmp


    /*
     * Order in final list by order of appearance:
     * 1) playerIds
     * 2) PlayerRoles without relative positionings
     * 3) PlayerRoles with relative positionings relative to absolute points
     * 4) PlayerRoles with relative positionings to other relative points (e.g. players)
     *
     * Checks if p1 < p2 in order to resolve dependencies on instantiation
     */

    ParticipationCmp::ParticipationCmp()
    {

    }

    bool ParticipationCmp::operator()(IParticipation* p1,
                                      IParticipation* p2)
    {
        // In general, if this depends on participation with concern to positioning then it should be placed after participation
        // For efficiency sake, the following assumptions are being made
        // 1) this < participation, when both are PlayerIDs
        // 2) this < participation when both are PlayerRoles and have no location
        if ( p1->player()->type() == TT_PLAYER_AFFILIATION )
        {
            if ( p2->player()->type() == TT_PLAYER_AFFILIATION )
            {
                /// Tie-breaker: The one with the lowest number
                const PlayerAffiliation* p1_affilliation = static_cast<PlayerAffiliation*>( p1->player() );
                const PlayerAffiliation* p2_affilliation = static_cast<PlayerAffiliation*>( p2->player() );
                std::cout << ((p1_affilliation->number() < p2_affilliation->number()) ? "true" : "false")
                          << std::endl;
                return p1_affilliation->number() < p2_affilliation->number();
            }
            return true;
        }

        if ( p2->player()->type() == TT_PLAYER_AFFILIATION )
        {
            return false; // participation is a PlayerID
        }

        // this and participation are both PlayerRoles
        if ( !p1->hasLocation() )
        {
            return true; // this is a PlayerRole that does not depend on others to be determine its position (ignore participation roles status)
        }

        if ( !p2->hasLocation() )
        {
            return false; // participation is a PlayerRole that does not depend on others to be determine its positions, but this does
        }

        /// TODO Both player vars have locations
        /// Not handling players other than ID and Var
        assert ( p1->player()->type() == p2->player()->type() &&
                 p1->player()->type() == TT_VAR_PLAYER );

        const VarPlayer* p2_var = static_cast<const VarPlayer*>( p2->player() );
        // Both this and participation are PlayerRoles => check which one depends on the other
        // Check if 1st depends on 2nd
        PlayerVarVisitor p1_deps;
        p1->location()->visit( p1_deps );

        if ( find( p1_deps.playerVars().begin(),
                   p1_deps.playerVars().end(),
                   p2_var->name()) == p1_deps.playerVars().end() )
        {
            /// P1var depends on P2var
            return true;
        }

        /// Assuming that p2 might depend on p1
        return false;
    }

}
