#ifndef FUNCTOR_NUMBER_H
#define FUNCTOR_NUMBER_H

#include <setplay/model/operation.h>
#include <setplay/model/geom/ipoint.h>
#include <setplay/model/participants/iplayer.h>
#include <setplay/model/numbers/number_var.h>
#include <setplay/model/types.h>

namespace setplay
{

/*!
    \brief Class to calculate the modulus of a number
*/
class NumberAbs
        : public IDecimal
        , public UnaryOperation<NumberAbs, IDecimal>
{
public:
    NumberAbs(IDecimal* number);
    NumberAbs(const NumberAbs& other);
    ~NumberAbs();

    TokenType type() const;

    void visit(IVisitor &visitor);
    NumberAbs& operator*();
    NumberAbs* clone() const;
};

/*!
    \brief Class to change the sign of a number
*/
class NumberSignChange
        : public IDecimal
        , public UnaryOperation<NumberSignChange, IDecimal>
{
public:
    NumberSignChange(IDecimal* number);
    NumberSignChange(const NumberSignChange& other);
    ~NumberSignChange();

    TokenType type() const;

    void visit(IVisitor &visitor);
    NumberSignChange& operator*();
    NumberSignChange* clone() const;
};

/*!
    \brief Class that represents an addition multiplication of 2 numbers
*/
class NumberMultiplication
        : public IDecimal
        , public BinaryOperation<NumberMultiplication, IDecimal>
{
public:
    NumberMultiplication(IDecimal* operand_left, IDecimal* operand_right);
    NumberMultiplication(const NumberMultiplication& other);
    ~NumberMultiplication();

    TokenType type() const;

    void visit(IVisitor &visitor);
    NumberMultiplication& operator*();
    NumberMultiplication* clone() const;
};

/*!
    \brief Class that represents an addition subtraction of 2 numbers
*/
class NumberSubtraction
        : public IDecimal
        , public BinaryOperation<NumberSubtraction, IDecimal>
{
public:
    NumberSubtraction(IDecimal* operand_left, IDecimal* operand_right);
    NumberSubtraction(const NumberSubtraction& other);
    ~NumberSubtraction();

    TokenType type() const;

    void visit(IVisitor &visitor);
    NumberSubtraction& operator*();
    NumberSubtraction* clone() const;
};

/*!
    \brief Class that represents an addition operation of 2 numbers
*/
class NumberAddition
        : public IDecimal
        , public BinaryOperation<NumberAddition, IDecimal>
{
public:
    NumberAddition(IDecimal* operand_left, IDecimal* operand_right);
    NumberAddition(const NumberAddition& other);
    ~NumberAddition();

    TokenType type() const;

    void visit(IVisitor &visitor);
    NumberAddition& operator*();
    NumberAddition* clone() const;
};

/*!
    \brief Class that represents a division operation of 2 numbers
*/
class NumberDivision
        : public IDecimal
        , public BinaryOperation<NumberDivision, IDecimal>
{
public:
    NumberDivision(IDecimal* operand_left, IDecimal* operand_right);
    NumberDivision(const NumberDivision& other);
    ~NumberDivision();

    TokenType type() const;

    void visit(IVisitor &visitor);
    NumberDivision& operator*();
    NumberDivision* clone() const;
};

/*!
    \brief Class to calculate the direction of a vector
*/
class VectorDirection
        : public IDecimal
        , public UnaryOperation<VectorDirection, IPoint2D>
{
public:
    VectorDirection(IPoint2D* point);
    VectorDirection(const VectorDirection& other);
    ~VectorDirection();

    TokenType type() const;

    void visit(IVisitor &visitor);
    VectorDirection& operator*();
    VectorDirection* clone() const;
};

/*!
    \brief Class to calculate the length of a vector
*/
class VectorLength
        : public IDecimal
        , public UnaryOperation<VectorLength, IPoint2D>
{
public:
    VectorLength(IPoint2D* point);
    VectorLength(const VectorLength& other);
    ~VectorLength();

    TokenType type() const;

    void visit(IVisitor &visitor);
    VectorLength& operator*();
    VectorLength* clone() const;
};

/*!
  \brief Class to represent a distance operation between two vectors
  */
class VectorDistance
        : public IDecimal
        , public BinaryOperation<VectorDistance, IPoint2D>
{
public:
    VectorDistance(IPoint2D* operand_left, IPoint2D* operand_right);
    VectorDistance(const VectorDistance& other);
    ~VectorDistance();

    TokenType type() const;

    void visit(IVisitor &visitor);
    VectorDistance& operator*();
    VectorDistance* clone() const;
};

/*!
  \brief Class to represent an angular difference operation between two vectors
  */
class VectorAngularDiff
        : public IDecimal
        , public BinaryOperation<VectorAngularDiff, IPoint2D>
{
public:
    VectorAngularDiff(IPoint2D* operand_left, IPoint2D* operand_right);
    VectorAngularDiff(const VectorAngularDiff& other);
    ~VectorAngularDiff();

    TokenType type() const;

    void visit(IVisitor &visitor);
    VectorAngularDiff& operator*();
    VectorAngularDiff* clone() const;
};

/*!
  \brief Class to represent an inner product operation between two vectors
  */
class VectorInnerProduct
        : public IDecimal
        , public BinaryOperation<VectorInnerProduct, IPoint2D>
{
public:
    VectorInnerProduct(IPoint2D* operand_left, IPoint2D* operand_right);
    VectorInnerProduct(const VectorInnerProduct& other);
    ~VectorInnerProduct();

    TokenType type() const;

    void visit(IVisitor &visitor);
    VectorInnerProduct& operator*();
    VectorInnerProduct* clone() const;
};

/*!
  \brief Class to represent an outer product operation between two vectors
  */
class VectorOuterProduct
        : public IDecimal
        , public BinaryOperation<VectorOuterProduct, IPoint2D>
{
public:
    VectorOuterProduct(IPoint2D* operand_left, IPoint2D* operand_right);
    VectorOuterProduct(const VectorOuterProduct& other);
    ~VectorOuterProduct();

    TokenType type() const;

    void visit(IVisitor &visitor);
    VectorOuterProduct& operator*();
    VectorOuterProduct* clone() const;
};

/*!
  \brief Functor class to determine the defense line (x-axis value) of a given team
  */
class DefenseLine
        : public IDecimal
{
public:
    DefenseLine( TeamAffiliation team_affiliation = TA_UNKNOWN );
    DefenseLine(const DefenseLine& other);

    TokenType type() const;

    TeamAffiliation teamAffiliation() const;

    DefenseLine& setTeamAffiliation(TeamAffiliation team_affiliation);
    DefenseLine& operator=(const DefenseLine& team_affiliation);

    DefenseLine* clone() const;
    DefenseLine& operator*();
    void visit(IVisitor &visitor);
private:
    TeamAffiliation M_team_affiliation;
};

/*!
  \brief Functor class to determine the offense line (x-axis value) of a given team
  */
class OffenseLine
        : public IDecimal
{
public:
    OffenseLine( TeamAffiliation team_affiliation = TA_UNKNOWN );
    OffenseLine(const OffenseLine& other);

    TokenType type() const;

    TeamAffiliation teamAffiliation() const;

    OffenseLine& setTeamAffiliation(TeamAffiliation team_affiliation);
    OffenseLine& operator=(const OffenseLine& team_affiliation);

    OffenseLine* clone() const;
    OffenseLine& operator*();
    void visit(IVisitor &visitor);
private:
    TeamAffiliation M_team_affiliation;
};

class OffsideLine
        : public IDecimal
{
public:
    OffsideLine( TeamAffiliation team_affiliation = TA_UNKNOWN );
    OffsideLine(const OffsideLine& other);

    TokenType type() const;

    TeamAffiliation teamAffiliation() const;

    OffsideLine& setTeamAffiliation(TeamAffiliation team_affiliation);
    OffsideLine& operator=(const OffsideLine& team_affiliation);

    OffsideLine* clone() const;
    OffsideLine& operator*();
    void visit(IVisitor &visitor);
private:
    TeamAffiliation M_team_affiliation;
};

/*!
  \brief Functor object to query a mobile object's speed
  */
class MobileObjectSpeed
        : public IDecimal
{
public:
    MobileObjectSpeed(IMobileObject* mobile_object = NULL);
    MobileObjectSpeed(const MobileObjectSpeed& other);
    ~MobileObjectSpeed();

    TokenType type() const;

    IMobileObject *mobileObject() const;

    MobileObjectSpeed& setMobileObject(IMobileObject* mobile_object);
    MobileObjectSpeed& operator=(const MobileObjectSpeed& team_affiliation);

    MobileObjectSpeed* clone() const;
    MobileObjectSpeed& operator*();
    void visit(IVisitor &visitor);

private:
    IMobileObject* M_mobile_object;
};




/*!
  \brief Functor object to return the current game time
  */
class CurrentTime
        : public IDecimal
{
public:
    CurrentTime();

    TokenType type() const;

    CurrentTime* clone() const;
    CurrentTime& operator*();
    void visit(IVisitor &visitor);
};

/*!
  \brief Functor object to get the time left until the end of the current half of the game
  */
class RemainingTime
        : public IDecimal
{
public:
    RemainingTime();

    TokenType type() const;

    RemainingTime* clone() const;
    RemainingTime& operator*();
    void visit(IVisitor &visitor);
};

/*!
  \brief Functor object to return the number of goals scored by a team
  */
class GoalsScored
        : public IDecimal
{
public:
    GoalsScored(TeamAffiliation team_affiliation);
    GoalsScored(const GoalsScored& other);

    TokenType type() const;
    TeamAffiliation teamAffiliation() const;

    GoalsScored* clone() const;
    GoalsScored& operator*();
    void visit(IVisitor &visitor);
private:
    TeamAffiliation M_team_affiliation;
};

/*!
  \brief Functor object to return the goal difference between our team and the opponent
  */
class GoalDiff
        : public IDecimal
{
public:
    GoalDiff();

    TokenType type() const;

    GoalDiff* clone() const;
    GoalDiff& operator*();
    void visit(IVisitor &visitor);
};

/*!
  \brief Functor object to return the estimated time needed for a player to be able to intercept a mobile object
  */
class InterceptionTime
        : public IDecimal
{
public:
    InterceptionTime(IPlayer* player = NULL,
                     IMobileObject* mobile_object = NULL);
    InterceptionTime(const InterceptionTime& other);
    ~InterceptionTime();

    TokenType type() const;

    InterceptionTime& setPlayer(IPlayer* player);
    InterceptionTime& setMobileObject(IMobileObject* mobile_object);

    IPlayer* player() const;
    IMobileObject* mobileObject() const;

    InterceptionTime* clone() const;
    InterceptionTime& operator*();
    void visit(IVisitor &visitor);
private:
    IPlayer* M_player;
    IMobileObject* M_mobile_object;
};

/*!
  \brief Functor object to return x-axis coordinate of a point
  */
class PointX
        : public IDecimal
{
public:
    PointX(IPoint2D* point = NULL);
    PointX(const PointX& other);
    ~PointX();

    TokenType type() const;

    PointX& setPoint(IPoint2D* point);

    IPoint2D* point() const;

    PointX* clone() const;
    PointX& operator*();
    void visit(IVisitor &visitor);
private:
    IPoint2D* M_point;
};

/*!
  \brief Functor object to return the y-axis coordinate of a point
  */
class PointY
        : public IDecimal
{
public:
    PointY(IPoint2D* point = NULL);
    PointY(const PointY& other);
    ~PointY();

    TokenType type() const;

    PointY& setPoint(IPoint2D* point);

    IPoint2D* point() const;

    PointY* clone() const;
    PointY& operator*();
    void visit(IVisitor &visitor);
private:
    IPoint2D* M_point;
};



}

#endif // FUNCTOR_NUMBER_H
