#include "player_vars_visitor.h"


#include <setplay/model/setplay.h>
#include <setplay/model/actions.h>
#include <setplay/model/conditions.h>
#include <setplay/model/transitions.h>
#include <setplay/model/directives.h>
#include <setplay/model/numbers.h>
#include <setplay/model/geom.h>
#include <setplay/model/assign.h>
#include <setplay/model/types.h>

#include <algorithm>

namespace setplay {

PlayerVarVisitor::PlayerVarVisitor()
{

}

void PlayerVarVisitor::visit(Setplay& setplay)
{
    // Visit parameters
    if ( setplay.hasParameters() )
    {
        for( Parameters::const_iterator itp = setplay.parameters()->begin();
             itp != setplay.parameters()->end(); itp++)
        {
            itp->second->visit( *this );
        }
    }

    // Visit players
    visitPlayers( setplay.players() );

    // Visit abortCond
    if ( setplay.hasAbortCondition() )
    {
        setplay.abortCondition()->visit( *this );
    }

    // Visit steps
    if ( setplay.hasSteps() )
    {
        for( Steps::const_iterator its = setplay.steps()->begin();
             its != setplay.steps()->end(); its++)
        {
            (*its).second->visit( *this );
        }
    }

    /// Visit assignments???
}

void PlayerVarVisitor::visit(Step& step)
{
    if ( step.hasStartCondition() )
    {
        step.startCondition()->visit( *this );
    }

    if ( step.hasAbortCondition() )
    {
        step.abortCondition()->visit( *this );
    }

    if ( step.numParticipations() > 0)
    {
        for ( Participations::const_iterator itp = step.participations()->begin();
              itp != step.participations()->end(); itp++)
        {
            (*itp)->visit( *this );
        }
    }

    // LeadPlayer
    if( step.hasLeader() )
    {
        step.leader()->visit( *this );
    }

    // Transitions
    if ( step.numTransitions() > 0 )
    {
        for( Transitions::const_iterator itt = step.transitions()->begin();
             itt != step.transitions()->end(); itt++)
        {
            (*itt)->visit( *this );
        }
    }
}

void PlayerVarVisitor::visit( IParticipation& participation)
{
    participation.player()->visit( *this );
    if ( participation.hasLocation() )
    {
        participation.location()->visit( *this );
    }
}

/// Transitions
void PlayerVarVisitor::visit(NextStep& transition)
{
    if ( transition.hasStartCondition() )
    {
        transition.startCondition()->visit( *this );
    }

    if ( transition.hasAbortCondition() )
    {
        transition.abortCondition()->visit( *this );
    }

    if ( transition.hasDirectives() )
    {
        visitDirectives( transition.directives() );
    }
}

void PlayerVarVisitor::visit(Finish& transition)
{
    if ( transition.hasStartCondition() )
    {
        transition.startCondition()->visit( *this );
    }

    if ( transition.hasDirectives() )
    {
        visitDirectives( transition.directives() );
    }
}

void PlayerVarVisitor::visit(Abort& transition)
{
    if ( transition.hasStartCondition() )
    {
        transition.startCondition()->visit( *this );
    }

    if ( transition.hasDirectives() )
    {
        visitDirectives( transition.directives() );
    }

}

void PlayerVarVisitor::visit(Do& directive)
{
    visitDirective( directive);
}

void PlayerVarVisitor::visit( DoWhile& directive)
{
    directive.condition()->visit( *this );
    visitDirective( directive );
}

void PlayerVarVisitor::visit(Dont& directive)
{
    visitDirective( directive );
}


void PlayerVarVisitor::visit(ActDribble& action)
{
    if ( action.location() != NULL)
    {
        action.location()->visit( *this );
    }
}

void PlayerVarVisitor::visit(ActFollow& action)
{
    /// Target
    if ( action.target() != NULL )
    {
        action.target()->visit( *this );
    }

    /// Displacement
    if (action.displacement() != NULL)
    {
        action.displacement()->visit( *this );
    }
}


void PlayerVarVisitor::visit(ActHold& )
{

}

void PlayerVarVisitor::visit(ActIntercept& action)
{
    if ( action.target() != NULL )
    {
        action.target()->visit( *this );
    }
}

void PlayerVarVisitor::visit(ActPass& action)
{
    if ( action.location() != NULL )
    {
        action.location()->visit( *this );
    }
}

void PlayerVarVisitor::visit(ActPos& action)
{
    if ( action.location() != NULL )
    {
        action.location()->visit( *this );
    }
}


void PlayerVarVisitor::visit(ActShoot& )
{

}

void PlayerVarVisitor::visit( ActLookAt& action)
{
    for ( vector<IPlayer*>::const_iterator itp = action.players().begin(); itp != action.players().end(); itp++ )
    {
        (*itp)->visit( *this );
    }
}


/// Conditions


/// CondAnd
void PlayerVarVisitor::visit(CondAnd& condition)
{
    visitConditions( condition.conditions() );
}

/// CondOr
void PlayerVarVisitor::visit(CondOr& condition)
{
    visitConditions( condition.conditions() );
}


/// CondNot
void PlayerVarVisitor::visit(CondNot& condition)
{
    condition.condition()->visit( *this );
}

void PlayerVarVisitor::visit(CondBallOwner& condition)
{
    condition.player()->visit( *this );
}

void PlayerVarVisitor::visit(CondTeamInPossession& condition)
{

}

void PlayerVarVisitor::visit(CondBool& condition)
{

}

void PlayerVarVisitor::visit(CondGameSituation& condition)
{

}

void PlayerVarVisitor::visit(CondAtLocation& condition)
{
    condition.object()->visit( *this );
    condition.location()->visit( *this );
}

void PlayerVarVisitor::visit(CondTeamFormation& condition)
{

}

void PlayerVarVisitor::visit(CondOffside& condition)
{
    condition.player()->visit( *this );
}

void PlayerVarVisitor::visit(CondCompDecimal& condition)
{
    condition.operandLeft()->visit( *this );
    condition.operandRight()->visit( *this );
}

void PlayerVarVisitor::visit(CondKickable& condition)
{
    condition.player()->visit( *this );
}

void PlayerVarVisitor::visit( CondCanIntercept& condition)
{
    condition.player()->visit( *this );
    condition.mobileObject()->visit( *this );
}

void PlayerVarVisitor::visit( CondCanShoot& condition)
{
    condition.player()->visit( *this );
}

void PlayerVarVisitor::visit( CondCanPass& condition)
{
    condition.player()->visit( *this );
    condition.location()->visit( *this );
}

/// Locations
void PlayerVarVisitor::visit(Rect2D& location)
{
    location.topLeft()->visit( *this );
    location.bottomRight()->visit( *this );
}

void PlayerVarVisitor::visit(Triangle2D& location)
{
    location.a()->visit( *this );
    location.b()->visit( *this );
    location.c()->visit( *this );
}

void PlayerVarVisitor::visit(Sector2D& location)
{
    location.center()->visit( *this );
    location.minRadius()->visit( *this );
    location.maxRadius()->visit( *this );
    location.startAngle()->visit( *this );
    location.endAngle()->visit( *this );
}

void PlayerVarVisitor::visit(Circle2D& location)
{
    location.center()->visit( *this );
    location.radius()->visit( *this );
}

void PlayerVarVisitor::visit(Ray2D& location)
{
    location.origin()->visit( *this );
    location.direction()->visit( *this );
}

void PlayerVarVisitor::visit(Line2D& location)
{
    location.a()->visit( *this );
    location.b()->visit( *this );
    location.c()->visit( *this );
}

/// Participant
void PlayerVarVisitor::visit(PlayerAffiliation& participant)
{

}

/// Player functors
void PlayerVarVisitor::visit( PlayerSelf& player_functor)
{

}

void PlayerVarVisitor::visit( NearestPlayer& player_functor)
{
    player_functor.location()->visit( *this );
}

void PlayerVarVisitor::visit( NearestTeammate& player_functor)
{
    player_functor.location()->visit( *this );
}

void PlayerVarVisitor::visit( NearestOpponent& player_functor)
{
    player_functor.location()->visit( *this );
}

void PlayerVarVisitor::visit( Goalie& )
{

}

/// Points

void PlayerVarVisitor::visit(Point2D& point)
{
    point.x()->visit( *this );
    point.y()->visit( *this );
}

/// Variables

void PlayerVarVisitor::visit( Variable& )
{
    // TODO If type is Player => Write
}

void PlayerVarVisitor::visit(VarLocation& variable)
{

}

void PlayerVarVisitor::visit( VarRegion& variable)
{

}

void PlayerVarVisitor::visit( VarPoint& variable)
{

}

void PlayerVarVisitor::visit(VarCondition& variable)
{

}

void PlayerVarVisitor::visit(VarDecimal& variable)
{

}

void PlayerVarVisitor::visit(VarInteger& variable)
{

}

void PlayerVarVisitor::visit(VarPlayer& variable)
{
    if ( std::find( M_player_vars.begin(),
                    M_player_vars.end(),
                    variable.name() ) == M_player_vars.end() )
    {
        M_player_vars.push_back( variable.name() );
    }
}

/// Point Functors
void PlayerVarVisitor::visit( MobileObjectLocation& point_functor)
{
    point_functor.mobileObject()->visit( *this );
}

void PlayerVarVisitor::visit(PlayerPosition& point_functor)
{
    point_functor.player()->visit( *this );
}

void PlayerVarVisitor::visit( BallPosition& )
{

}

// Vector unary operations
void PlayerVarVisitor::visit( VectorNormalize& point_functor )
{
    point_functor.operand()->visit( *this );
}

void PlayerVarVisitor::visit( VectorInvertX& point_functor )
{
    point_functor.operand()->visit( *this );
}

void PlayerVarVisitor::visit( VectorInvertY& point_functor )
{
    point_functor.operand()->visit( *this );
}

void PlayerVarVisitor::visit( VectorInvertXY& point_functor )
{
    point_functor.operand()->visit( *this );
}

void PlayerVarVisitor::visit( VectorSubtraction& point_functor )
{
    point_functor.operandLeft()->visit( *this );
    point_functor.operandRight()->visit( *this );
}

void PlayerVarVisitor::visit( VectorAddition& point_functor )
{
    point_functor.operandLeft()->visit( *this );
    point_functor.operandRight()->visit( *this );
}

// Vector scalar operations
void PlayerVarVisitor::visit( VectorScale& point_functor )
{
    point_functor.operandLeft()->visit( *this );
    point_functor.operandRight()->visit( *this );
}

void PlayerVarVisitor::visit( VectorRotation& point_functor )
{
    point_functor.operandLeft()->visit( *this );
    point_functor.operandRight()->visit( *this );
}

void PlayerVarVisitor::visit( VectorDivision& point_functor )
{
    point_functor.operandLeft()->visit( *this );
    point_functor.operandRight()->visit( *this );
}

void PlayerVarVisitor::visit( NearestIntersectionPoint& point_functor)
{
    point_functor.operandLeft()->visit( *this );
    point_functor.operandRight()->visit( *this );
}

void PlayerVarVisitor::visit( RegionCentroid& point_functor)
{
    point_functor.operand()->visit( *this );
}

/// Basic Data Types
void PlayerVarVisitor::visit(Decimal& value)
{

}

void PlayerVarVisitor::visit(Integer& value)
{

}

void PlayerVarVisitor::visit(Bool& value)
{

}

/// Decimal Functors

// Number operations
void PlayerVarVisitor::visit( NumberAbs& decimal_functor )
{
    decimal_functor.operand()->visit( *this );
}

void PlayerVarVisitor::visit( NumberSignChange& decimal_functor )
{
    decimal_functor.operand()->visit( *this );
}

void PlayerVarVisitor::visit( NumberMultiplication& decimal_functor )
{
    decimal_functor.operandLeft()->visit( *this );
    decimal_functor.operandRight()->visit( *this );
}

void PlayerVarVisitor::visit( NumberSubtraction& decimal_functor )
{
    decimal_functor.operandLeft()->visit( *this );
    decimal_functor.operandRight()->visit( *this );
}

void PlayerVarVisitor::visit( NumberDivision& decimal_functor )
{
    decimal_functor.operandLeft()->visit( *this );
    decimal_functor.operandRight()->visit( *this );
}

void PlayerVarVisitor::visit( NumberAddition& decimal_functor )
{
    decimal_functor.operandLeft()->visit( *this );
    decimal_functor.operandRight()->visit( *this );
}

// Vector to scalar operations
void PlayerVarVisitor::visit( VectorDirection& decimal_functor )
{
    decimal_functor.operand()->visit( *this );
}

void PlayerVarVisitor::visit( VectorLength& decimal_functor )
{
    decimal_functor.operand()->visit( *this );
}

// Two vector to scalar operations
void PlayerVarVisitor::visit( VectorDistance& decimal_functor )
{
    decimal_functor.operandLeft()->visit( *this );
    decimal_functor.operandRight()->visit( *this );
}

void PlayerVarVisitor::visit( VectorAngularDiff& decimal_functor )
{
    decimal_functor.operandLeft()->visit( *this );
    decimal_functor.operandRight()->visit( *this );
}

void PlayerVarVisitor::visit( VectorInnerProduct& decimal_functor )
{
    decimal_functor.operandLeft()->visit( *this );
    decimal_functor.operandRight()->visit( *this );
}

void PlayerVarVisitor::visit( VectorOuterProduct& decimal_functor )
{
    decimal_functor.operandLeft()->visit( *this );
    decimal_functor.operandRight()->visit( *this );
}

// In game queries

void PlayerVarVisitor::visit(DefenseLine& decimal_functor)
{

}

void PlayerVarVisitor::visit(OffenseLine& decimal_functor)
{

}

void PlayerVarVisitor::visit(OffsideLine& decimal_functor)
{

}

void PlayerVarVisitor::visit(MobileObjectSpeed& decimal_functor)
{
    decimal_functor.mobileObject()->visit( *this );
}

void PlayerVarVisitor::visit( PointX& decimal_functor )
{
    decimal_functor.point()->visit( *this );
}

void PlayerVarVisitor::visit( PointY& decimal_functor )
{
    decimal_functor.point()->visit( *this );
}


/// Integer Functors
void PlayerVarVisitor::visit(CurrentTime& )
{

}

void PlayerVarVisitor::visit(RemainingTime& )
{

}

void PlayerVarVisitor::visit( GoalsScored& integer_functor)
{

}

void PlayerVarVisitor::visit( GoalDiff& integer_functor)
{

}

void PlayerVarVisitor::visit( InterceptionTime& integer_functor )
{
    integer_functor.player()->visit( *this );
    integer_functor.mobileObject()->visit( *this );
}

void PlayerVarVisitor::visit( Assignment& assign )
{
    assign.variable()->visit( *this );
    assign.expression()->visit( *this );
}

void PlayerVarVisitor::visit( Ball& )
{

}

/// Helper methods implementation

void PlayerVarVisitor::visitPlayers( vector<IPlayer* >* participants)
{
    for( Participants::const_iterator itp = participants->begin();
             itp != participants->end(); itp++ )
    {
            ( *itp )->visit( *this );
    }
}

void PlayerVarVisitor::visitDirectives( Directives* directives )
{
    for ( Directives::const_iterator itd = directives->begin();
          itd != directives->end(); itd++ )
    {
        (*itd)->visit( *this );
    }
}

void PlayerVarVisitor::visitDirective( IDirective &directive)
{
    // Visit players
    visitPlayers( directive.players() );

    // Actions to execute
    visitActions( directive.actions() );
}

void PlayerVarVisitor::visitActions( Actions* actions )
{
    for ( Actions::const_iterator ita = actions->begin();
          ita != actions->end(); ita++)
    {
        (*ita)->visit( *this );
    }
}

void PlayerVarVisitor::visitConditions( Conditions* conditions )
{
    for ( Conditions::const_iterator itc = conditions->begin();
          itc != conditions->end(); itc++)
    {
        (*itc)->visit( *this );
    }
}

}
