#include "step.h"

#include <setplay/model/ivisitor.h>
#include <setplay/model/transitions/next_step.h>

namespace setplay {


    const unsigned int Step::FIRST_STEP_ID = 0;

    Step::Step()
        : M_id(0u)
        , M_leader( NULL )
        , M_transitions( NULL )
        , M_participations( NULL )
        , M_abort_condition( NULL )
        , M_start_condition( NULL )
        , M_wait_condition( NULL )
    {
        M_transitions = new Transitions();
        M_participations = new Participations();
        M_assignments = new Assignments();
    }

    Step::Step(const Step &other)
        : M_id( other.M_id )
    {
        M_start_condition = other.M_start_condition != NULL
                               ? other.M_start_condition->clone()
                               : NULL;

        M_abort_condition = other.M_abort_condition != NULL
                               ? other.M_abort_condition->clone()
                               : NULL;

        M_wait_condition = other.M_wait_condition != NULL
                               ? other.M_wait_condition->clone()
                               : NULL;

        M_leader = other.M_leader != NULL
                     ? other.M_leader->clone()
                     : NULL;

        M_transitions = new Transitions();
        if ( other.M_transitions != NULL )
        {
            for ( Transitions::const_iterator itt = other.M_transitions->begin();
                  itt != other.M_transitions->end(); itt++)
            {
                addTransition( (*itt)->clone() );
            }
        }

        M_participations = new Participations();
        if ( other.M_participations != NULL )
        {
            for ( Participations::const_iterator itp = other.M_participations->begin();
                  itp != other.M_participations->end(); itp++)
            {
                addParticipation( (*itp)->clone() );
            }
        }

        M_assignments = new Assignments();
        if ( other.M_assignments != NULL )
        {
            for ( Assignments::const_iterator ita = other.M_assignments->begin();
                  ita != other.M_assignments->end(); ita++)
            {
                addAssignment( (*ita)->clone() );
            }
        }
    }

    Step::~Step()
    {
        if ( M_start_condition != NULL )
        {
            delete M_start_condition;
        }

        if ( M_abort_condition != NULL )
        {
            delete M_abort_condition;
        }

        if ( M_leader != NULL )
        {
            delete M_leader;
        }

        if ( M_participations != NULL )
        {
            for ( Participations::iterator itp = M_participations->begin();
                  itp != M_participations->end();
                  itp++)
            {
                delete *itp;
            }
        }

        if ( M_transitions != NULL )
        {
            for ( Transitions::iterator itt = M_transitions->begin();
                  itt != M_transitions->end();
                  itt++)
            {
                delete *itt;
            }
        }

        if ( M_assignments != NULL )
        {
            for ( Assignments::iterator ita = M_assignments->begin();
                  ita != M_assignments->end();
                  ita++)
            {
                delete *ita;
            }
            delete M_assignments;
        }
    }

    Participations *Step::participations() const
    {
        return M_participations;
    }

    Assignments* Step::assignments() const
    {
        return M_assignments;
    }

    Participants Step::participants() const
    {
        return M_participants;
    }

    Transitions* Step::transitions() const
    {
        return M_transitions;
    }


    unsigned int Step::id () const
    {
        return M_id;
    }

    IPlayer *Step::leader() const
    {
        return M_leader;
    }

    bool Step::hasLeader() const
    {
        return M_leader != NULL;
    }

    bool Step::hasAbortCondition() const
    {
        return M_abort_condition != NULL;
    }

    bool Step::hasStartCondition() const
    {
        return M_start_condition != NULL;
    }

    bool Step::hasWaitCondition() const
    {
        return M_wait_condition != NULL;
    }

    unsigned int Step::numTransitions() const
    {
        return (M_transitions != NULL && !M_transitions->empty())
                ? M_transitions->size()
                : 0;
    }

    unsigned int Step::numParticipations() const
    {
        return (M_participations != NULL && !M_participations->empty())
                ? M_participations->size()
                : 0;
    }

    bool Step::hasAssignments() const
    {
        return M_assignments != NULL && !M_assignments->empty();
    }

    ICondition* Step::abortCondition() const
    {
        return M_abort_condition;
    }

    ICondition* Step::startCondition() const
    {
        return M_start_condition;
    }

    ICondition* Step::waitCondition() const
    {
        return M_wait_condition;
    }


    // FA 10/06/11 - Check if this is the first step
    bool Step::is_first() const
    {
        return M_id == FIRST_STEP_ID;
    }

    // FA 14/06/11 - Get transition that leads to a given step or NULL otherwise
    /**
     * @param new_step_id The step number to which the retrieved transition should lead
     * @returns A pointer to the transition or NULL otherwise
     */
    ITransition* Step::getTransitionToStep(unsigned int new_step_id) const
    {
        for (unsigned int i = 0; i < M_transitions->size() ; i++)
        {
            if (M_transitions->at(i)->type() == TT_TRANSITION_NEXT_STEP &&
                ((NextStep*) M_transitions->at(i))->nextStepId() == new_step_id)
            {
                return M_transitions->at(i); // Found a transition that leads to the next step
            }
        }
        return NULL; // No next step transition found to the step number

    }

    // Modifiers
    Step& Step::setId(unsigned int id)
    {
        M_id = id;
        return *this;
    }

    Step& Step::setLeader(IPlayer* leader)
    {
        if ( M_leader != NULL ) { delete M_leader; }
        M_leader = leader;
        return *this;
    }

    Step& Step::setStartCondition(ICondition* condition)
    {
        if ( M_start_condition != NULL ) { delete M_start_condition; }
        M_start_condition = condition;
        return *this;
    }

    Step& Step::setAbortCondition(ICondition* condition)
    {
        if ( M_abort_condition != NULL ) { delete M_abort_condition; }
        M_abort_condition = condition;
        return *this;
    }

    Step& Step::setWaitCondition(ICondition* condition)
    {
        if ( M_wait_condition != NULL ) { delete M_wait_condition; }
        M_wait_condition = condition;
        return *this;
    }

    Step& Step::setParticipations(Participations* participations)
    {
        if ( M_participations != NULL )
        {
            for ( Participations::iterator it = M_participations->begin();
                  it != M_participations->end();
                  it++)
            {
                delete *it;
            }
            delete M_participations;
        }
        M_participations = participations;
        return *this;
    }


    Step& Step::setAssignments(Assignments* assignments)
    {
        if ( M_assignments != NULL )
        {
            for ( Assignments::iterator it = M_assignments->begin();
                  it != M_assignments->end(); it++)
            {
                delete *it;
            }
            delete M_assignments;
        }
        M_assignments = assignments;
        return *this;
    }

    Step& Step::addParticipation(IParticipation* participation)
    {
        M_participations->push_back(participation);
        M_participants.push_back(participation->player());
        return *this;
    }

    Step& Step::setTransitions(Transitions* transitions)
    {
        if ( M_transitions != NULL )
        {
            for ( Transitions::iterator it = M_transitions->begin();
                  it != M_transitions->end();
                  it++)
            {
                delete *it;
            }
            delete M_transitions;
        }
        M_transitions = transitions;
        return *this;
    }

    Step& Step::addTransition(ITransition* transition)
    {
        M_transitions->push_back( transition );
        return *this;
    }

    Step& Step::addAssignment(Assignment *assignment)
    {
        M_assignments->push_back( assignment );
        return *this;
    }

    void Step::visit(IVisitor& visitor)
    {
        visitor.visit( *this );
    }

    Step* Step::clone() const
    {
        return new Step( *this );
    }

}
