#include "player_selector.h"

#include <setplay/runtime/setplay_interpreter.h>

#include <setplay/generators/sexpr/setplay_generator_sexpr.h>
#include <setplay/model/player_vars_visitor.h>

#include <lemon/matching.h>
#include <lemon/concepts/graph.h>
#include <lemon/list_graph.h>

#include <rcsc/common/logger.h>

using namespace lemon;
using namespace lemon::concepts;

namespace setplay {

PlayersMinGlobalDistanceSelector::PlayersMinGlobalDistanceSelector( IWorldModel* world )
    : M_world( world )
{

}

PlayersMinGlobalDistanceSelector::~PlayersMinGlobalDistanceSelector()
{

}

bool minWeightBipartiteMatching( SetplayInterpreter& interpreter,
                                 const std::map<std::string, rcsc::Vector2D>& player_vars,
                                 const std::map<unsigned int, rcsc::Vector2D>& concrete_players)
{
    ListGraph bipartite_graph;

    // Initialize player nodes
    std::map<unsigned int, ListGraph::Node> concrete_player_to_node_map;
    std::map<ListGraph::Node, unsigned int> node_to_concrete_player_map;

    // Create a node each concrete player
    for ( std::map<unsigned int, rcsc::Vector2D>::const_iterator itp = concrete_players.begin();
          itp != concrete_players.end(); itp++)
    {
        ListGraph::Node node = bipartite_graph.addNode();
        concrete_player_to_node_map.insert(std::pair<unsigned int, ListGraph::Node>( itp->first, node));
        node_to_concrete_player_map.insert(std::pair<ListGraph::Node, unsigned int>( node, itp->first));
    }

    std::map<string, ListGraph::Node> player_var_to_node_map;

    // Create a node each player variable
    for ( std::map<string, rcsc::Vector2D>::const_iterator itp = player_vars.begin();
          itp != player_vars.end(); itp++)
    {
        player_var_to_node_map.insert(std::pair<string, ListGraph::Node>( itp->first, bipartite_graph.addNode()) );
    }

    MaxWeightedMatching<ListGraph,
                        ListGraph::EdgeMap<float> >::WeightMap players_weight_values(bipartite_graph);

    // Calculate players positionings
    for ( std::map<unsigned int, rcsc::Vector2D>::const_iterator itc = concrete_players.begin( );
          itc != concrete_players.end(); itc++)
    {
        for ( std::map<std::string, rcsc::Vector2D>::const_iterator itv = player_vars.begin( );
              itv != player_vars.end(); itv++)
        {
            ListGraph::Edge edge = bipartite_graph.addEdge( concrete_player_to_node_map.at( itc->first ),
                                                            player_var_to_node_map.at( itv->first ));
            // NOTE: cannot use negative value. Figure out another way to do this
            float weight =  1 / itc->second.dist2( itv->second );
            players_weight_values[edge] = abs(weight);
            rcsc::dlog.addText( rcsc::Logger::PLAN,
                                __FILE__ " (%d, %s)=%.5f",
                                itc->first,
                                itv->first.c_str(),
                                weight);
        }
    }

    MaxWeightedMatching<ListGraph,
                        ListGraph::EdgeMap<float> > matching(bipartite_graph,
                                                             players_weight_values);
    matching.run();

    /// Check that the matching has an appropriate number of elements
    // std::cout << "Matching size " << matching.matchingSize() << std::endl;
    if ( matching.matchingSize() != (int) player_vars.size() )
    {
        return false;
    }

    for ( std::map<std::string, rcsc::Vector2D>::const_iterator itp = player_vars.begin();
          itp != player_vars.end();
          itp++ )
    {
        unsigned int matched_pl_unum = node_to_concrete_player_map.at( matching.mate( player_var_to_node_map[ itp->first ] ) );
        rcsc::dlog.addText( rcsc::Logger::PLAN,
                            __FILE__ " Registering %s as #%d",
                            itp->first.c_str(),
                            matched_pl_unum);
        interpreter.push( itp->first, new setplay::PlayerAffiliation( matched_pl_unum, TA_OUR ));
    }
    return true;
}

bool PlayersMinGlobalDistanceSelector::selectPlayers(SetplayInterpreter &interpreter )
{
    // Leader of the first step will be self?

    vector<std::string> players_instantiated;

    // 1) Determine eligible players and their respective positions
    vector<unsigned int> eligible_players = M_world->activePlayers( TA_OUR );
    rcsc::dlog.addText( rcsc::Logger::PLAN,
                        __FILE__ " PlayerSelector\n Initial eligible players=%d",
                        eligible_players.size());

    // 2) Determine players variables positions
    const Setplay* setplay = interpreter.currentSetplay();

    /// Assuming all parameters are players for now
    if ( eligible_players.size() < setplay->parameters()->size() )
    {
        rcsc::dlog.addText( rcsc::Logger::PLAN,
                            __FILE__ " Not enough eligible players to instantiante set-play players %d < %d",
                            eligible_players.size(), setplay->parameters()->size());
        return false;
    }

    /// Get eligible player positions
    map<unsigned int, rcsc::Vector2D> eligible_players_positions;
    rcsc::dlog.addText( rcsc::Logger::PLAN,
                        __FILE__ " PlayerSelector: Eligible players positions");
    for ( vector<unsigned int>::const_iterator itu = eligible_players.begin();
          itu != eligible_players.end() ; itu++ )
    {
        rcsc::Vector2D ppos = M_world->playerPosition( PlayerAffiliation(*itu, TA_OUR) );
        if ( ppos == rcsc::Vector2D::INVALIDATED )
        {
            rcsc::dlog.addText( rcsc::Logger::PLAN,
                                __FILE__ " _PlayerSelector: Could not find %d position",
                                *itu);
            continue;
        }
        eligible_players_positions.insert(pair<unsigned int,
                                          rcsc::Vector2D>( *itu, ppos));
        rcsc::dlog.addText( rcsc::Logger::PLAN,
                            __FILE__  " (%d,(%.2f,%.2f)) ",
                            *itu, ppos.x, ppos.y);
    }

    /// Assuming all parameters are players for now
    if ( eligible_players_positions.size() < setplay->parameters()->size() )
    {
        rcsc::dlog.addText( rcsc::Logger::PLAN,
                            __FILE__ " Not enough eligible players and positions to instantiante set-play players %d < %d",
                            eligible_players.size(), setplay->parameters()->size());
        return false;
    }

    const Step* first_step = setplay->firstStep();

    std::map<std::string, rcsc::Vector2D> player_vars_positions;

    bool self_allocated = false;

    // Assuming that all possible dependencies between player locations have been met
    for ( Participations::const_iterator itp = first_step->participations()->begin();
          itp != first_step->participations()->end();
          itp++)
    {
        /// Players: PlayerAffiliation, PlayerVar, NearestPlayer, NearestTeammate, NearestOpponent, Goalie
        stringstream ss;
        SetplayGeneratorSexpr sexpr( ss );
        IParticipation* participation = *itp;
        participation->visit( sexpr );
        ss << std::endl;

        if ( participation->player()->type() == TT_PLAYER_AFFILIATION )
        {
            // Concrete player => No need to instantiate, must remove from eligible players
            const PlayerAffiliation* player_affiliation = static_cast<PlayerAffiliation*>( participation->player() );
            map<unsigned int, rcsc::Vector2D>::iterator itp = eligible_players_positions.find( player_affiliation->number() );
            if ( itp == eligible_players_positions.end() )
            {
                rcsc::dlog.addText( rcsc::Logger::PLAN,
                                    __FILE__ " Player #%d is not eligible!!! ",
                                    player_affiliation->number());
                assert (false );
            }
            eligible_players_positions.erase( itp );
            continue;
        }
        rcsc::dlog.addText( rcsc::Logger::PLAN,
                            __FILE__ " PlayerSelector: Participations\n %s", ss.str().c_str());

        /// Assume that the player is a PlayerVar!!!!
        assert( participation->player()->type() == TT_VAR_PLAYER );
        const VarPlayer* player_var = static_cast<VarPlayer*>( participation->player() );
        if ( !participation->hasLocation() )
        {
            rcsc::dlog.addText( rcsc::Logger::PLAN,
                                __FILE__ " Participation %s has no location",
                                player_var->name().c_str());
            // Free participant? Assume that it will be self
            if ( !self_allocated )
            {
                interpreter.push( player_var->name(), new PlayerAffiliation( M_world->self(), TA_OUR ));
                rcsc::dlog.addText( rcsc::Logger::PLAN,
                                    __FILE__ " Registered %s as self",
                                    player_var->name().c_str());
                map<unsigned int, rcsc::Vector2D>::iterator its = eligible_players_positions.find( M_world->self() );
                if ( its == eligible_players_positions.end() )
                {
                    rcsc::dlog.addText( rcsc::Logger::PLAN,
                                        __FILE__ " Self is not eligible for player allocation");
                    assert( false );
                }
                eligible_players_positions.erase( its );
                players_instantiated.push_back( player_var->name() );
                interpreter.push( player_var->name(), new PlayerAffiliation( M_world->self(), TA_OUR ));
                self_allocated = true;
                continue;
            }
            rcsc::dlog.addText( rcsc::Logger::PLAN,
                                __FILE__ " TODO More than one free participant? Who should we choose!!! ");
            assert( false );
        }

        // Player location can be safely determined? Supposedly, as it shouln't depend on remaining players
        // TODO To fully solve this one must use the notion of allocation groups that rely on dependencies between players
        // For now this will only with one free player and others depending on his location
        switch ( participation->location()->locationType() )
        {
            case ILocation2D::LT_POINT:
            {
                const IPoint2D* ipoint = static_cast<IPoint2D*>( participation->location() );
                rcsc::Vector2D position = interpreter.interpret( *ipoint );
                player_vars_positions.insert(pair<std::string, rcsc::Vector2D>( player_var->name(),
                                                                                position));
            }
            break;
            case ILocation2D::LT_REGION:
            {
                const IRegion2D* iregion = static_cast<IRegion2D*>( participation->location() );
                rcsc::Region2D* region = interpreter.interpret( *iregion );
                rcsc::Vector2D position = region->centroid();
                player_vars_positions.insert(pair<std::string, rcsc::Vector2D>( player_var->name(),
                                                                                position));
                delete region;
            }
            break;
            case ILocation2D::LT_MOBILE_OBJECT:
            {
                const MobileObjectLocation* mol = static_cast<MobileObjectLocation*>( participation->location() );
                switch ( mol->mobileObject()->mobileObjectType() )
                {
                    case IMobileObject::MOT_BALL:
                    {
                        rcsc::Vector2D position = M_world->ballPosition();
                        player_vars_positions.insert(pair<std::string, rcsc::Vector2D>( player_var->name(),
                                                                                        position));
                    }
                    break;
                    case IMobileObject::MOT_PLAYER:
                    {
                        const IPlayer* iplayer = static_cast<IPlayer*>( mol->mobileObject() );
                        PlayerAffiliation player = interpreter.interpret( *iplayer );
                        rcsc::Vector2D position = M_world->playerPosition( player );
                        player_vars_positions.insert(pair<std::string, rcsc::Vector2D>( player_var->name(),
                                                                                        position));
                    }
                    break;
                }
            }
            break;
            default:
            {
                std::cerr << " Unhandled location type " << participation->location()->locationType() << std::endl;
                assert(false);
            }
                break;
        }
    }

    rcsc::dlog.addText( rcsc::Logger::PLAN,
                        __FILE__ " Players remaining for allocation=%d, Eligible=%d",
                        player_vars_positions.size(),
                        eligible_players_positions.size());
    // Map player vars to players available according to their distance to the points
    bool res = minWeightBipartiteMatching( interpreter,
                                           player_vars_positions,
                                           eligible_players_positions );
    if ( !res )
    {
        rcsc::dlog.addText( rcsc::Logger::PLAN,
                            __FILE__ " PlayerSelector: Could not allocate all players needed for setplay");
        // Remove all temporary instantiations made for players
        for ( std::vector<std::string>::const_iterator itp = players_instantiated.begin();
              itp != players_instantiated.end(); itp++ )
        {
            interpreter.unset( *itp );
        }
    }
    return res;
}

/*
    map<string, PlayerAffiliation*>*
    SetplayInterpreter::participatingPlayersDistanceFromSelf()
    {
        vector<unsigned int> players;
        //LMOTA TODO
        assert(false);

        return NULL;
    }

    map<std::string, PlayerAffiliation*>*
    SetplayInterpreter::participatingPlayersDistanceFromPositions()
    {
      //DEBUG
      cerr<<"SP: using normal distances"<<endl;

      // Will output player numbers starting with 1, up to max 11...

      //Jersey number of players participating
      vector<unsigned int>* participating_players = new vector<unsigned int>();

      const Step::Participations* first_step_participations = M_active_setplay->firstStep()->participations();

      // 30/06/2011 FA Get eligible players
      vector<const IPlayerID*>* eligible_players = M_context_provider->eligible_plan_participants(M_active_setplay->participants()->size());

      // Should first try to decide who is the lead player


      const IPlayerID* first_step_leader = getPlayerID( M_active_setplay->firstStep()->leader() );
      assert(first_step_leader!= NULL);

      //Go through all players and choose closest player
      for(unsigned int i = 0; i!= M_active_setplay->participants()->size(); i++)
      {
        const IPlayerID* setplay_participant = getPlayerID( M_active_setplay->participants()->at(i) );

        assert(setplay_participant != NULL);


        // If participation is the leadplayer, or is a playerId with this player's
        // number, then it will be taken by this player
        if(*setplay_participant == *first_step_leader )
        { // TODO Check if this equality can be tested properly !!!!
           participating_players->push_back(M_self->number());
           // TODO Player instantiation
           // ((IPlayerID*) players->at(i))->set(new IPlayerID(TA_OUR,me->number));
           cout << "Chosen leader " << M_self->number() << endl;
       }
       else
       {
          // Current player player reference is a role
          float distance=2000;//initial distance very high
          int chosenPlayer=0;
          rcsc::Vector2D pos = rcsc::Vector2D::INVALIDATED;

          for(unsigned int j=0; j < eligible_players->size(); j++)
          { // j!= M_context_provider->numPlayersPerTeam(); j++){
            // Will check all players, except the goalie, which can only be part
            // of a setplay when starting it
            // Must first check if player has already been chosen, or if it's me,
            // in which case it cannot be chosen
            pos = rcsc::Vector2D::INVALIDATED;
            bool already_chosen = false;
            for(unsigned int k=0; k!= participating_players->size();k++)

              // Check if player was already chosen
              if( eligible_players->at(j)->number() == participating_players->at(k) ||
                  eligible_players->at(j)->number() == M_self->number())
              {
                already_chosen =true;
                break;
              }

              if ( !already_chosen  )
              { //check if distance is smaller than current
                //Must first get corresponding location in participation vector
                for(unsigned int k=0; k!= first_step_participations->size();k++)
                {
                  const IParticipation* first_step_participation = first_step_participations->at(k);

                  const IPlayerID* first_step_participant = getPlayerID(first_step_participation->player());

                  if(*setplay_participant == *first_step_participant)
                  {

                   if( first_step_participation->has_location() )
                   {
                      // Figure out the point location (assume that finding a centralPoint location for all regions is sufficient to figure out the best player choice)
                      // TODO Figure out player instantiation
                      const IPlayerID* player_var = first_step_participation->player();
                      if ( !player_var )
                      {     // TODO Will always be non-instantiated as no prior instantiation is being made before calling this methods
                            cerr << "Player Role isn't instantiated" << endl;
                            // player_var->print(cerr);
                            // Figure out how to decide player instantiation

                             // pos=new IPoint2D(*participations->at(k)->location()
                               //                 ->centroid(M_context_provider)
                                 //               ->asPoint2D(M_context_provider));
                      }
                    } else
                   { // Location is null
                            // Decide which player should fullfil the current role
                            // Is current player the lead player
                    }

                  }
                }


                // if location is not defined, this player will be chosen,
                // since there is no other criterion...
                if (pos==NULL)
                {
                   chosenPlayer=j;
                   break;
                }
                else
                {
                     const IPlayerID* our_current_player = eligible_players->at(j);
                     if( M_context_provider->playerPosition( *our_current_player)!= NULL &&
                         pos->distance(*M_context_provider->playerPosition(*our_current_player), M_context_provider) < distance)
                     {  // Assuming that player returns a position!!
                        // TODO Register this player's name in the variable registry
                        // distance = pos->distance(*M_context_provider->playerPosition(*our_current_player), M_context_provider);
                        chosenPlayer=j;
                    }
                }
            }
            //else
              //DEBUG
              //cerr<<"SP: player "<<j+1<<" discarded, was already chosen!"<<endl;
          }
          //DEBUG
          // cerr<<"SP: "<<i<<"th Position being searched:";
    //       if(pos!=NULL)pos->print(cerr);
    //       cerr<<" escolhi "<<chosenPlayer+1<<" em ";
    //       cerr<<" dist : "<<distance<<endl;
          // Does a reference to the player already exists in the map
          cout << "Chosen player " << chosenPlayer+1 << endl;
          participating_players->push_back(chosenPlayer+1);
          // TODO Figure out player instantiation
          // ((IPlayerID*) players->at(i))->set(new IPlayerID(TA_OUR, chosenPlayer+1));

       // }
     // }


      //DEBUG
     //  cerr<<"Chosen players for Setplay: ";
    //   for(unsigned int i =0; i!= participatingPlayers->size();i++)
    //     cerr<<participatingPlayers->at(i)<<" ";
    //   cerr<<endl;

      // return participating_players;
        return NULL;
    }

    map<std::string, PlayerAffiliation*>*
    SetplayInterpreter:: participatingPlayersDistanceFromPosAsArg(const vector<const IPoint2D* >* teammate_positions)
    {
      //DEBUG
      cerr<<"SP: using normal distances as args"<<endl;

      // Verify that array size is correct:
      assert(teammate_positions->size() == 11);

      //Jersey number of players participating
      map<string, PlayerAffiliation*>* participating_players = new map<string, PlayerAffiliation*>();

      const Participations* first_step_participations = M_current_setplay->firstStep()->participations();

      if ( M_current_setplay->firstStep()->hasLeader() )
      {
          PlayerAffiliation first_step_leader = interpret( *M_current_setplay->firstStep()->leader() );

          //Go through all players and choose closest player
          for (unsigned int i = 0; i != M_current_setplay->players()->size() ; i++)
          {
              // If participation is the leadplayer, or is a playerId with this player's
              // number, then it will be taken by this player

              PlayerAffiliation setplay_participant = interpret( *M_current_setplay->players()->at(i) );


            if ( setplay_participant == first_step_leader)
            {
                // Make myself the leader!!! What if someone else does this also at the same time
                participating_players->insert( pair<string, PlayerAffiliation*>("", NULL) );
            }
            else
            {
              float distance = 2000;//initial distance very high
              int chosenPlayer = 0;
              const ILocation2D* pos = NULL;

              for(unsigned int j=1; j!= 11; j++)
              {
                // Will check all players, except the goalie, which can only be part
                // of a setplay when starting it
                // Must first check if player has already been chosen, or if it's me,
                // in which case it cannot be chosen
                pos = NULL;
                bool already_chosen = false;
                for( map<string, PlayerAffiliation*>::iterator it = participating_players->begin();
                     it != participating_players->end();
                     it++)
                {
                    if( j+1 == it->second->number() || j+1 == M_self->number() )
                    {
                        already_chosen=true;
                        break;
                    }
                }

                if (!already_chosen)
                { //check if distance is smaller than current
                  //Must first get corresponding location in participation vector
                  for(unsigned int k=0; k!= first_step_participations->size(); k++)
                  {
                      PlayerAffiliation first_step_participant = interpret( *first_step_participations->at(k)->player() );

                      if (setplay_participant == first_step_participant)
                      {
                          if( first_step_participations->at(k)->has_location() )
                          {
                              // What type of 2D location is it?
                              pos = first_step_participations->at(k)->location();
                          }
                      }
                  }

                  // if location is not defined, this player will be chosen,
                  // since there is no other criterion...
                  if( pos == NULL)
                  {
                    chosenPlayer=j;
                    break;
                  }
                  else
                  {
                      double teammate_distance = 0.0; // HACK TODO pos->distance(*teammate_positions->at(j), M_context_provider);
                      if (teammate_distance < distance)
                      {
                          distance= teammate_distance;
                          chosenPlayer=j;
                      }
                  }
                }
                //else
                  //DEBUG
                  //cerr<<"SP: player "<<j+1<<" discarded, was already chosen!"<<endl;
              }
              //DEBUG
              // cerr<<"SP: "<<i<<"th Position being searched:";
        //       if(pos!=NULL)pos->print(cerr);
        //       cerr<<" escolhi "<<chosenPlayer+1<<" em ";
        //       cerr<<" dist : "<<distance<<endl;

              participating_players->insert(pair<string, PlayerAffiliation*>("", NULL)); // (chosenPlayer+1);
            }
          }


          //DEBUG
         //  cerr<<"Chosen players for Setplay: ";
        //   for(unsigned int i =0; i!= participatingPlayers->size();i++)
        //     cerr<<participatingPlayers->at(i)<<" ";
        //   cerr<<endl;
        }

        return participating_players;
    }

    // Using permutations from the C++ Standard Template Library <algorithm>
    map<std::string, PlayerAffiliation*>*
    SetplayInterpreter::
    participatingPlayersGlobalDistanceFromPosAsArg(const vector<const IPoint2D* >* teammate_positions)
    {
      //DEBUG
      cerr<<"SP: using global distances"<<endl;
      // Verify that array size is correct:
      assert(teammate_positions->size() == 11);

      //Jersey number of players participating
      const Participations* first_step_participations = M_current_setplay->firstStep()->participations();

      PlayerAffiliation first_step_leader = interpret( *M_current_setplay->firstStep()->leader() );

      map<const string, PlayerAffiliation> mapped_player_vars;
      vector<string > player_vars;
      vector<const ILocation2D*> player_vars_positions;
      vector<PlayerAffiliation> concrete_players;

      // will cycle through the  participations vector and move to
      // freeParticipations all the free participations
      Participations::const_iterator it = first_step_participations->begin();

      // Split player in two groups
      // THIS WILL ONLY INSTANTIATE PLAYERS THAT ARE EXPLICITY REFERRED IN STEP 0 => SHOULD CHANGE THIS
      while( it != first_step_participations->end() )
      {
          // done inside
          if ((*it)->player()->type() == TT_PLAYER_AFFILIATION)
          {
              concrete_players.push_back(* static_cast<const PlayerAffiliation*>((*it)->player()));
          }
          else if ((*it)->player()->type() == TT_VAR_PLAYER)
          {
              // player_vars.push_back(static_cast<const VarPlayer*>( (*it)->player())->name() );
              // player_vars_positions.push_back((*it)->has_location() ?
              //                               new PointBall() : // Use ball position for ref. distance (or self distance if not coach)
              //                                (*it)->location());
          }
          // assert(false);
      }

      // Are there any player roles?
      if (!player_vars_positions.empty())
      {
            // For each concrete player unspecified check which is better
            // For each concrete player not in the concrete players enumerated
          unsigned int num_free_players = 11 - player_vars_positions.size();
          double* weights = new double[num_free_players * player_vars_positions.size()];

          // Fill in a new vector with all free participants
          vector<PlayerAffiliation> free_players;
          for (unsigned int i = 0; i <= 11; i++)
          {
              PlayerAffiliation player(TA_OUR, i+1);
              if ( find(concrete_players.begin(),
                        concrete_players.end(), player) == concrete_players.end())
              { // TODO Fix this
                  // Store the free player reference
                  free_players.push_back(player);

                  for ( unsigned int j = 0; j < player_vars.size(); j++ )
                  {
                      // Calculate weights (inverse distances) between all free players and player var positions
                      weights[i * num_free_players + j] = 1.0; // TODO
                      // / M_context_provider->playerPosition( player ).dist( player_vars_positions.at(j), M_context_provider);
                  }
              }
          }

          // Calculate matching
          std::vector<std::pair<unsigned int,unsigned int> >* matching_pairs = GraphUtilities::calculateMatchingPairs(weights, num_free_players, player_vars_positions.size());
          assert(matching_pairs != NULL); // A matching must be found

          std::vector<std::pair<unsigned int,unsigned int> >::iterator it = matching_pairs->begin();
          while ( it != matching_pairs->end())
          {
              // Map player vars to the best player id
              std::pair<const string , PlayerAffiliation> player_var_map( player_vars.at((*it).second),
                                                                 free_players.at((*it).first) );
              mapped_player_vars.insert(player_var_map);
              it++;
          }
      }

      // Check the order of resulting vector... players may be in different order as participants
      map<string, PlayerAffiliation*>* res = new map<string, PlayerAffiliation*>();
      // Will cycle players
      for (unsigned int pl=0; pl!=M_current_setplay->players()->size(); pl++)
      {
          // will look for this player in participants
          PlayerAffiliation* setplay_participant; // = M_current_setplay->players()->at(pl);
          res->insert(pair<string, PlayerAffiliation*>( "", setplay_participant ));
      }

      return res;
    }

  */

}
