#include "setplay_communication.h"

#include <setplay/runtime/setplay_database.h>
#include <setplay/runtime/iworld_model.h>

#include <sstream>
#include <rcsc/common/logger.h>

namespace setplay {

/// MessageBox
MessageBox::MessageBox()
{

}

MessageBox::~MessageBox()
{
    if ( !M_messages.empty() )
    {
        for ( Messages::iterator it = M_messages.begin();
              it != M_messages.end();
              it++)
        {
            delete *it;
        }
    }
}

void MessageBox::pushMessage(IMessage* message)
{
    M_messages.push_back( message );
}

MessageBox::Messages* MessageBox::messages()
{
    return &M_messages;
}

unsigned int MessageBox::numMessages() const
{
    return M_messages.size();
}

void MessageBox::reset()
{
    for ( Messages::iterator it = M_messages.begin();
          it != M_messages.end() ;
          it++)
    {
        delete *it;
    }
    M_messages.clear();
}

void MessageBox::print(std::ostream& os) const
{
    if ( M_messages.empty() )
    {
        os << "No messages" << std::endl;
    }

    os << M_messages.size() << " messages " << std::endl;
    for ( Messages::const_iterator it = M_messages.begin();
          it != M_messages.end(); it++)
    {
        (*it)->print( os );
        os << std::endl;
    }
    os << std::endl;
}

/// SetplayCommunication

SetplayCommunication::SetplayCommunication()
    : M_last_execution_state( SetplayInterpreter::Stopped )
    , M_last_inboung_msg_time( 0u )
    , M_last_outbound_msg_time( 0u )
{

}

SetplayCommunication::~SetplayCommunication()
{

}

MessageBox* SetplayCommunication::outbox()
{
    return &M_outbox;
}

MessageBox* SetplayCommunication::inbox()
{
    return &M_inbox;
}

bool SetplayCommunication::repeatMessage() const
{
    return false; // (M_world_model->time() - M_step_start_time) < M_message_repeat_time && M_world_model->time()!=0;
}

void SetplayCommunication::processOutbox()
{
    if ( M_interpreter->executionState() == SetplayInterpreter::Stopped ||
         (M_interpreter->executionState() == M_last_execution_state &&
          ( M_last_execution_state == SetplayInterpreter::SetplayFinished ||
            M_last_execution_state == SetplayInterpreter::SetplayAborted)))
    {
        return;
    }

    // TODO Find out which messages should be sent based on the current interpreter state

    // Generate outbox messages
    const IWorldModel* world = M_interpreter->worldModel();

    // TODO Fix this issues with PlayerAffiliation
    PlayerAffiliation self = M_interpreter->self();
    unsigned int time = world->currentTime();

    /// Should send messages? Only if I am the current step-leader or the next step leader
    bool leadership = ( M_interpreter->hasNextStep() && M_interpreter->nextStepLeader() == self) ||
                      ( M_interpreter->hasCurrentStep() && M_interpreter->currentStepLeader() == self );
    bool should_send_messages = leadership;
    if ( !should_send_messages )
    {
        log("Should not send setplay messages ");
        return;
    }

    log("Going to generate setplay messages ");

    stringstream ss;
    IMessage* msg = NULL;
    switch ( M_interpreter->executionState() )
    {
        case SetplayInterpreter::Stopped:
            return;
        case SetplayInterpreter::SetplayStarted: // StepWait condition was not yet verified
        {
            SetplayStartMessage* setplay_start_msg = new SetplayStartMessage( self,
                                                                              time,
                                                                              M_interpreter->currentSetplay()->id(),
                                                                              M_interpreter->inverted() );
            if ( M_interpreter->currentSetplay()->hasParameters() )
            {
                for ( Parameters::const_iterator itp = M_interpreter->currentSetplay()->parameters()->begin();
                      itp != M_interpreter->currentSetplay()->parameters()->end(); itp++ )
                {
                    setplay_start_msg->push( itp->first, M_interpreter->resolve( itp->first )->clone() );
                }
            }

            msg = setplay_start_msg;
        }
        break;
        case SetplayInterpreter::StepChosen:
        case SetplayInterpreter::StepWait:
        case SetplayInterpreter::StepStarted: // StepStart condition was not yet verified
        {
            StepEnterMessage* step_enter_msg = new StepEnterMessage( self,
                                                                     time,
                                                                     M_interpreter->currentSetplay()->id(),
                                                                     M_interpreter->currentStep()->id(),
                                                                     M_interpreter->inverted() );
            // Add parameters
            if ( M_interpreter->currentSetplay()->hasParameters() )
            {
                for ( Parameters::const_iterator itp = M_interpreter->currentSetplay()->parameters()->begin();
                      itp != M_interpreter->currentSetplay()->parameters()->end(); itp++ )
                {
                    const Object* object = M_interpreter->resolve( itp->first );
                    assert( object != NULL );
                    step_enter_msg->push( itp->first, object->clone() );
                }
            }

            msg = step_enter_msg;
        }
        break;
        case SetplayInterpreter::StepAborted: // StepStart condition was not yet verified
            msg = new SetplayAbortMessage( self,
                                           time,
                                           M_interpreter->currentSetplay()->id(),
                                           M_interpreter->inverted());
        break;
        case SetplayInterpreter::TransitionChosen:
        case SetplayInterpreter::TransitionStarted:
        {
            // Check the type of transition
            StepTransitionMessage* step_transition_msg = new StepTransitionMessage( self,
                                                                                    time,
                                                                                    M_interpreter->currentSetplay()->id(),
                                                                                    M_interpreter->currentStep()->id(),
                                                                                    M_interpreter->nextStep()->id(),
                                                                                    M_interpreter->inverted() );
            // TODO Ignore parameters at this stage?? All should be known if any
            if ( M_interpreter->currentSetplay()->hasParameters() )
            {
                for ( Parameters::const_iterator ita = M_interpreter->currentSetplay()->parameters()->begin();
                      ita != M_interpreter->currentSetplay()->parameters()->end() ; ita++)
                {
                    const Object* obj = M_interpreter->resolve( ita->first );
                    step_transition_msg->push( ita->first, obj->clone() );
                }
            }
            msg = step_transition_msg;
        }
        break;
        case SetplayInterpreter::SetplayAborted:
        {
            msg = new SetplayAbortMessage( self,
                                          time,
                                          M_interpreter->currentSetplay()->id(),
                                          M_interpreter->inverted());
            // TODO Should change set-play interpreter state to stopped to prevent broadcast of messages continuosuly
        }
        break;
        case SetplayInterpreter::SetplayFinished:
        {
            msg = new SetplayFinishMessage( self,
                                            time,
                                            M_interpreter->currentSetplay()->id(),
                                            M_interpreter->inverted());
        }
        break;
        default:
        {
            std::cerr << "Unhandled interpreter execution state " << M_interpreter->executionStateStr(M_interpreter->executionState()) << std::endl;
            assert(false);
        }
            break;
   }

   if ( msg != NULL )
   {
       M_last_outbound_msg_time = time;
       msg->print(ss);
       M_outbox.pushMessage( msg );
   }

   log(ss.str());

   /// TODO Using a state variable might not be enough to work well when a set-play ends
   M_last_execution_state = M_interpreter->executionState();
}

bool SetplayCommunication::shouldProcessMessage(const IMessage &message) const
{
    if ( M_interpreter->executionState() != SetplayInterpreter::Stopped )
    {
        // Ignore messages if I am the leader
        if ( M_interpreter->hasNextStep() )
        {
            if ( M_interpreter->nextStepLeader() == M_interpreter->self() )
            {
                return false;
            }
        }
        else
        {
            if ( M_interpreter->hasCurrentStep() && M_interpreter->currentStepLeader() == M_interpreter->self() )
            {
                return false;
            }
        }
    }

    return true;
}

void SetplayCommunication::processInbox()
{
    // Clear outbox
    M_outbox.reset();

    if ( ( M_interpreter->executionState() == M_last_execution_state &&
           ( M_last_execution_state == SetplayInterpreter::SetplayFinished ||
             M_last_execution_state == SetplayInterpreter::SetplayAborted)))
    {
        log("ProcessInbox: Reseting interpreter (1st cond)");
        M_interpreter->reset();
        return;
    }

    // HACK Don't hear a message for 5 cycles => reset
    if ( !M_interpreter->canUpdateState() &&
         (M_interpreter->worldModel()->currentTime() - M_last_inboung_msg_time) > 5 )
    {
        log("ProcessInbox: Reseting interpreter (2nd cond)");
        M_interpreter->reset();
        return;
    }

    // If I am participant and the message is coming from a non-participant ignore it
    // Process all messages in the inbox
    MessageBox::Messages::iterator itm = M_inbox.messages()->begin();
    while ( itm != M_inbox.messages()->end() )
    {
        IMessage* msg = *itm;

        if ( shouldProcessMessage( *msg ) )
        {
            log("\nProcessed message in inbox");
            switch ( msg->type() )
            {
                case MSG_SETPLAY_START:
                    process( static_cast<SetplayStartMessage&>( *msg ));
                break;
                case MSG_STEP_ENTER:
                    process( static_cast<StepEnterMessage&>( *msg ));
                break;
                case MSG_STEP_TRANSITION:
                    process( static_cast<StepTransitionMessage&>( *msg ));
                break;
                case MSG_ACTION_DONE:
                    process( static_cast<ActionDoneMessage&>( *msg ));
                break;
                case MSG_ABORT:
                    process( static_cast<SetplayAbortMessage&>( *msg ));
                break;
                case MSG_FINISH:
                    process( static_cast<SetplayFinishMessage&>( *msg ));
                break;
            }
        }

        // TODO Should I broadcast received messages?
        // Only broadcast messages if they came from current leader
        // Non-participants will broadcast messages
        if ( M_interpreter->executionState() != SetplayInterpreter::Stopped && !M_interpreter->participatesInSetplay( M_interpreter->self() ) )
        {
            M_outbox.pushMessage( msg->clone() );
        }

        /// Discard received message
        delete msg;
        itm = M_inbox.messages()->erase( itm );
    }
    if ( !M_inbox.messages()->empty() )
    {
        std::cerr << M_inbox.messages()->size() << std::endl;
        std::cerr.flush();
        assert ( false );
    }
}

/// Setplay State Machine logic

bool SetplayCommunication::shouldChangeSetplay(const SetplayStartMessage& msg) const
{

    PlayerAffiliation self = M_interpreter->self();
    /// Accept a setplay change if the message came from the current step leader or next step leader
    /// Convert this to a policy? How??
    return ( M_interpreter->hasCurrentStep() && M_interpreter->currentStepLeader() == self ) ||
           ( M_interpreter->hasNextStep() && M_interpreter->nextStepLeader() ==  self );
}

bool SetplayCommunication::shouldRestartSetplay(const SetplayStartMessage& msg) const
{
    return false;
}

bool SetplayCommunication::shouldChangeStep(const StepTransitionMessage& msg) const
{
    PlayerAffiliation step_leader = M_interpreter->currentStepLeader();

    if ( step_leader == msg.sender() )
    {
        return true;
    }

    if ( M_interpreter->hasNextStep() )
    {
        PlayerAffiliation next_step_leader = M_interpreter->currentStepLeader();
        return msg.sender() == next_step_leader;
    }
    return false;
}


bool SetplayCommunication::shouldRestartStep(const StepTransitionMessage &step) const
{
    return false;
}


/// Communication

void SetplayCommunication::receive(IMessage* message)
{
    // Enqueue message
    M_inbox.pushMessage( message );
    M_last_inboung_msg_time = message->time();
}

void SetplayCommunication::process(const StepEnterMessage& msg)
{
    Setplay* setplay = M_database->getSetplay( msg.setplayId() );
    if ( setplay == NULL )
    {
        std::cerr << "Received a StepEnterMessage of a non-existent set-play" << std::endl;
        return;
    }

    Step* step = setplay->getStep( msg.stepId() );
    if ( step == NULL )
    {
        std::cerr << "Received a StepEnterMessage and non-existent step" << std::endl;
        return;
    }

    /* if ( M_interpreter == NULL )
    {
        // TODO
        log("StepEnterMessage: No set-play interpreter");
        assert(false);
        // TODO Setplay parameters should always be transmited??
        assert ( !setplay->hasParameters() );
        // Spawn a new interpreter for the set-play and the step
        M_interpreter = new SetplayInterpreter();
        M_interpreter->startSetplay( setplay );
        M_interpreter->setInverted( msg.inverted() );
        M_interpreter->enterStep( step );
        // M_setplay_mediator->pushInterpreter( M_interpreter );
    } */

    if ( M_interpreter->executionState() == SetplayInterpreter::Stopped )
    {
        M_interpreter->loadSetplay( setplay );

        /// Load all arguments of the set-play
        assert( setplay->parameters()->size() != msg.numArguments() );

        for ( Parameters::const_iterator ita = setplay->parameters()->begin();
              ita != setplay->parameters()->end(); ita++ )
        {
            const Object* obj = msg.argument( ita->first );
            M_interpreter->push( ita->first, obj->clone() );
        }

        M_interpreter->setInverted( msg.inverted() );
        M_interpreter->loadStep( step );
        return;
    }

    /// A set-play is executing. Does it have a different id?
    if ( M_interpreter->currentSetplay()->id() != msg.setplayId() )
    {
        log("Aborting the current set-play to start a new one");
        M_interpreter->abortSetplay();
        M_interpreter->loadSetplay( setplay );
        /// Load all arguments of the set-play
        assert( setplay->parameters()->size() != msg.numArguments() );
        for ( Parameters::const_iterator ita = setplay->parameters()->begin();
              ita != setplay->parameters()->end(); ita++ )
        {
            const Object* obj = msg.argument( ita->first );
            M_interpreter->push( ita->first, obj->clone() );
        }
        M_interpreter->setInverted( msg.inverted() );
        M_interpreter->loadStep( step );
        return;
    }

    /// TODO Does it have the same inversion? How about parameters? Are they instantiated the same?

    /// A set-play with the same id is executing. Is any current step being executed?
    if ( M_interpreter->executionState() == SetplayInterpreter::SetplayAborted ||
         M_interpreter->executionState() == SetplayInterpreter::SetplayFinished )
    {
        log("Received a StepEnterMessage and I've already aborted/Finished the set-play. Ignore it");
        return;
    }

    if ( !M_interpreter->hasCurrentStep() )
    {
        M_interpreter->loadStep( step );
        return;
    }

    /// Set-play with same id is executing a step. Does it have the same id?
    if ( M_interpreter->currentStep()->id() != msg.stepId() )
    {
        log("Ending current step and loading the new step");
        M_interpreter->endCurrentStep();
        M_interpreter->loadStep( step );
    }

}

void
SetplayCommunication::process(const SetplayStartMessage& msg)
{
    Setplay* setplay = M_database->getSetplay( msg.setplayId() );
    if ( setplay == NULL )
    {
        std::cerr << "Received a SetplayStartMessage of a non-existent set-play" << std::endl;
        return;
    }

    /// TODO Not handling the case when a set-play start message might be received of the same active set-play and the state is aborted or finished

    if ( M_interpreter == NULL )
    {
        // TODO Spawn a new setplay-interpreter?
        log("SetplayStartMessage: No set-play interpreter");
        assert( false );
    }

    if ( M_interpreter->executionState() != SetplayInterpreter::Stopped )
    {
        if ( M_interpreter->currentSetplay()->id() != msg.setplayId() &&
             shouldChangeSetplay(msg) )
        {
            /// Changing set-play
            log("Received a SetplayStartMessage but I am following a different set-play");
            M_interpreter->abortSetplay();
        }
        else if ( shouldRestartSetplay(msg) )
        {
            /// Restarting set-play
            log("Restarting an active set-play after having received a SetplayStartMessage");
            M_interpreter->abortSetplay();
        }
        else
        {
            log("Received a set-play start message but ignoring it");
            return;
        }
    }

    /// Starting a new setplay
    log("Starting a new set-play after having received a SetplayStartMessage");
    M_interpreter->loadSetplay( setplay );
    M_interpreter->setInverted( msg.inverted() );

    assert ( setplay->parameters()->size() != msg.numArguments() );

    for ( Parameters::const_iterator ita = setplay->parameters()->begin();
          ita != setplay->parameters()->end(); ita++ )
    {
        const Object* obj = msg.argument( ita->first );
        M_interpreter->push( ita->first, obj->clone() );
    }
}

void SetplayCommunication::process(const StepTransitionMessage& msg)
{
    if ( M_interpreter == NULL )
    {
        // TODO Spawn a new setplay-interpreter?
        log("StepTransitionMessage: No set-play interpreter");
        assert( false );
    }

    Setplay* setplay = M_database->getSetplay( msg.setplayId() );
    if ( setplay == NULL )
    {
        log("Received a StepTransitionMessage of a non-existent set-play");
        return;
    }

    Step* step = setplay->getStep( msg.currentStepId() );
    if ( step == NULL )
    {
        log("Received a StepTransitionMessage of a non-existent step");
        return;
    }

    Step* next_step = setplay->getStep( msg.nextStepId() );
    if ( next_step == NULL )
    {
        log("Received a StepTransitionMessage of a non-existent next step");
        return;
    }

    ITransition* transition = step ->getTransitionToStep( msg.nextStepId() );
    assert( transition != NULL );

    if ( M_interpreter->executionState() == SetplayInterpreter::Stopped )
    {
        /// Received a step message and no set-play was active

        /// No set-play was active => Start the new set-play
        log("StepTransitionMessage: No set-play was active => Start the new set-play");
        M_interpreter->loadSetplay( setplay  );
        /// Load all arguments of the set-play
        assert( setplay->parameters()->size() != msg.numArguments() );

        for ( Parameters::const_iterator ita = setplay->parameters()->begin();
              ita != setplay->parameters()->end(); ita++ )
        {
            const Object* obj = msg.argument( ita->first );
            M_interpreter->push( ita->first, obj->clone() );
        }

        M_interpreter->enterStep( step  );
        M_interpreter->followTransition( transition  );
        return;
    }

    /// Active set-play is different
    if ( M_interpreter->currentSetplay()->id() != msg.setplayId() )
    {
        log("StepTransitionMessage: Active set-play has a different id. Doing nothing.");
        return;
    }

    /// Active set-play is the same
    if ( M_interpreter->executionState() == SetplayInterpreter::SetplayAborted ||
         M_interpreter->executionState() == SetplayInterpreter::SetplayFinished )
    {
        log("Received a StepTransitionMessage and I've already aborted/Finished the set-play. Ignore it");
        return;
    }

    if ( !M_interpreter->hasCurrentStep() )
    {
        /// Setplay is active but no current step is not defined
        log("StepTransitionMessage: Setplay is active but no current step is not defined");
        M_interpreter->enterStep( step  );
        M_interpreter->followTransition( transition  );
        return;
    }

    /// Set-play is active and has a current step

    if ( M_interpreter->currentStep()->id() != msg.currentStepId() )
    {
        /// TODO CHECK IF THE NEXT STEP HAS LOWER IDS (AND IGNORE THEM)
        if ( !M_interpreter->hasNextStep() )
        {
            /// Change the current step of the set-play and follow the given transition
            log("StepTransitionMessage: No next step and msg current step id is different than the current step of the set-play. Change step and follow the given transition.");
            M_interpreter->endCurrentStep();
            M_interpreter->enterStep( step  );
            M_interpreter->followTransition( transition  );
            return;
        }

        /// Has next step
        /// Check if next step in the message is already the current step. If so we would be going backwards
        if ( M_interpreter->currentStep()->id() == msg.nextStepId() )
        {
            /// HACK Accept unconditionally if the message came from the next step leader
            if ( M_interpreter->nextStepLeader() == msg.sender() )
            {
                M_interpreter->endTransition();
                M_interpreter->loadStep( step );
                M_interpreter->followTransition( transition );
                log("StepTransitionMessage: StepTransition message came from next step leader. Accepting the change in transition.");
                return;
            }
            log("StepTransitionMessage: Ignore the message since the next step id of the message is already the current step.");
            return;
        }

        /// TODO Msg.currentStepId != currentStepId && Msg.currentStepId != nextStepId
        log("StepTransitionMessage: Msg.currentStepId != currentStepId && Msg.currentStepId != nextStepId. Follow the msg transition");
        M_interpreter->endCurrentStep();
        M_interpreter->enterStep( step );
        M_interpreter->followTransition( transition );
        return;
    }

    /// Set-play is active and has the same current step
    if ( !M_interpreter->hasNextStep() )
    {
        /// Set-play is active, has the same current step but has no transition
        log("StepTransitionMessage: Set-play is active, has the same current step but has no transition");
        M_interpreter->followTransition( transition  );
        return;
    }

    /// Set-play is active, has the same current step and has a current transition
    if ( M_interpreter->nextStep()->id() != msg.nextStepId() )
    {
        /// Set-play is active, has the same current step but has a different transition
        log("StepTransitionMessage: Set-play is active, has the same current step but has a different transition");
        M_interpreter->endTransition();
        M_interpreter->followTransition( transition  );
        return;
    }

    /// Set-play is active, has the same current step and has the same current transition
    // TODO Restart transition or do nothing?
}


void
SetplayCommunication::process(const SetplayAbortMessage& msg)
{
   /* if ( M_interpreter == NULL )
    {
        // TODO Spawn a new setplay-interpreter?
        log("SetplayAbortMessage: No set-play interpreter");
        assert( false );
    } */
    if ( M_interpreter->executionState() == SetplayInterpreter::Stopped )
    {
        log("Received a setplay abort message but no setplay is active");
        return;
    }

    if ( M_interpreter->currentSetplay()->id() != msg.setplayId() )
    {
        log("Received a setplay abort message but the id of the active setplay is does not match");
        return;
    }

    if ( M_interpreter->executionState() == SetplayInterpreter::SetplayAborted )
    {
        log("Received a setplay abort message and I've already aborted the set-play. Reset the interpreter");
        M_interpreter->reset();
        return;
    }

    log("Received a setplay abort message and I am executing the set-play. Aborting the set-play");
    M_interpreter->abortSetplay();
}

void
SetplayCommunication::process(const SetplayFinishMessage& msg)
{
   /* if ( M_interpreter == NULL )
    {
        // TODO Spawn a new setplay-interpreter?
        log("SetplayAbortMessage: No set-play interpreter");
        assert( false );
    } */

    if ( M_interpreter->executionState() == SetplayInterpreter::Stopped )
    {
        log("Received a setplay finish message but no setplay is active");
        return;
    }

    if ( M_interpreter->currentSetplay()->id() != msg.setplayId() )
    {
        log("Received a setplay finish message but the id of the active setplay does not match");
        return;
    }

    if ( M_interpreter->executionState() == SetplayInterpreter::SetplayFinished )
    {
        log("Received a setplay finish message and I've already finished the set-play. Reset the interpreter");
        M_interpreter->reset();
        return;
    }

    log("Received a setplay finish message and I am executing the set-play. Finishing the set-play");
    M_interpreter->finishSetplay();
}

void
SetplayCommunication::process(const ActionDoneMessage&)
{
    log("Received an action done message which is not yet implemented");
    assert(false);
    // TODO
}


void SetplayCommunication::reset()
{
    M_inbox.reset();
    M_outbox.reset();
}

void SetplayCommunication::log(const std::string& message) const
{
    rcsc::dlog.addText( rcsc::Logger::COMMUNICATION,
                        __FILE__ " SetplayCommunication: %s", message.c_str());
}


/// TimedMessage
SetplayCommunication::TimedMessage::TimedMessage( IMessage* message, unsigned int time_to_live )
    : M_message( message )
    , M_time_to_live( time_to_live )
{
    assert( message != NULL );
}

const IMessage* SetplayCommunication::TimedMessage::getMessage()
{
    assert( M_time_to_live >= 0 );
    M_time_to_live--;
    return M_message;
}

unsigned int SetplayCommunication::TimedMessage::timeToLive() const
{
    return M_time_to_live;
}

}
