#include "setplay_interpreter.h"

#include <cassert>
#include <fstream>
#include <utility>
#include <algorithm>
#include <typeinfo>

#include <setplay/model/setplay.h>

#include <setplay/model/geom/ipoint.h>
#include <setplay/model/geom/iregion.h>
#include <setplay/model/conditions.h>
#include <setplay/model/actions.h>
#include <setplay/model/numbers.h>
#include <setplay/model/participants.h>
#include <setplay/model/transitions.h>
#include <setplay/model/directives.h>
#include <setplay/model/assign.h>

#include <setplay/runtime/icondition_evaluator.h>
#include <setplay/runtime/iaction_executor.h>
#include <setplay/runtime/transition_selector.h>
#include <setplay/runtime/abort_decider.h>
#include <setplay/runtime/state_update_decider.h>

#include <setplay/runtime/isetplay_execution_observer.h>
#include <setplay/model/variable.h>

#include <setplay/model/types.h>
#include <setplay/model/participants/player_affiliation.h>

#include <sstream>
#include <rcsc/common/logger.h>
#include <setplay/generators/sexpr/setplay_generator_sexpr.h>

namespace setplay {

    SetplayInterpreter::SetplayInterpreter()
        : M_execution_state( Stopped )
        , M_current_setplay(NULL)
        , M_last_step(NULL)
        , M_current_step(NULL)
        , M_next_step(NULL)
        , M_current_transition(NULL)
        , M_inverted(false)
        , M_action_executor( NULL )
        , M_condition_evaluator( NULL )
        , M_transition_selector( NULL )
        , M_abort_decider( NULL )
        , M_state_update_decider( NULL )
    {

    }

    /// Accessors
    SetplayInterpreter::ExecutionState SetplayInterpreter::executionState() const
    {
        return M_execution_state;
    }

    const Setplay* SetplayInterpreter::currentSetplay() const
    {
        return M_current_setplay;
    }

    const Step* SetplayInterpreter::currentStep() const
    {
        return M_current_step;
    }

    const Step* SetplayInterpreter::nextStep() const
    {
        return M_next_step;
    }

    bool SetplayInterpreter::hasActiveSetplay() const
    {
        return M_execution_state != Stopped &&
               M_execution_state != SetplayFinished &&
               M_execution_state != SetplayAborted;
    }

    bool SetplayInterpreter::hasCurrentSetplay() const
    {
        return M_current_setplay != NULL;
    }

    bool SetplayInterpreter::hasCurrentStep() const
    {
        return M_current_step != NULL;
    }

    bool SetplayInterpreter::hasCurrentTransition() const
    {
        return M_current_transition != NULL;
    }

    bool SetplayInterpreter::hasNextStep() const
    {
        return M_next_step != NULL;
    }

    PlayerAffiliation SetplayInterpreter::currentStepLeader() const
    {
        assert( hasCurrentStep() );

        return interpret( *M_current_step->leader() );
    }

    PlayerAffiliation SetplayInterpreter::nextStepLeader() const
    {
        assert( hasNextStep() );

        return interpret( *M_next_step->leader() );
    }

    const ITransition* SetplayInterpreter::currentTransition() const
    {
        return M_current_transition;
    }

    bool SetplayInterpreter::participatesInSetplay(const PlayerAffiliation& player)
    {
        if ( M_execution_state == Stopped  )
        {
            return false;
        }

        for ( Participants::const_iterator itp = M_current_setplay->players()->begin();
              itp != M_current_setplay->players()->end();
              itp++)
        {
            PlayerAffiliation setplay_player = interpret( **itp );
            if ( setplay_player == player )
            {
                return true;
            }
        }
        return false;
    }

    bool SetplayInterpreter::selfParticipatesInSetplay()
    {
        return participatesInSetplay( self() );
    }

    bool SetplayInterpreter::participatesInCurrentStep(const PlayerAffiliation& player)
    {
        assert ( hasCurrentStep() );

        return participatesInStep( *M_current_step, player );
    }

    PlayerAffiliation SetplayInterpreter::self() const
    {
        return PlayerAffiliation( M_world_model->self(), TA_OUR);
    }


    bool SetplayInterpreter::selfParticipatesInCurrentStep()
    {
        return participatesInCurrentStep( self() );
    }

    bool SetplayInterpreter::participatesInNextStep(const PlayerAffiliation& player)
    {
        assert ( hasNextStep() );

        return participatesInStep( *M_next_step, player );
    }

    bool SetplayInterpreter::selfParticipatesInNextStep()
    {
        return participatesInNextStep( self() );
    }

    /// Helper methods
    bool SetplayInterpreter::participatesInStep(const Step& step, const PlayerAffiliation& player)
    {
        for ( Participations::const_iterator itp = step.participations()->begin();
              itp != step.participations()->end();
              itp++)
        {
            PlayerAffiliation step_player = interpret( *(*itp)->player() );
            if ( step_player == player )
            {
                return true;
            }
        }
        return false;
    }

    bool SetplayInterpreter::canUpdateState() const
    {
        bool can_update_state = M_state_update_decider == NULL ||
                                M_state_update_decider->canUpdateInterpreterState( *this );
        rcsc::dlog.addText( rcsc::Logger::PLAN,
                            __FILE__ " SetplayInterpreter: canUpdateState=%s CurrentState=%s",
                            can_update_state ? "true" : "false",
                            executionStateStr(M_execution_state).c_str() );
        return can_update_state;
    }

    /// Modifiers


    void SetplayInterpreter::update()
    {
        rcsc::dlog.addText( rcsc::Logger::PLAN,
                            __FILE__ " Updating set-play interpreter with State=%s",
                            executionStateStr( M_execution_state ).c_str());

        // Any abort situation overrides the progress of the set-play (non optimistic)
        if ( canUpdateState() && shouldAbort() )
        {
            abortSetplay();
            return;
        }

        switch ( M_execution_state )
        {
            case Stopped:
                return;
            break;
            case SetplayStarted: // StepWait condition was not yet verified
            {
                std::cerr << "TODO Implement" << std::endl;
                assert(false);
            }
            break;
            case StepChosen:
            {
                /// START Check if participations have locations and generate an implicit condition and actions
                /// TODO Decide who should evaluate wait condition!!!

                if ( M_current_step->hasWaitCondition() )
                {
                    std::vector<IPlayer*> players_of_interest;

                    ILocation2D* self_location = NULL;
                    PlayerAffiliation self = this->self();
                    PlayerAffiliation step_leader = currentStepLeader();
                    for ( Participations::const_iterator itp = M_current_step->participations()->begin() ;
                          itp != M_current_step->participations()->end() ; itp++)
                    {
                        IParticipation* step_participation = *itp;

                        stringstream ss;
                        SetplayGeneratorSexpr sexpr(ss);
                        step_participation->visit( sexpr );
                        rcsc::dlog.addText( rcsc::Logger::PLAN,
                                            __FILE__ " SeptlayInterpreter: ImplicitActStepWait StepParticipation=%s",
                                            ss.str().c_str());
                        PlayerAffiliation step_participant = interpret( *step_participation->player() );
                        if ( step_participation->hasLocation() )
                        {
                            CondAtLocation* cond_at_location = new CondAtLocation( step_participation->player()->clone(),
                                                                                   step_participation->location()->clone());

                            stringstream ss1;
                            SetplayGeneratorSexpr sexpr( ss1 );
                            // HACK Forcing interpretation of location as ipoint2D
                            if ( step_participant != self )
                            {
                                IPlayer* poi = step_participation->player()->clone();
                                players_of_interest.push_back( poi );
                            }

                            cond_at_location->visit( sexpr );
                            step_participation->location()->visit( sexpr );
                            rcsc::dlog.addText( rcsc::Logger::PLAN,
                                                __FILE__ " SeptlayInterpreter: CondAtLocation =%s",
                                                ss1.str().c_str());

                            PlayerAffiliation step_participant = interpret( *step_participation->player() );
                            if ( step_participant == self )
                            {
                                if ( self_location != NULL )
                                {
                                    std::cerr << "Player appears >1x in set-play participation" << std::endl;
                                    assert(false);
                                }
                                self_location = step_participation->location();
                            }
                        }
                    }

                    bool am_step_leader = self == step_leader;

                    stringstream ss;
                    SetplayGeneratorSexpr gen(ss);
                    if ( self_location != NULL )
                    {
                        if ( am_step_leader )
                        {
                            M_current_actions.push_back( new ActDribble( self_location->clone() ) );
                            M_current_actions.push_back( new ActHold() );

                            if ( !players_of_interest.empty() )
                            {
                                M_current_actions.push_back( new ActLookAt(players_of_interest) );
                            }
                        }
                        else
                        {
                            M_current_actions.push_back( new ActPos( self_location->clone() ) );
                            /// Look at leader
                            M_current_actions.push_back( new ActLookAt( M_current_step->leader()->clone() ) );
                        }
                    }
                    else
                    {
                        if ( am_step_leader )
                        {
                            M_current_actions.push_back( new ActHold() );
                        }

                        if ( !players_of_interest.empty() )
                        {
                            M_current_actions.push_back( new ActLookAt( players_of_interest ) );
                        }
                    }

                    ss << "NumActions=" << M_current_actions.size() << " ";
                    for ( Actions::const_iterator ita = M_current_actions.begin(); ita != M_current_actions.end() ; ita++ )
                    {
                        (*ita)->visit( gen );
                    }
                    rcsc::dlog.addText( rcsc::Logger::PLAN,
                                        __FILE__ " SetplayInterpreter: Implicit actions? => %s",
                                        ss.str().c_str());

                    M_execution_state = StepWait;
                }
                else
                {
                    M_execution_state = StepStarted;
                }

                notifyObservers();
                update();
                return;
            }
            break;
            case StepWait:
            {
                if ( canUpdateState() )
                {
                    /*bool should_start_step = !M_current_step->hasStartCondition() || interpret( *M_current_step->startCondition() );
                    // Should start step?
                    rcsc::dlog.addText( rcsc::Logger::PLAN,
                                        __FILE__ " ShouldStartStep=%s", should_start_step ? "true" : "false" );
                    if ( should_start_step )
                    {
                        M_execution_state = StepStarted;
                        notifyObservers();
                        update();
                    } */
                    bool should_wait = !M_current_step->hasWaitCondition() || interpret( *M_current_step->waitCondition() );
                    // Should start step?
                    rcsc::dlog.addText( rcsc::Logger::PLAN,
                                        __FILE__ " StepWait=%s", should_wait  ? "true" : "false" );
                    if ( !should_wait  )
                    {
                        M_execution_state = StepStarted;
                        notifyObservers();
                        update();
                        return;
                    }
                }
            }
            break;
            case StepStarted: // StepStart condition was verified
            {
                if ( canUpdateState() )
                {
                    // Choose transition
                    if ( M_current_step->numTransitions() == 0 )
                    {
                        finishSetplay();
                        return;
                    }

                    if ( M_current_step->numTransitions() == 1  )
                    {
                        ITransition* transition = M_current_step->transitions()->at(0);
                        followTransition( transition );

                        M_execution_state = TransitionChosen;
                        notifyObservers();
                    }
                    else
                    {
                        const ITransition* transition = M_transition_selector->selectTransition( *this );
                        assert ( transition != NULL );

                        if ( transition != NULL )
                        {
                            followTransition( transition );
                            M_execution_state = TransitionChosen;
                            notifyObservers();
                        }
                    }

                    if ( M_execution_state != StepStarted )
                    {
                        update();
                        return;
                    }
                }
            }
            break;
            case TransitionChosen:
            {
                bool should_start_transition = !M_current_transition->hasStartCondition() ||
                                               interpret( *M_current_transition->startCondition());
                rcsc::dlog.addText( rcsc::Logger::PLAN,
                                    __FILE__ " ShouldStartTransition=%s", should_start_transition ? "true" : "false");

                // Should start transition?
                if ( should_start_transition )
                {
                    // Load actions
                    loadActions( *M_current_transition );
                    M_execution_state = TransitionStarted;
                    notifyObservers();
                    update();
                    return;
                }
            }
            break;
            case TransitionStarted:
            {
                // Action execution code was here

                if ( canUpdateState() )
                {

                    // IMPORTANT NOTE
                    // Due to the concept of transition in the grammar there might not exist a next-step
                    // It makes more sense to me that this could become the property of a step

                    switch ( M_current_transition->transitionType()  )
                    {
                        case ITransition::TT_ABORT:
                        {
                            M_execution_state = SetplayAborted;
                            notifyObservers();
                        }
                        break;
                        case ITransition::TT_FINISH:
                        {
                            M_execution_state = SetplayFinished;
                            notifyObservers();
                        }
                        break;
                        case ITransition::TT_NEXT_STEP:
                        {
                            assert ( hasNextStep() );
                            bool should_enter_next_step = M_next_step != NULL &&
                                                          ( !M_next_step->hasStartCondition() ||
                                                            interpret( *M_next_step->startCondition()) );

                            if ( should_enter_next_step )
                            {
                                loadStep( M_next_step );
                                update();
                                return;
                            }
                        }
                        break;
                    }
                }
            }
            break;
            case StepAborted:
            case SetplayAborted:
            case SetplayFinished:
            {
                // Don't do anything
                clearActions();
            }
                break;
        }


        /// Execute all actions
        rcsc::dlog.addText( rcsc::Logger::PLAN,
                            __FILE__ " Actions to execute =%d ",
                            M_current_actions.size());

        /// Action execution
        if ( M_current_actions.empty() )
        {
            M_action_executor->noAction();
        }
        else
        {
            while ( !M_current_actions.empty() )
            {
                // Hack to remove constness to allow visit
                IAction* action = const_cast<IAction*>( M_current_actions.front() );
                stringstream ss;
                SetplayGeneratorSexpr sexpr(ss);
                action->visit( sexpr );
                rcsc::dlog.addText( rcsc::Logger::PLAN,
                                    __FILE__ " SetplayInterpreter: Going to execute %s",
                                    ss.str().c_str());
                if ( !interpret( *action ) )
                {
                    rcsc::dlog.addText( rcsc::Logger::PLAN,
                                        __FILE__ " SetplayInterpreter: Action execution not successful");
                    break;
                }
                rcsc::dlog.addText( rcsc::Logger::PLAN,
                                    __FILE__ " SetplayInterpreter: Action execution successful");
                delete action; // Free action space
                M_current_actions.erase(M_current_actions.begin());
            }
        }
    }


    void SetplayInterpreter::loadSetplay(const Setplay* setplay)
    {
        assert( !hasActiveSetplay() );

        if ( hasActiveSetplay() )
        {
            reset();
        }

        // Check that the set-play parameters and the arguments size match => TODO Should also check if their types match!!
        M_current_setplay = setplay;

        // Push the setplay space to the stack
        MemorySpace* setplay_space = new MemorySpace("setplay");
        M_objects_spaces.push_back(setplay_space);

        M_execution_state = SetplayStarted;
        notifyObservers(); /// TODO Set-play should be started explicitly externally

        // Enter first step??
        loadStep( setplay->firstStep() );
    }

    void SetplayInterpreter::loadStep(const Step* new_step)
    {
        if ( M_current_step == new_step )
        {
            return;
        }

        /// Save last step
        M_last_step = M_current_step;

        /// StepChanged
        if ( M_current_step != NULL )
        {
            // Uninstantiate variables
            if ( M_current_transition != NULL && M_current_transition->hasAssignments() )
            {
                M_objects_spaces.pop_back();
            }
            if ( M_current_step->hasAssignments() )
            {
                M_objects_spaces.pop_back();
            }
        }

        M_current_step = new_step;
        M_current_transition = NULL;
        M_next_step = NULL;
        clearActions();

        if ( M_current_step == NULL ) { return; }

        if ( M_current_step->hasAssignments() )
        {
            MemorySpace* step_space = new MemorySpace("");
            M_objects_spaces.push_back( step_space );

            /// Execute step assignments
            for ( Assignments::const_iterator ita = M_current_step->assignments()->begin();
                  ita != M_current_step->assignments()->end();
                  ita++)
            {
                interpret( **ita );
            }
        }

        M_execution_state = StepChosen;
        notifyObservers();
    }

    /// Assumes start condition was already verified
    void SetplayInterpreter::enterStep(const Step* new_step)
    {
        loadStep( new_step );
        M_execution_state = StepStarted;
        notifyObservers();
    }

    void SetplayInterpreter::endCurrentStep()
    {
        assert ( M_current_step != NULL );

        if ( M_current_transition != NULL  )
        {
            endTransition();
        }

        if ( M_current_step->hasAssignments() )
        {
            MemorySpace* step_space = M_objects_spaces.back();
            M_objects_spaces.pop_back();
            delete step_space;
        }

        M_current_step = NULL;
    }

    void SetplayInterpreter::followTransition(const ITransition* new_transition)
    {
        if ( M_current_transition == new_transition )
        {
            return;
        }

        /// TransitionChanged
        if ( M_current_transition != NULL )
        {
            // Uninstantiate transition spaces
            if ( M_current_transition->hasAssignments() )
            {
                M_objects_spaces.pop_back();
            }
        }

        // transitionChanged( old_transition, new_transition );
        M_current_transition = new_transition;
        M_next_step = NULL;
        clearActions();

        if ( M_current_transition == NULL ) { return; }

        if ( M_current_transition->type() == TT_TRANSITION_NEXT_STEP )
        {
            M_next_step = M_current_setplay->getStep(static_cast<const NextStep*>(M_current_transition)->nextStepId());
        }

        if ( M_current_transition->hasAssignments() )
        {
            MemorySpace* transition_space = new MemorySpace("");
            M_objects_spaces.push_back( transition_space );

            /// Execute step assignments
            for ( Assignments::const_iterator ita = M_current_transition->assignments()->begin();
                  ita != M_current_transition->assignments()->end();
                  ita++)
            {
                interpret( **ita );
            }
        }

        // Load actions for player!
        loadActions( *M_current_transition );

        M_execution_state = TransitionStarted;
        notifyObservers();
    }

    void SetplayInterpreter::endTransition()
    {
        assert ( M_current_transition != NULL );

        if ( M_current_transition->hasAssignments() )
        {
            MemorySpace* transition_space = M_objects_spaces.back();
            M_objects_spaces.pop_back();
            delete transition_space;
        }

        M_current_transition = NULL;
        M_next_step = NULL;
    }

    void SetplayInterpreter::abortSetplay()
    {
        assert( M_current_setplay != NULL );

        M_execution_state = SetplayAborted;
        notifyObservers();
    }

    void SetplayInterpreter::finishSetplay()
    {
        assert(M_current_setplay != NULL);

        M_execution_state = SetplayFinished;

        notifyObservers();
    }

    void SetplayInterpreter::clearActions()
    {
        for ( Actions::const_iterator ita = M_current_actions.begin() ; ita != M_current_actions.end() ; ita++ )
        {
            delete *ita;
        }
        M_current_actions.clear();
    }

    void SetplayInterpreter::reset()
    {
        if ( hasNextStep() )
        {
            endTransition();
        }

        if ( hasCurrentStep() )
        {
            endCurrentStep();
        }

        clearActions();
        resetSpace();

        M_execution_state = SetplayInterpreter::Stopped;
        M_current_setplay = NULL;
        M_current_step = NULL;
        M_next_step = NULL;
        M_current_transition = NULL;
        M_inverted = false;
    }

    void SetplayInterpreter::resetSpace()
    {
        for ( std::vector<MemorySpace*>::iterator its = M_objects_spaces.begin();
              its != M_objects_spaces.end();
              its++)
        {
            delete *its;
        }
        M_objects_spaces.clear();
    }

    void SetplayInterpreter::setInverted(bool inverted)
    {
        M_inverted = inverted;
    }

    bool SetplayInterpreter::inverted() const
    {
        return M_inverted;
    }

    bool SetplayInterpreter::shouldAbort()
    {
        assert ( M_current_setplay != NULL );

        if ( M_abort_decider != NULL && !M_abort_decider->shouldAbort( *this ) )
        {
            rcsc::dlog.addText( rcsc::Logger::PLAN,
                                __FILE__ " SetplayInterpreter: Cannot decide abort");
            return false;
        }

        if ( shouldAbortSetplay() )
        {
            rcsc::dlog.addText( rcsc::Logger::PLAN,
                                __FILE__ " SetplayInterpreter: Aborting due to set-play abort condition");
            return true;
        }

        if ( shouldAbortStep() )
        {
            rcsc::dlog.addText( rcsc::Logger::PLAN,
                                __FILE__ " SetplayInterpreter: Aborting due to current step abort condition");
            return true;
        }

        return false;
    }

    bool SetplayInterpreter::shouldAbortSetplay()
    {
        assert( M_current_setplay != NULL );

        return M_current_setplay->hasAbortCondition() &&
               interpret( *M_current_setplay->abortCondition() );
    }

    bool SetplayInterpreter::shouldAbortStep()
    {
        return hasCurrentStep() &&
               M_current_step->hasAbortCondition() &&
               interpret( *M_current_step->abortCondition() );
    }


    void SetplayInterpreter::loadActions(const ITransition& transition )
    {
        // Clear previous actions?
        clearActions();

        /// Load player actions, if any
        if ( !transition.hasDirectives() )
        {
            return;
        }

        for( Directives::const_iterator itd = transition.directives()->begin();
             itd != transition.directives()->end();
             itd++)
        {
            switch ( (*itd)->directiveType() )
            {
                case IDirective::DT_DO:
                    interpret( static_cast<const Do&>( **itd ) );
                break;
                case IDirective::DT_DONT:
                    interpret( static_cast<const Dont&>( **itd ) );
                break;
                case IDirective::DT_DO_WHILE:
                    interpret( static_cast<const DoWhile&>( **itd ) );
                break;
            }
        }
    }


    // FA 28/05/11 - Retrieve a non-interdiction region to be used by players not participating in the setplay
    /*IRegion2D* SetplayInterpreter::getExecutionRegion()
    {
        assert(M_active_setplay != NULL);

        const vector<ParticipationSituated*>* participations = M_active_setplay->firstStep().participations();
        if (participations == NULL || participations->size() == 0) {
            cerr << "SP (interdictedRegion): No participants are available in the current step #" << M_active_setplay->id() << " (CRITICAL ERROR!!!)" << endl;
            return NULL;
        }

        // Step class: if no transition is active, return a region based on the current step participants initial positions.

        IRegion* interdictedRegion = NULL;
        if (participations->size() == 3) {
            const IPoint2D* pointPlayer1 = //M_context_provider->playerPos(*participants->at(0)->player()->value());
            participations->at(0)->location() == NULL
                                            ? M_context_provider->playerPos(participations->at(0)->player()->asPlayerID())
                                            : participations->at(0)->location()->centroid(M_context_provider);
            const IPoint2D* pointPlayer2 = // M_context_provider->playerPos(*participants->at(1)->player()->value());
            participations->at(1)->location() == NULL
                                        ? M_context_provider->playerPos(participations->at(1)->player()->asPlayerID())
                                        : participations->at(1)->location()->centroid(M_context_provider);
            const IPoint2D* pointPlayer3 = // M_context_provider->playerPos(*participants->at(2)->player()->value());
            participations->at(2)->location() == NULL
                                        ? M_context_provider->playerPos(participations->at(2)->player()->asPlayerID())
                                        : participations->at(2)->location()->centroid(M_context_provider);

            // Create a triangular region triangle between the 3 players
            interdictedRegion = new RegTri(const_cast < IPoint2D* > (pointPlayer1),
                                           const_cast < IPoint2D* > (pointPlayer2),
                                           const_cast < IPoint2D* > (pointPlayer3));
        } else {
            if (participations->size() == 2) {
                const IPoint2D* pointPlayer1 = //M_context_provider->playerPos(*participants->at(0)->player()->value());
                participations->at(0)->location() == NULL
                                            ? M_context_provider->playerPos(participations->at(0)->player()->asPlayerID())
                                            : participations->at(0)->location()->centroid(M_context_provider);
                const IPoint2D* pointPlayer2 = M_context_provider->playerPos(participations->at(1)->player()->asPlayerID());
                participations->at(1)->location() == NULL
                                            ? M_context_provider->playerPos(participations->at(1)->player()->asPlayerID())
                                            : participations->at(1)->location()->centroid(M_context_provider);
                // Create a rectangular region between the 2 players
                interdictedRegion = new RegRec(const_cast < IPoint2D* > (pointPlayer1), const_cast < IPoint2D* > (pointPlayer2));
            } else {
                // More than 3 players => Calculate highestX, lowestX, highestY, lowestY
                const IPoint2D* currentPlayerPos = //M_context_provider->playerPos(*participants->at(0)->player()->value());
                participations->at(0)->location() == NULL
                                            ? M_context_provider->playerPos(participations->at(0)->player()->asPlayerID())
                                                    : participations->at(0)->location()->centroid(M_context_provider);
                float lowestY = currentPlayerPos->getY()->value();
                float lowestX = currentPlayerPos->getX()->value();
                float highestY = lowestY;
                float highestX = lowestX;

                for (unsigned int i = 1; i < participations->size(); i++) {
                    currentPlayerPos = M_context_provider->playerPos(participations->at(i)->player()->asPlayerID());

                    if (lowestY > currentPlayerPos->getY()->value()) {
                        lowestY = currentPlayerPos->getY()->value();
                    }
                    if (highestY < currentPlayerPos->getY()->value()) {
                        highestY = currentPlayerPos->getY()->value();
                    }
                    if (lowestX > currentPlayerPos->getX()->value()) {
                        lowestX = currentPlayerPos->getX()->value();
                    }
                    if (highestX < currentPlayerPos->getX()->value()) {
                        highestX = currentPlayerPos->getX()->value();
                    }
                }

                IPoint2D* topLeftInterdictedRegionPt = new IPoint2D(new Decimal(lowestX),
                                                                          new Decimal(lowestY));
                IPoint2D* bottomRightInterdictedRegionPt = new IPoint2D(new Decimal(highestX),
                                                                              new Decimal(highestY));

                // Create a rectangular region between the N players
                interdictedRegion = new RegRec(topLeftInterdictedRegionPt, bottomRightInterdictedRegionPt);

            }
        }
        return interdictedRegion;
        return NULL;
    } */

    SetplayInterpreter::~SetplayInterpreter()
    {

    }

    MemorySpace* SetplayInterpreter::currentSpace( )
    {
        assert ( !M_objects_spaces.empty() );

        return M_objects_spaces.back();
    }

    MemorySpace* SetplayInterpreter::getSpaceWithSymbol( const std::string& name )
    {
        // Iterate existing memory spaces in inverse order
        for ( vector<MemorySpace*>::const_reverse_iterator itm = M_objects_spaces.rbegin();
              itm != M_objects_spaces.rend(); itm++)
        {
            if ( (*itm)->get(name) != NULL )
            {
                return *itm;
            }
        }

        // Not found in stack => Look in global space
        /*if ( M_global_space->get( name ) != NULL )
        {
            return M_global_space;
        }*/

        // No space found with that symbol
        return NULL;
    }

    const Object* SetplayInterpreter::resolve( const std::string& name ) const
    {
        const Object* object = NULL;

        // Iterate existing memory spaces in inverse order
        for ( vector<MemorySpace*>::const_reverse_iterator itm = M_objects_spaces.rbegin();
              itm != M_objects_spaces.rend(); itm++)
        {
            object = ( (*itm)->get(name) );
            if ( object != NULL )
            {
                return object;
            }
        }

        if ( M_global_space != NULL )
        {
            object = M_global_space->get(name);
            if ( object != NULL )
            {
                return object;
            }
        }

        // Die here if no object found
        stringstream ss;
        ss << "No object found with name " << name << std::endl;
        ss << "Memory spaces: " << M_objects_spaces.size() << std::endl;
        for ( std::vector<MemorySpace*>::const_reverse_iterator it = M_objects_spaces.rbegin();
              it != M_objects_spaces.rend(); it++ )
        {
            (*it)->print( ss );
            ss << std::endl;
        }
        M_global_space->print(ss);
        ss << std::endl;

        std::cerr << ss.str() << std::endl;
        assert( false );

        return object;
    }

    bool SetplayInterpreter::unset( const std::string& name )
    {
        MemorySpace* space = getSpaceWithSymbol( name );
        if ( space != NULL )
        {
            return space->free( name );
        }

        return false;
    }

    void SetplayInterpreter::push( std::string name, Object* object)
    {
        MemorySpace* space = currentSpace();

        space->put( name, object );
    }

    void SetplayInterpreter::setWorldModel( IWorldModel *world_model)
    {
        M_world_model = world_model;
    }

    void SetplayInterpreter::setGlobalSpace( const MemorySpace* global_space)
    {
        M_global_space = global_space;
    }

    void SetplayInterpreter::setConditionEvaluator( IConditionEvaluator* condition_evaluator)
    {
        M_condition_evaluator = condition_evaluator;
    }

    void SetplayInterpreter::setActionExecutor( IActionExecutor* action_executor)
    {
        M_action_executor = action_executor;
    }

    void SetplayInterpreter::setTransitionSelector( ITransitionSelector* transition_selector)
    {
        M_transition_selector = transition_selector;
    }

    void SetplayInterpreter::setAbortDecider( IAbortDecider* abort_decider)
    {
        M_abort_decider = abort_decider;
    }

    void SetplayInterpreter::setStateUpdateDecider( IStateUpdateDecider* state_update_decider)
    {
        M_state_update_decider = state_update_decider;
    }

    /// Interpretation functions

    Object* SetplayInterpreter::interpret(const Object* object)
    {
        const IPoint2D* point = dynamic_cast<const IPoint2D*>( object );
        if ( point != NULL )
        {
            rcsc::Vector2D vector = interpret( *point );
            return new Point2D( new Decimal(vector.x),
                                new Decimal(vector.y));
        }

        const ICondition* condition = dynamic_cast<const ICondition*>( object );
        if ( condition != NULL )
        {
            return new Bool( interpret( *condition ) );
        }

        const IRegion2D* region = dynamic_cast<const IRegion2D*>( object );
        if ( region != NULL )
        {
            return calculate( *region );
        }

        const IDecimal* decimal = dynamic_cast<const IDecimal*>( object );
        if ( decimal != NULL )
        {
            return new Decimal( interpret( *decimal ));
        }

        const IPlayer* player = dynamic_cast<const IPlayer*>( object );
        if ( player != NULL )
        {
            PlayerAffiliation player_ref = interpret( *player );
            return player_ref.clone();
        }


        assert(false);
        return NULL;
    }


    void SetplayInterpreter::interpret(const Setplay& setplay)
    {
        /// Create memory space for step
        // Execute all assignments before entering steps
        if ( setplay.hasAssignments() )
        {
            // Should only do this once
            for ( Assignments::const_iterator ita = setplay.assignments()->begin();
                  ita != setplay.assignments()->end();
                  ita++ )
            {
                // Execute each assignment
                interpret( **ita );
            }
        }

        /*M_current_step = setplay.steps()->at(0);
        interpret( *M_current_step );*/
    }

    void SetplayInterpreter::interpret(const Step& step)
    {
        /// Create memory space for step
        MemorySpace* step_space = new MemorySpace("");
        M_objects_spaces.push_back(step_space);

        // Execute all assignments before entering transitions
        if ( step.hasAssignments() )
        {
            // Should only do this once
            for ( Assignments::const_iterator ita = step.assignments()->begin();
                  ita != step.assignments()->end();
                  ita++ )
            {
                // Execute each assignment
                interpret( **ita );
            }
        }
    }


    /// Transitions
    void SetplayInterpreter::interpret(const ITransition& transition)
    {
        MemorySpace* transition_space = new MemorySpace("");
        M_objects_spaces.push_back(transition_space);

        // Execute all assignments before entering transitions
        if ( transition.hasAssignments() )
        {
            // Should only do this once
            for ( Assignments::const_iterator ita = transition.assignments()->begin();
                  ita != transition.assignments()->end();
                  ita++ )
            {
                // Execute each assignment
                interpret( **ita );
            }
        }

        switch ( transition.type() )
        {
            case TT_TRANSITION_NEXT_STEP:
                interpret( static_cast<const NextStep&>(transition) );
                break;
            case TT_TRANSITION_ABORT:
                interpret( static_cast<const Abort&>(transition) );
                break;
            case TT_TRANSITION_FINISH:
                interpret( static_cast<const Finish&>(transition) );
                break;
            default:
                assert(false);
                break;
        }
    }

    void SetplayInterpreter::interpret(const NextStep& transition)
    {
        if ( transition.hasStartCondition() )
        {
            if ( interpret( *transition.startCondition() ))
            {
                // Start the execution of the transition => Change state

                // Get participant do actions
                if ( !transition.hasDirectives() )
                {
                    return;
                }

                for ( Directives::const_iterator itd = transition.directives()->begin();
                      itd != transition.directives()->end();
                      itd++)
                {
                    switch ( (*itd)->directiveType() )
                    {
                        case IDirective::DT_DO:
                            interpret( static_cast<const Do&>( **itd ));
                            break;
                        case IDirective::DT_DONT:
                            interpret( static_cast<const Dont&>( **itd ));
                            break;
                    }
                }
            }
        }
    }

    void SetplayInterpreter::interpret(const Finish& transition)
    {
        if ( transition.hasStartCondition() )
        {
            bool res = interpret( *transition.startCondition() );
            if ( res )
            {
                // Define this transition as the current transition
                // M_current_transition = transition.clone();
            }
        }

    }

    void SetplayInterpreter::interpret(const Abort& transition)
    {
        if ( transition.hasStartCondition() )
        {
            bool res = interpret( *transition.startCondition() );
            if ( res )
            {
                // Push actions to be executed
            }
        }

        // Evaluate the abort condition
        // if ( M_abort )
    }

    /// Directives

    void SetplayInterpreter::interpret(const Do& directive)
    {
        // Am I the player for which this directive is meant
        PlayerAffiliation self( M_world_model->self(),
                                        TA_OUR);

        bool self_in_directive = false;
        for ( Players::const_iterator itp = directive.players()->begin();
              itp != directive.players()->end() && !self_in_directive; itp++ )
        {
            PlayerAffiliation directive_player = interpret( **itp );
            self_in_directive = self == directive_player;
        }

        if ( !self_in_directive )
        {
            return;
        }

        if ( directive.hasActions() )
        {
            rcsc::dlog.addText( rcsc::Logger::PLAN,
                                __FILE__ " SetplayInterpreter: Appending %d actions",
                                directive.actions()->size() );

            for ( Actions::const_iterator ita = directive.actions()->begin();
                  ita != directive.actions()->end() ; ita++ )
            {
                M_current_actions.push_back( (*ita)->clone() );
                stringstream ss;
                SetplayGeneratorSexpr sexpr(ss);
                (*ita)->visit( sexpr );
                rcsc::dlog.addText( rcsc::Logger::PLAN,
                                    __FILE__ " SetplayInterpreter: Action %s",
                                    ss.str().c_str() );
            }
        }
    }

    void SetplayInterpreter::interpret(const Dont& )
    {
        return;
    }

    void SetplayInterpreter::interpret(const DoWhile& )
    {
        return;
    }

    /// Actions
    bool SetplayInterpreter::interpret(const IAction& action) const
    {
        switch ( action.type() )
        {
            case TT_ACTION_DRIBBLE:
                return interpret( static_cast<const ActDribble&>(action) );
            case TT_ACTION_FOLLOW:
                return interpret( static_cast<const ActFollow&>(action) );
            case TT_ACTION_INTERCEPT:
                return interpret( static_cast<const ActIntercept&>(action) );
            case TT_ACTION_PASS:
                return interpret( static_cast<const ActPass&>(action) );
            case TT_ACTION_POS:
                return interpret( static_cast<const ActPos&>(action) );
            case TT_ACTION_SHOOT:
                return M_action_executor->shoot();
            case TT_ACTION_HOLD:
                return M_action_executor->holdBall();
            case TT_ACTION_LOOK_AT:
                return interpret( static_cast<const ActLookAt&>(action) );
            default:
            {
                std::cerr << "Unhandled action type" << action.type() << " - "  << action.actionType() << std::endl;
                assert(false);
            }
        }
    }

    bool SetplayInterpreter::interpret(const ActDribble& action) const
    {
        bool res;
        switch ( action.location()->type() )
        {
            case TT_LINE:
            {
                assert(false);
                /*rcsc::Line2D line = interpret( static_cast<Line2D&>(*action.location()));
                rcsc::Vector2D self_position = M_world_model->playerPosition( M_self->affiliation(),
                                                                              M_self->number());
                rcsc::Vector2D closest_intersection = line.projection( self_position );
                res = M_action_executor->moveTo( closest_intersection );*/
            }
                break;
            case TT_RAY:
            {
                assert(false);
                /*rcsc::Ray2D ray = interpret( static_cast<const Ray2D&>(*action.location()));
                rcsc::Vector2D self_position = M_world_model->playerPosition( M_self->affiliation(),
                                                                              M_self->number());
                rcsc::Vector2D closest_intersection= ray.nearestPoint(self_position);
                res = M_action_executor->moveTo( closest_intersection );*/
            }
                break;
            case TT_RECTANGLE:
            case TT_CIRCLE:
            case TT_TRIANGLE:
            case TT_SECTOR:
            {
                rcsc::Region2D* region = interpret( static_cast<const IRegion2D&>( *action.location()) );
                res = M_action_executor->dribbleTo(*region);
                delete region;
            }
            break;
            case TT_POINT:
            {
                rcsc::Vector2D point = interpret( static_cast<const IPoint2D&>( *action.location() ));
                res = M_action_executor->dribbleTo(point);
            }
            break;
            default:
                assert(false);
                break;
        }
        return res;
    }

    bool SetplayInterpreter::interpret(const ActFollow& action) const
    {
        bool res = false;
        switch( action.target()->mobileObjectType() )
        {
            case IMobileObject::MOT_BALL:
            {
                rcsc::Vector2D point = interpret( *action.displacement() );
                res = M_action_executor->followBall(point);
            }
            break;
            case IMobileObject::MOT_PLAYER:
            {
                PlayerAffiliation player_affiliation = interpret( *static_cast<const IPlayer*>( action.target() ) );
                res = M_action_executor->followPlayer( player_affiliation );
            }
            break;
        }
        return res;
    }

    bool SetplayInterpreter::interpret(const ActHold& action) const
    {
        return M_action_executor->holdBall();
    }

    bool SetplayInterpreter::interpret(const ActIntercept& action) const
    {
        bool res = false;
        switch( action.target()->mobileObjectType() )
        {
            case IMobileObject::MOT_BALL:
                res = M_action_executor->interceptBall( );
                break;
            case IMobileObject::MOT_PLAYER:
            {
                PlayerAffiliation player_affiliation = interpret( *static_cast<const IPlayer*>( action.target() ) );
                res = M_action_executor->interceptPlayer( player_affiliation );
            }
            break;
        }
        return res;
    }

    bool SetplayInterpreter::interpret(const ActPass& action) const
    {
        bool res = false;
        switch ( action.location()->locationType() )
        {
            case ILocation2D::LT_POINT:
            {
                const IPoint2D& ipoint = static_cast< IPoint2D& >( *action.location() );
                if ( ipoint.type() == TT_PLAYER_POSITION )
                {
                    const PlayerPosition& player_position = static_cast<const PlayerPosition&>( ipoint );
                    PlayerAffiliation player = interpret( *player_position.player() );
                    res = M_action_executor->passToPlayer( player );
                }
                else
                {
                    // TODO Check if it is a player, otherwise just resolve as point
                    rcsc::Vector2D point = interpret( ipoint );
                    res = M_action_executor->passToPoint( point );
                }
            }
            break;
            case ILocation2D::LT_REGION:
            {
                const IRegion2D& iregion = static_cast< IRegion2D& >( *action.location() );
                rcsc::Region2D* region = interpret( iregion);
                res = M_action_executor->passToRegion( *region );
            }
            break;
            default:
            std::cerr << "Wrong location type " << typeid(action.location()).name() << std::endl;
                assert(false);
        }

        return res;
    }

    bool SetplayInterpreter::interpret(const ActPos& action) const
    {
        bool res = false;
        switch ( action.location()->locationType() )
        {
            case ILocation2D::LT_REGION:
            {
                rcsc::Region2D* region = interpret( *static_cast<const IRegion2D*>(action.location()) );
                res = M_action_executor->moveTo(region->centroid());
                delete region;
            }
            break;
            case ILocation2D::LT_POINT:
            {
                rcsc::Vector2D point = interpret( *static_cast<const IPoint2D*>( action.location() ) );
                res = M_action_executor->moveTo(point);
            }
            break;
            case ILocation2D::LT_RAY:
            {
                assert(false);
            }
            break;
            case ILocation2D::LT_MOBILE_OBJECT:
            {
                assert(false);
            }
            break;
            case ILocation2D::LT_LINE:
            {
                assert(false);
            }
            break;
            case ILocation2D::LT_VARIABLE:
            {
                const VarLocation* var_location = static_cast<const VarLocation*>( action.location() );
                ILocation2D* location = static_cast<ILocation2D*>(resolve( var_location->name() )->clone());
                res = interpret( ActPos( location ) );
            }
            break;
        }
        return res;
    }

    bool SetplayInterpreter::interpret(const ActLookAt& action) const
    {
        std::vector<PlayerAffiliation> players;
        for ( std::vector<IPlayer*>::const_iterator itp = action.players().begin();
              itp != action.players().end(); itp++ )
        {
            PlayerAffiliation player = interpret( **itp );
            players.push_back( player );
        }
        return M_action_executor->lookAt( players );
    }

    /// Conditions
    bool SetplayInterpreter::interpret(const CondAnd& condition) const
    {
        for ( Conditions::const_iterator itc = condition.conditions()->begin();
              itc != condition.conditions()->end();
              itc++)
        {
            if ( interpret( **itc ) == false ) { return false; }
        }
        return true;
    }

    bool SetplayInterpreter::interpret(const CondOr& condition) const
    {
        for ( Conditions::const_iterator itc = condition.conditions()->begin();
              itc != condition.conditions()->end();
              itc++)
        {
            if ( interpret( **itc ) == true ) { return true; }
        }
        return false;
    }

    bool SetplayInterpreter::interpret(const CondNot& condition) const
    {
        return !interpret( *condition.condition() );
    }

    bool SetplayInterpreter::interpret(const CondBallOwner& condition) const
    {
        PlayerAffiliation player = interpret( *condition.player() );
        return M_condition_evaluator->ownsBall( player );
    }

    bool SetplayInterpreter::interpret(const CondTeamInPossession& condition) const
    {
        return M_condition_evaluator->isTeamInPossession( condition.teamInPossession() );
        //  condition.teamInPossession() == M_world_model->teamInPossession();
    }

    bool SetplayInterpreter::interpret(const CondBool& condition) const
    {
        return condition.state() == true;
    }

    bool SetplayInterpreter::interpret(const CondGameSituation& condition) const
    {
        return M_condition_evaluator->isGameSituation( condition.gameSituation() );
    }

    bool SetplayInterpreter::interpret(const CondAtLocation& condition) const
    {
        bool res = false;
        switch ( condition.location()->locationType() )
        {
            case ILocation2D::LT_LINE:
            {
                assert(false);
            }
            break;
            case ILocation2D::LT_RAY:
            {
                assert(false);
            }
            break;
            case ILocation2D::LT_POINT:
            {
                rcsc::Vector2D position = interpret( static_cast<const IPoint2D&>( *condition.location() ));

                switch ( condition.object()->mobileObjectType() )
                {
                    case IMobileObject::MOT_BALL:
                    {
                        res = M_condition_evaluator->ballAtPosition( position  );
                    }
                    break;
                    case IMobileObject::MOT_PLAYER:
                    {
                        PlayerAffiliation player = interpret( static_cast<const IPlayer&>( *condition.object()) );
                        res = M_condition_evaluator->playerAtPosition( player, position  );
                    }
                    break;
                }
            }
            break;
            case ILocation2D::LT_REGION:
            {
                rcsc::Region2D* region = interpret( static_cast<const IRegion2D&>( *condition.location() ));
                switch ( condition.object()->mobileObjectType() )
                {
                    case IMobileObject::MOT_BALL: // Wrong type
                        res = M_condition_evaluator->ballAtRegion( *region );
                        break;
                    case IMobileObject::MOT_PLAYER:
                    {
                        PlayerAffiliation player = interpret( static_cast<const IPlayer&>( *condition.object()) );
                        res = M_condition_evaluator->playerAtRegion( player, *region );
                    }
                    break;
                    default:
                        assert(false);
                    break;
                }
                delete region;
            }
            break;
            case ILocation2D::LT_MOBILE_OBJECT:
            {
                assert(false);
            }
            break;
            case ILocation2D::LT_VARIABLE:
            {
                const VarLocation* var_location = static_cast<const VarLocation*>( condition.location() );
                ILocation2D* location = static_cast<const ILocation2D*>(resolve( var_location->name() ))->clone();
                location->visit( SetplayGeneratorSexprLog );
                CondAtLocation cond( condition.object()->clone(), location );
                cond.visit( SetplayGeneratorSexprLog );
                res = interpret( cond );
            }
            break;
        }
        return res;
    }


    bool SetplayInterpreter::interpret(const CondTeamFormation& ) const
    {
        assert(false);
        return false; // HACK
    }

    bool SetplayInterpreter::interpret(const CondOffside& condition) const
    {
        PlayerAffiliation player = interpret( *condition.player() );
        return M_condition_evaluator->isOffside( player );
    }

    bool SetplayInterpreter::interpret(const CondCanPass& condition) const
    {
        // TODO HACK
        return true;
        // return M_condition_evaluator->isOffside( player.affiliation(), player.number() );
    }


    bool SetplayInterpreter::interpret(const CondCanShoot& condition) const
    {
        // TODO HACK
        return true;
    }

    bool SetplayInterpreter::interpret(const CondCanIntercept& condition) const
    {
        PlayerAffiliation player = interpret( *condition.player() );
        return M_condition_evaluator->canInterceptBall( player );
    }

    bool SetplayInterpreter::interpret(const CondCompDecimal& condition) const
    {
        float operand_left = interpret( *condition.operandLeft() );
        float operand_right = interpret( *condition.operandRight() );
        bool res;
        switch ( condition.operationType() )
        {
            case NCOT_LOWER:
                res = operand_left < operand_right;
                break;
            case NCOT_LOWER_OR_EQUAL:
                res = operand_left <= operand_right;
                break;
            case NCOT_EQUAL:
                res = operand_left == operand_right;
                break;
            break;
            case NCOT_NOT_EQUAL:
                res = operand_left != operand_right;
                break;
            case NCOT_GREATER_OR_EQUAL:
                res = operand_left >= operand_right;
                break;
            case NCOT_GREATER:
                res = operand_left > operand_right;
                break;
            default:
                assert(false);
        }
        return res;
    }

    bool SetplayInterpreter::interpret(const CondKickable& condition) const
    {
        PlayerAffiliation player = interpret( *condition.player() );
        return M_condition_evaluator->isBallKickable( player );
    }

    /// Locations
    rcsc::Rect2D* SetplayInterpreter::interpret(const Rect2D& location) const
    {
        rcsc::Vector2D top_left = interpret( *location.topLeft() );
        rcsc::Vector2D bottom_right = interpret( *location.bottomRight() );
        return new rcsc::Rect2D(top_left, bottom_right);
    }

    rcsc::Triangle2D* SetplayInterpreter::interpret(const Triangle2D& location) const
    {
        rcsc::Vector2D a = interpret( *location.a() );
        rcsc::Vector2D b = interpret( *location.b() );
        rcsc::Vector2D c = interpret( *location.c() );
        return new rcsc::Triangle2D(a, b, c);
    }

    rcsc::Sector2D* SetplayInterpreter::interpret(const Sector2D& location) const
    {
        rcsc::Vector2D center = interpret( *location.center() );
        float minRadius = interpret( *location.minRadius() );
        float maxRadius = interpret( *location.maxRadius() );
        float startRadius = interpret( *location.startAngle() );
        float endRadius = interpret( *location.endAngle() );
        return new rcsc::Sector2D(center,
                                  minRadius, maxRadius,
                                  rcsc::AngleDeg(startRadius), rcsc::AngleDeg(endRadius));
    }

    rcsc::Circle2D* SetplayInterpreter::interpret(const Circle2D& location) const
    {
        rcsc::Vector2D center = interpret( *location.center() );
        float radius = interpret( *location.radius() );
        return new rcsc::Circle2D(center, radius);
    }

    rcsc::Ray2D SetplayInterpreter::interpret(const Ray2D& location) const
    {
        rcsc::Vector2D origin = interpret( *location.origin() );
        float direction = interpret( *location.direction() );

        return rcsc::Ray2D(origin, rcsc::AngleDeg(direction));
    }

    rcsc::Line2D SetplayInterpreter::interpret(const Line2D& location) const
    {
        float a = interpret( *location.a() );
        float b = interpret( *location.b() );
        float c = interpret( *location.c() );
        return rcsc::Line2D(a, b, c);
    }

    /// Numbers
    float SetplayInterpreter::interpret(const IDecimal& decimal) const
    {
        switch ( decimal.type() )
        {
            case TT_DECIMAL:
                return interpret( static_cast<const Decimal&>(decimal) );
            case TT_NUMBER_ABS:
                return interpret( static_cast<const NumberAbs&>(decimal) );
            case TT_NUMBER_SIGN_CHANGE:
                return interpret( static_cast<const NumberSignChange&>(decimal) );
            case TT_NUMBER_MULTIPLICATION:
                return interpret( static_cast<const NumberMultiplication&>(decimal) );
            case TT_NUMBER_SUBTRACTION:
                return interpret( static_cast<const NumberSubtraction&>(decimal) );
            case TT_NUMBER_ADDITION:
                return interpret( static_cast<const NumberAddition&>(decimal) );
            case TT_NUMBER_DIVISION:
                return interpret( static_cast<const NumberDivision&>(decimal) );
            case TT_OFFENSE_LINE:
                return interpret( static_cast<const OffenseLine&>(decimal) );
            case TT_OFFSIDE_LINE:
                return interpret( static_cast<const OffsideLine&>(decimal) );
            case TT_DEFENSE_LINE:
                return interpret( static_cast<const DefenseLine&>(decimal) );
            case TT_VECTOR_DIRECTION:
                return interpret( static_cast<const VectorDirection&>(decimal) );
            case TT_VECTOR_LENGTH:
                return interpret( static_cast<const VectorLength&>(decimal) );
            case TT_VECTOR_DISTANCE:
                return interpret( static_cast<const VectorDistance&>(decimal) );
            case TT_VECTOR_ANGULAR_DIFF:
                return interpret( static_cast<const VectorAngularDiff&>(decimal) );
            case TT_VECTOR_INNER_PRODUCT:
                return interpret( static_cast<const VectorInnerProduct&>(decimal) );
            case TT_VECTOR_OUTER_PRODUCT:
                return interpret( static_cast<const VectorOuterProduct&>(decimal) );
            case TT_CURRENT_TIME:
                return interpret( static_cast<const CurrentTime&>(decimal) );
            case TT_REMAINING_TIME:
                return interpret( static_cast<const RemainingTime&>(decimal) );
            case TT_GOALS_SCORED:
                return interpret( static_cast<const GoalsScored&>(decimal) );
            case TT_GOAL_DIFF:
                return interpret( static_cast<const GoalDiff&>(decimal) );
            case TT_MOBILE_OBJECT_SPEED:
                return interpret( static_cast<const MobileObjectSpeed&>(decimal) );
            case TT_VAR_DECIMAL:
            {
                const VarDecimal& decimal_var = static_cast<const VarDecimal&>( decimal );
                const IDecimal* idecimal = static_cast<const IDecimal*>( resolve( decimal_var.name() ) );
                return interpret( *idecimal );
            }
            default:
                assert(false);
        }
        return 0.0;
    }


    float SetplayInterpreter::interpret(const Decimal& number) const
    {
        return number.value();
    }

    float SetplayInterpreter::interpret(const Integer& number) const
    {
        return number.value();
    }

    const Object *SetplayInterpreter::interpret(const Variable &number) const
    {
        return resolve( number.name() );
    }

    /// Decimal functors
    // Number operations
    float SetplayInterpreter::interpret( const NumberAbs& decimal_functor ) const
    {
        float operand = interpret( *decimal_functor.operand() );
        return fabs(operand);
    }

    float SetplayInterpreter::interpret( const NumberSignChange& decimal_functor ) const
    {
        float operand = interpret( *decimal_functor.operand() );
        return -operand;
    }

    float SetplayInterpreter::interpret( const NumberMultiplication& decimal_functor ) const
    {
        float operand_left = interpret( *decimal_functor.operandLeft() );
        float operand_right = interpret( *decimal_functor.operandRight() );
        return operand_left * operand_right;
    }

    float SetplayInterpreter::interpret( const NumberSubtraction& decimal_functor ) const
    {
        float operand_left = interpret( *decimal_functor.operandLeft() );
        float operand_right = interpret( *decimal_functor.operandRight() );
        return operand_left - operand_right;
    }

    float SetplayInterpreter::interpret( const NumberDivision& decimal_functor ) const
    {
        float operand_left = interpret( *decimal_functor.operandLeft() );
        float operand_right = interpret( *decimal_functor.operandRight() );
        assert( operand_right != 0.0 );
        return operand_left / operand_right;
    }

    float SetplayInterpreter::interpret( const NumberAddition& decimal_functor ) const
    {
        float operand_left = interpret( *decimal_functor.operandLeft() );
        float operand_right = interpret( *decimal_functor.operandRight() );
        return operand_left + operand_right;
    }

    // Vector to scalar operations
    float SetplayInterpreter::interpret( const VectorDirection& decimal_functor ) const
    {
        rcsc::Vector2D operand = interpret( *decimal_functor.operand() );
        return operand.dir().degree();
    }

    float SetplayInterpreter::interpret( const VectorLength& decimal_functor ) const
    {
        rcsc::Vector2D operand = interpret( *decimal_functor.operand() );
        return operand.length();
    }

    // Two vector to scalar operations
    float SetplayInterpreter::interpret( const VectorDistance& decimal_functor ) const
    {
        rcsc::Vector2D operand_left = interpret( *decimal_functor.operandLeft() );
        rcsc::Vector2D operand_right = interpret( *decimal_functor.operandRight() );
        return operand_left.dist(operand_right);
    }

    float SetplayInterpreter::interpret( const VectorAngularDiff& decimal_functor ) const
    {
        rcsc::Vector2D operand_left = interpret( *decimal_functor.operandLeft() );
        rcsc::Vector2D operand_right = interpret( *decimal_functor.operandRight() );
        return (operand_left - operand_right).dir().degree();
    }

    float SetplayInterpreter::interpret( const VectorInnerProduct& decimal_functor ) const
    {
        rcsc::Vector2D operand_left = interpret( *decimal_functor.operandLeft() );
        rcsc::Vector2D operand_right = interpret( *decimal_functor.operandRight() );
        return operand_left.innerProduct(operand_right);
    }

    float SetplayInterpreter::interpret( const VectorOuterProduct& decimal_functor ) const
    {
        rcsc::Vector2D operand_left = interpret( *decimal_functor.operandLeft() );
        rcsc::Vector2D operand_right = interpret( *decimal_functor.operandRight() );
        return operand_left.outerProduct(operand_right);
    }

    // In game queries

    float SetplayInterpreter::interpret(const DefenseLine& decimal_functor) const
    {
        return M_world_model->defenseLine( decimal_functor.teamAffiliation() );
    }

    float SetplayInterpreter::interpret(const OffenseLine& decimal_functor) const
    {
        return M_world_model->offenseLine( decimal_functor.teamAffiliation() );
    }

    float SetplayInterpreter::interpret(const OffsideLine& decimal_functor) const
    {
        return M_world_model->offsideLine( decimal_functor.teamAffiliation() );
    }

    float SetplayInterpreter::interpret(const MobileObjectSpeed& decimal_functor) const
    {
        switch ( decimal_functor.mobileObject()->mobileObjectType() )
        {
            case IMobileObject::MOT_BALL:
                return M_world_model->ballSpeed();
            case IMobileObject::MOT_PLAYER:
            {
                PlayerAffiliation player = interpret( static_cast<IPlayer&>( *decimal_functor.mobileObject() ) );
                return M_world_model->playerSpeed( player );
            }
        }
    }

    float SetplayInterpreter::interpret(const PointX& decimal_functor) const
    {
        rcsc::Vector2D point = interpret( *decimal_functor.point() );
        return point.x;
    }

    float SetplayInterpreter::interpret(const PointY& decimal_functor) const
    {
        rcsc::Vector2D point = interpret( *decimal_functor.point() );
        return point.y;
    }

    float SetplayInterpreter::interpret(const CurrentTime& ) const
    {
        return M_world_model->currentTime();
    }

    float SetplayInterpreter::interpret(const RemainingTime& ) const
    {
        return M_world_model->remainingTimeInPeriod();
    }

    float SetplayInterpreter::interpret(const GoalsScored& decimal_functor ) const
    {
        return M_world_model->goalsScored( decimal_functor.teamAffiliation() );
    }

    float SetplayInterpreter::interpret(const GoalDiff& ) const
    {
        return M_world_model->goalsScored( TA_OUR ) -
               M_world_model->goalsScored( TA_OPP );
    }

    float SetplayInterpreter::interpret(const InterceptionTime& decimal_functor ) const
    {
        PlayerAffiliation player1 = interpret( *decimal_functor.player() );
        switch ( decimal_functor.mobileObject()->mobileObjectType() )
        {
            case IMobileObject::MOT_BALL:
                return M_world_model->ballInterceptionTime( player1 );
            case IMobileObject::MOT_PLAYER:
            {
                PlayerAffiliation player2 = interpret( static_cast<IPlayer&>( *decimal_functor.mobileObject()) );
                return M_world_model->playerInterceptionTime( player1 , player2 );
            }
            break;
        }
    }

    /// Points
    rcsc::Vector2D SetplayInterpreter::interpret(const Point2D& point) const
    {
        double x = interpret( *point.x() );
        double y = interpret( *point.y() );

        return rcsc::Vector2D(x, inverted() ? y : -y );
    }

    /// Point functors
    rcsc::Vector2D SetplayInterpreter::interpret(const PlayerPosition& point_functor) const
    {
        PlayerAffiliation player = interpret( *point_functor.player() );
        return M_world_model->playerPosition( player );
    }

    rcsc::Vector2D SetplayInterpreter::interpret(const BallPosition& ) const
    {
        return M_world_model->ballPosition();
    }

    // Vector unary operations
    rcsc::Vector2D SetplayInterpreter::interpret( const VectorNormalize& point_functor ) const
    {
        rcsc::Vector2D operand = interpret( *point_functor.operand() );
        return operand.normalize();
    }

    rcsc::Vector2D SetplayInterpreter::interpret( const VectorInvertX& point_functor ) const
    {
        rcsc::Vector2D operand = interpret( *point_functor.operand() );
        return rcsc::Vector2D(-operand.x, operand.y);
    }

    rcsc::Vector2D SetplayInterpreter::interpret( const VectorInvertY& point_functor ) const
    {
        rcsc::Vector2D operand = interpret( *point_functor.operand() );
        return rcsc::Vector2D(operand.x, -operand.y);
    }

    rcsc::Vector2D SetplayInterpreter::interpret( const VectorInvertXY& point_functor ) const
    {
        rcsc::Vector2D operand = interpret( *point_functor.operand() );
        return rcsc::Vector2D(-operand.x, -operand.y);
    }

    rcsc::Vector2D SetplayInterpreter::interpret( const VectorSubtraction& point_functor ) const
    {
        rcsc::Vector2D operand_left = interpret( *point_functor.operandLeft() );
        rcsc::Vector2D operand_right = interpret( *point_functor.operandRight() );
        return operand_left - operand_right;
    }

    rcsc::Vector2D SetplayInterpreter::interpret( const VectorAddition& point_functor ) const
    {
        rcsc::Vector2D operand_left = interpret( *point_functor.operandLeft() );
        rcsc::Vector2D operand_right = interpret( *point_functor.operandRight() );
        return operand_left + operand_right;
    }

    // Vector scalar operations
    rcsc::Vector2D SetplayInterpreter::interpret( const VectorScale& point_functor ) const
    {
        rcsc::Vector2D operand_left = interpret( *point_functor.operandLeft() );
        float operand_right = interpret( *point_functor.operandRight() );
        return operand_left * operand_right;
    }

    rcsc::Vector2D SetplayInterpreter::interpret( const VectorRotation& point_functor ) const
    {
        rcsc::Vector2D operand_left = interpret( *point_functor.operandLeft() );
        float operand_right = interpret( *point_functor.operandRight() );
        return operand_left.rotate(operand_right);
    }

    rcsc::Vector2D SetplayInterpreter::interpret( const VectorDivision& point_functor ) const
    {
        rcsc::Vector2D operand_left = interpret( *point_functor.operandLeft() );
        float operand_right = interpret( *point_functor.operandRight() );
        return operand_left / operand_right;
    }


    /// Variables


    /// End Implementation of IVisitor interface


    bool SetplayInterpreter::interpret(const ICondition& condition) const
    {
        switch ( condition.type() )
        {
            case TT_CONDITION_AND:
                return interpret( static_cast<const CondAnd&>( condition ) );
            case TT_CONDITION_OR:
                return interpret( static_cast<const CondOr&>( condition ) );
            case TT_CONDITION_NOT:
                return interpret( static_cast<const CondNot&>( condition ) );
            case TT_CONDITION_BOOL:
                return interpret ( static_cast<const CondBool&>( condition ) );
            case TT_CONDITION_OBJECT_LOCATION:
                return interpret( static_cast<const CondAtLocation&>( condition ) );
            case TT_CONDITION_COMP_DECIMAL:
                return interpret( static_cast<const CondCompDecimal&>( condition ) );
            case TT_CONDITION_BALL_OWNER:
                return interpret( static_cast<const CondBallOwner&>( condition ) );
            case TT_CONDITION_TEAM_FORMATION:
                return interpret( static_cast<const CondTeamFormation&>( condition ) );
            case TT_CONDITION_GAME_SITUATION:
                return interpret( static_cast<const CondGameSituation&>( condition ) );
            case TT_CONDITION_KICKABLE:
                return interpret( static_cast<const CondKickable&>( condition ));
            case TT_CONDITION_TEAM_IN_POSSESSION:
                return interpret ( static_cast<const CondTeamInPossession&>( condition ));
            case TT_CONDITION_OFFSIDE:
                return interpret ( static_cast<const CondOffside&>( condition ));
            case TT_CONDITION_CAN_PASS:
                return interpret ( static_cast<const CondCanPass&>( condition ));
            case TT_CONDITION_CAN_INTERCEPT:
                return interpret ( static_cast<const CondCanIntercept&>( condition ));
            case TT_CONDITION_CAN_SHOOT:
                return interpret ( static_cast<const CondCanShoot&>( condition ));
            default:
                std::cerr << "SetplayInterpreter: Condition interpreter unknown condition "
                          << typeid(condition).name()
                          << std::endl;
                assert(false);
                break;
        }
    }


    rcsc::Vector2D SetplayInterpreter::interpret(const IPoint2D& point) const
    {
        switch ( point.type() )
        {
            case TT_POINT:
                return interpret( static_cast<const Point2D&>( point ) );
            // Point functors
            case TT_VECTOR_NORMALIZE:
                return interpret( static_cast<const VectorNormalize&>( point ) );
            case TT_VECTOR_INVERT_X:
                return interpret( static_cast<const VectorInvertX&>( point ) );
            case TT_VECTOR_INVERT_Y:
                return interpret( static_cast<const VectorInvertY&>( point ) );
            case TT_VECTOR_INVERT_XY:
                return interpret( static_cast<const VectorInvertXY&>( point ) );
            case TT_VECTOR_SUBTRACTION:
                return interpret( static_cast<const VectorSubtraction&>( point ) );
            case TT_VECTOR_ADDITION:
                return interpret( static_cast<const VectorAddition&>( point ) );
            case TT_VECTOR_SCALE:
                return interpret( static_cast<const VectorScale&>( point ) );
            case TT_VECTOR_ROTATION:
                return interpret( static_cast<const VectorRotation&>( point ) );
            case TT_VECTOR_DIVISION:
                return interpret( static_cast<const VectorDivision&>( point ) );
            case TT_PLAYER_POSITION:
                return interpret( static_cast<const PlayerPosition&>( point ) );
            case TT_BALL_POSITION:
                return interpret( static_cast<const BallPosition&>( point ) );
            case TT_VAR_POINT:
            {
                const VarPoint& point_var = static_cast<const VarPoint&>( point  );
                const IPoint2D* ipoint = static_cast<const IPoint2D*>( resolve( point_var.name()) );
                return interpret( *ipoint );
            }
        default:
                std::cerr << "SetplayInterpreter: Point interpreter unknown point "
                      << typeid(point).name()
                      << std::endl;
                assert(false);
                break;
        }
    }

    /* Calculations used for assignment operations */
    Object* SetplayInterpreter::calculate(const Object* object)
    {
        const IPoint2D* ipoint = dynamic_cast<const IPoint2D*>( object );
        if ( ipoint != NULL )
        {
            rcsc::Vector2D point = interpret( *ipoint );
            return new Point2D( new Decimal(point.x) , new Decimal(point.y) );
        }

        const IRegion2D* region = dynamic_cast<const IRegion2D*>( object );
        if ( region != NULL )
        {
            return calculate( *region );
        }

        assert(false);
        return NULL;
    }

    IRegion2D* SetplayInterpreter::calculate(const IRegion2D& region)
    {
        IRegion2D* iregion;
        switch ( iregion->type() )
        {
            case TT_RECTANGLE:
            {
                rcsc::Rect2D* rect = interpret(static_cast<const Rect2D&>( region ) );
                iregion = new Rect2D( new Point2D( new Decimal( rect->left() ), new Decimal( rect->top())),
                                      new Point2D( new Decimal( rect->right() ), new Decimal( rect->bottom())));
                delete rect;
            }
            break;
            case TT_TRIANGLE:            
            {
                rcsc::Triangle2D* triangle = interpret(static_cast<const Triangle2D&>( region ) );
                iregion = new Triangle2D( new Point2D( new Decimal( triangle->a().x ), new Decimal( triangle->a().y )),
                                          new Point2D( new Decimal( triangle->b().x ), new Decimal( triangle->b().y )),
                                          new Point2D( new Decimal( triangle->c().x ), new Decimal( triangle->c().y )));
                delete triangle;
            }
            break;
            case TT_CIRCLE:
            {
                rcsc::Circle2D* circle = interpret(static_cast<const Circle2D&>( region ) );
                iregion = new Circle2D( new Point2D( new Decimal( circle->center().x ), new Decimal( circle->center().y )),
                                        new Decimal( circle->radius() ));
                delete circle;
            }
            break;
            case TT_SECTOR:
            {
                rcsc::Sector2D* sector = interpret(static_cast<const Sector2D&>( region ) );
                iregion = new Sector2D( new Point2D( new Decimal( sector->center().x ), new Decimal( sector->center().y )),
                                        new Decimal( sector->radiusMin() ),
                                        new Decimal( sector->radiusMax() ),
                                        new Decimal( sector->angleLeftStart().degree() ),
                                        new Decimal( sector->angleRightEnd().degree() ) );
                delete sector;
            }
            break;
            case TT_VAR_REGION:
            {
                const VarRegion& region_var = static_cast<const VarRegion&>( region_var ) ;
                iregion = calculate( static_cast<const IRegion2D&>( *resolve( region_var.name()) ));
            }
            break;
            default:
                std::cerr << "SetplayInterpreter: Region interpreter unknown region "
                      << typeid(region).name()
                      << std::endl;
                assert(false);
            break;
        }
        return iregion;
    }

    rcsc::Region2D* SetplayInterpreter::interpret(const IRegion2D& region) const
    {
        switch ( region.type() )
        {
            case TT_RECTANGLE:
            {
                const Rect2D& rect = static_cast<const Rect2D&>(region);
                rcsc::Vector2D top_left = interpret( *rect.topLeft() );
                rcsc::Vector2D bottom_right = interpret( *rect.bottomRight() );
                return new rcsc::Rect2D(top_left, bottom_right);
            }
            break;
            case TT_TRIANGLE:
            {
                const Triangle2D& triangle = static_cast<const Triangle2D&>(region);
                rcsc::Vector2D a = interpret( *triangle.a() );
                rcsc::Vector2D b = interpret( *triangle.b() );
                rcsc::Vector2D c = interpret( *triangle.c() );
                return new rcsc::Triangle2D(a, b, c);
            }
            break;
            case TT_CIRCLE:
            {
                const Circle2D& circle = static_cast<const Circle2D&>(region);
                rcsc::Vector2D center = interpret( *circle.center() );
                double radius = interpret( *circle.radius() );
                return new rcsc::Circle2D( center, radius );
            }
            break;
            case TT_SECTOR:
            {
                const Sector2D& sector = static_cast<const Sector2D&>(region);
                rcsc::Vector2D center = interpret( *sector.center() );
                double min_radius = interpret( *sector.minRadius() );
                double max_radius = interpret( *sector.maxRadius() );
                double start_angle = interpret( *sector.startAngle() );
                double end_angle = interpret( *sector.endAngle() );
                return new rcsc::Sector2D( center,
                                           min_radius, max_radius,
                                           rcsc::AngleDeg(start_angle), rcsc::AngleDeg(end_angle));
            }
            break;
            case TT_VAR_REGION:
            {
                const VarRegion& var_region = static_cast<const VarRegion&>( region );
                return interpret ( static_cast<const IRegion2D&>( *resolve( var_region.name() )));
            }
            break;
            default:
                assert(false);
                break;
        }
    }

    /// Participant
    PlayerAffiliation SetplayInterpreter::interpret(const IPlayer& player) const
    {
        switch ( player.type() )
        {
            case TT_PLAYER_AFFILIATION:
                return static_cast<const PlayerAffiliation&>(player);
            case TT_VAR_PLAYER:
                return interpret( static_cast<const VarPlayer&>( player ) );
            case TT_NEAREST_PLAYER:
                return interpret( static_cast<const NearestPlayer&>( player ) );
            case TT_NEAREST_TEAMMATE:
                return interpret( static_cast<const NearestTeammate&>( player ) );
            case TT_NEAREST_OPPONENT:
                return interpret( static_cast<const NearestOpponent&>( player ) );
            case TT_GOALIE:
                return interpret( static_cast<const Goalie&>( player ) );
            case TT_PLAYER_SELF:
                return interpret( static_cast<const PlayerSelf&>( player ) );
            default:
                std::cerr << typeid(player).name() << " " << player.type() << std::endl;
                assert(false);
                break;
        }
    }

    PlayerAffiliation SetplayInterpreter::interpret(const PlayerAffiliation& player) const
    {
        return player;
    }


    PlayerAffiliation SetplayInterpreter::interpret(const VarPlayer& player) const
    {
        return interpret( *static_cast<const IPlayer*>( resolve( player.name() ) ) );
    }

    /// Player functors
    PlayerAffiliation SetplayInterpreter::interpret(const PlayerSelf& ) const
    {
        return PlayerAffiliation( M_world_model->self(), TA_OUR );
    }

    PlayerAffiliation SetplayInterpreter::interpret(const NearestPlayer& player_functor) const
    {
        /// Switch location type: Region, Point, Player
        switch ( player_functor.location()->locationType() )
        {
            case ILocation2D::LT_REGION:
            {
                rcsc::Region2D* region = interpret( static_cast<const IRegion2D&>( *player_functor.location() ));
                PlayerAffiliation player = M_world_model->playerFastestTo( *region );
                delete region;
                return player;
            }
            case ILocation2D::LT_POINT:
            {
                rcsc::Vector2D position = interpret( static_cast<const IPoint2D&>( *player_functor.location() ));
                PlayerAffiliation player = M_world_model->playerFastestTo( position );
                return player;
            }
            case ILocation2D::LT_MOBILE_OBJECT:
            {
                const MobileObjectLocation& mol = static_cast<const MobileObjectLocation&>( *player_functor.location() );
                switch ( mol.mobileObject()->mobileObjectType() )
                {
                    case IMobileObject::MOT_PLAYER:
                    {
                        PlayerAffiliation player_ref = interpret( static_cast<const IPlayer&>( *mol.mobileObject() ));
                        return M_world_model->playerFastestTo( player_ref );
                    }
                    case IMobileObject::MOT_BALL:
                    {
                        PlayerAffiliation player = M_world_model->playerFastestToBall();
                        return player;
                    }
                }
            }
            default:
                assert( false );
        }
    }

    PlayerAffiliation SetplayInterpreter::interpret(const NearestTeammate& player_functor) const
    {
        /// Switch location type: Region, Point, Player
        switch ( player_functor.location()->locationType() )
        {
            case ILocation2D::LT_REGION:
            {
                rcsc::Region2D* region = interpret( static_cast<const IRegion2D&>( *player_functor.location() ));
                unsigned int number = M_world_model->teammateFastestTo( *region );
                delete region;
                return PlayerAffiliation( number, TA_OUR);
            }
            case ILocation2D::LT_POINT:
            {
                rcsc::Vector2D position = interpret( static_cast<const IPoint2D&>( *player_functor.location() ));
                unsigned int number = M_world_model->teammateFastestTo( position );
                return PlayerAffiliation( number, TA_OUR);
            }
            case ILocation2D::LT_MOBILE_OBJECT:
            {
                const MobileObjectLocation& mol = static_cast<const MobileObjectLocation&>( *player_functor.location() );
                switch ( mol.mobileObject()->mobileObjectType() )
                {
                    case IMobileObject::MOT_PLAYER:
                    {
                        PlayerAffiliation player_ref = interpret( static_cast<const IPlayer&>( *mol.mobileObject() ));
                        unsigned int number = M_world_model->teammateFastestTo( player_ref );
                        return PlayerAffiliation( number, TA_OUR );
                    }
                    case IMobileObject::MOT_BALL:
                    {
                        PlayerAffiliation player( M_world_model->teammateFastestToBall(), TA_OUR);
                        return player;
                    }
                }
            }
        default:
            assert( false );
        }
    }

    PlayerAffiliation SetplayInterpreter::interpret(const NearestOpponent& player_functor) const
    {
        switch ( player_functor.location()->locationType() )
        {
            case ILocation2D::LT_REGION:
            {
                rcsc::Region2D* region = interpret( static_cast<const IRegion2D&>( *player_functor.location() ));
                unsigned int number = M_world_model->opponentFastestTo( *region );
                delete region;
                return PlayerAffiliation( number, TA_OPP);
            }
            case ILocation2D::LT_POINT:
            {
                rcsc::Vector2D position = interpret( static_cast<const IPoint2D&>( *player_functor.location() ));
                unsigned int number = M_world_model->opponentFastestTo( position );
                return PlayerAffiliation( number, TA_OPP);
            }
            case ILocation2D::LT_MOBILE_OBJECT:
            {
                const MobileObjectLocation& mol = static_cast<const MobileObjectLocation&>( *player_functor.location() );
                switch ( mol.mobileObject()->mobileObjectType() )
                {
                    case IMobileObject::MOT_PLAYER:
                    {
                        PlayerAffiliation player_ref = interpret( static_cast<const IPlayer&>( *mol.mobileObject() ));
                        unsigned int number = M_world_model->opponentFastestTo( player_ref);
                        return PlayerAffiliation( number, TA_OPP);
                    }
                    case IMobileObject::MOT_BALL:
                    {
                        PlayerAffiliation player( M_world_model->opponentFastestToBall(),
                                                          TA_OPP);

                        return player;
                    }
                }
            }
        default:
            assert( false );
        }
    }

    PlayerAffiliation SetplayInterpreter::interpret(const Goalie& player_functor) const
    {
        return PlayerAffiliation( M_world_model->goalie( player_functor.affiliation()),
                                  player_functor.affiliation() );
    }


    /// Assignment operations
    void SetplayInterpreter::interpret(const Assignment& assign)
    {
        MemorySpace* current_space = currentSpace();

        /// TODO Check for errors by trying to determine the result type of an expression
        Object* object = calculate( assign.expression() );
        current_space->put( assign.variable()->name(), object );
    }


    /// Helper methods for Execution State

    std::map<SetplayInterpreter::ExecutionState, string> SetplayInterpreter::createExecutionStateStringsMap()
    {
        std::map<SetplayInterpreter::ExecutionState, string> map;
        map.insert(pair<ExecutionState, string>( Stopped, "Stopped" ));
        map.insert(pair<ExecutionState, string>( SetplayStarted, "SetplayStarted" ));
        map.insert(pair<ExecutionState, string>( StepWait, "StepWait" ));
        map.insert(pair<ExecutionState, string>( StepChosen, "StepChosen" ));
        map.insert(pair<ExecutionState, string>( StepStarted, "StepStarted" ));
        map.insert(pair<ExecutionState, string>( StepAborted, "StepAborted" ));
        map.insert(pair<ExecutionState, string>( TransitionChosen, "TransitionChosen" ));
        map.insert(pair<ExecutionState, string>( TransitionStarted, "TransitionStarted" ));
        map.insert(pair<ExecutionState, string>( TransitionEnded, "TransitionEnded" ));
        map.insert(pair<ExecutionState, string>( TransitionAborted, "TransitionAborted" ));
        map.insert(pair<ExecutionState, string>( SetplayAborted, "SetplayAborted" ));
        map.insert(pair<ExecutionState, string>( SetplayFinished, "SetplayFinished" ));

        return map;
    }

    const string &SetplayInterpreter::executionStateStr(SetplayInterpreter::ExecutionState execution_state)
    {
        static std::map<SetplayInterpreter::ExecutionState, string> execution_state_strings = createExecutionStateStringsMap();
        return execution_state_strings.at( execution_state );
    }

    void SetplayInterpreter::registerObserver( ISetplayExecutionObserver* observer )
    {
        M_observers.push_back( observer );
    }

    void SetplayInterpreter::unregisterObserver( ISetplayExecutionObserver* observer )
    {
        M_observers.remove( observer );
    }

    void SetplayInterpreter::notifyObservers()
    {
        for ( list<ISetplayExecutionObserver*>::iterator ito = M_observers.begin();
              ito != M_observers.end(); ito++ )
        {
            (*ito)->stateChanged( *this );
        }
    }
}
