// -*-c++-*-
#ifndef SP_SETPLAY_EXECUTOR_H
#define SP_SETPLAY_EXECUTOR_H

#include <list>
#include <vector>
#include <queue>
#include <string>
#include <iostream>

#include <setplay/model/setplay.h>
#include <setplay/model/variable.h>
#include <setplay/model/actions.h>
#include <setplay/model/scope.h>

#include <setplay/model/scope.h>
#include <setplay/runtime/memory_space.h>
#include <rcsc/geom.h>

#include <setplay/runtime/iworld_model.h>

using namespace std;

namespace setplay
{
    class ISetplayExecutionObserver;
    class ICondition;
    class IAction;
    class IConditionEvaluator;
    class IActionExecutor;
    class ITransitionSelector;
    class IAbortDecider;
    class IStateUpdateDecider;
    class IRegion2D;
    class ILocation2D;
    class IPlayer;

    // 1) SetplayManager has a pointer to a global execution context
    // 2) Upon instantiation, each setplay has its own execution context which points to the previous
    // 3) Upon instantiation, each step has its own execution context which points to the setplay
    // 4) Upon instantiation, each step has its own execution context which points to the setplay

    class SetplayInterpreter
    {
    public:
        typedef enum
        {
            Stopped,
            SetplayStarted, // StepWait condition was not yet verified
            SetplayAborted,
            SetplayFinished,
            StepChosen,
            StepWait,
            StepStarted, // StepStart condition was not yet verified
            StepAborted, // StepStart condition was not yet verified
            StepEnded,
            TransitionChosen,
            TransitionStarted,
            TransitionEnded,
            TransitionAborted
        } ExecutionState;

      SetplayInterpreter();
      virtual ~SetplayInterpreter();

      /// Accessors
      ExecutionState executionState() const;
      const Setplay* currentSetplay() const;
      const Step* currentStep() const;
      const Step* nextStep() const;
      const ITransition* currentTransition() const;

      bool hasActiveSetplay() const;
      bool hasCurrentSetplay() const;
      bool hasCurrentStep() const;
      bool hasCurrentTransition() const;
      bool hasNextStep() const;

      PlayerAffiliation self() const;
      PlayerAffiliation currentStepLeader() const;
      PlayerAffiliation nextStepLeader() const;


      /*!
        \brief Check if a player participates in the set-play
        \param player The player whose participation should be checked
        \return True if the player participates in the setplay (regardless of the current and next steps), or False otherwise
        */
      bool participatesInSetplay(const PlayerAffiliation& player);
      bool selfParticipatesInSetplay();

      /*!
        \brief Check if a player participates in the current step of the set-play
        \param player The player whose participation should be checked
        \return True if the player participates in the current step, or False otherwise
        */
      bool participatesInCurrentStep(const PlayerAffiliation& player);
      bool selfParticipatesInCurrentStep();

      /*!
        \brief Check if a player participates in the next step of the set-play
        \param player The player whose participation should be checked
        \return True if the player participates in the next step, or False otherwise
        */
      bool participatesInNextStep(const PlayerAffiliation& player);
      bool selfParticipatesInNextStep();

      /// Modifiers
      void update();

      void loadSetplay(const Setplay* setplay);
      void loadStep(const Step* new_step);
      void enterStep(const Step* new_step);
      void endCurrentStep();
      void followTransition(const ITransition *new_transition);
      void endTransition();
      void abortSetplay();
      void finishSetplay();
      void reset();

      void setWorldModel( IWorldModel* context_provider);
      void setGlobalSpace( const MemorySpace* global_space);
      void setConditionEvaluator( IConditionEvaluator* condition_evaluator);
      void setActionExecutor( IActionExecutor* action_executor);
      void setTransitionSelector( ITransitionSelector* transition_selector);
      void setAbortDecider( IAbortDecider* abort_decider);
      void setStateUpdateDecider( IStateUpdateDecider *state_update_decider);

      const IWorldModel* worldModel() const
      {
          return M_world_model;
      }

      //Inspectors
      /*!
       * \
       * @param p The player whose IActions are to be retrieved
       * @return New vector with the IActions the player has to execute in the transition
      */
      vector<IAction*>* getParticipantDoActions(const IPlayer& p);

      void setInverted(bool inverted);
      bool inverted() const;

      /// Public interpretations
      bool interpret(const IAction& action) const;
      bool interpret(const ICondition& condition) const;

      Object* calculate(const Object *object);
      IRegion2D* calculate(const IRegion2D& region);

      rcsc::Region2D* interpret(const IRegion2D& region) const;
      rcsc::Vector2D interpret(const IPoint2D &point) const;
      PlayerAffiliation interpret(const IPlayer& player) const;
      float interpret(const IDecimal& decimal) const;

      /// Implementation of IVisitor interface for concept interpretation
      Object* interpret(const Object* object);

      void interpret(const Setplay& setplay);
      void interpret(const Step& step);

      /// Transitions
      void interpret(const ITransition& transition); // Dispatcher
      void interpret(const NextStep& transition);
      void interpret(const Finish& transition);
      void interpret(const Abort& transition);

      /// Directives
      void interpret(const Do& directive);
      void interpret(const Dont& );
      void interpret(const DoWhile& );

      const Object* interpret(const Variable& variable) const;

      /// Integer functors

      /// Variables
      const ILocation2D* interpret(const VarLocation& variable);
      const ICondition* interpret(const VarCondition& variable);


      /// Assignment operations
      void interpret(const Assignment& assign);

      /// Specific regions interpretation
      rcsc::Rect2D* interpret(const Rect2D& location) const;
      rcsc::Triangle2D* interpret(const Triangle2D& location) const;
      rcsc::Sector2D* interpret(const Sector2D& location) const;
      rcsc::Circle2D* interpret(const Circle2D& location) const;
      rcsc::Ray2D interpret(const Ray2D& location) const;
      rcsc::Line2D interpret(const Line2D& location) const;

      /// Specific points interpretation
      rcsc::Vector2D interpret(const Point2D& point) const;
      /// Point functors
      rcsc::Vector2D interpret(const PlayerPosition& point_functor) const;
      rcsc::Vector2D interpret(const BallPosition& ) const;
      // Vector unary operations
      rcsc::Vector2D interpret( const VectorNormalize& point_functor ) const;
      rcsc::Vector2D interpret( const VectorInvertX& point_functor ) const;
      rcsc::Vector2D interpret( const VectorInvertY& point_functor ) const;
      rcsc::Vector2D interpret( const VectorInvertXY& point_functor ) const;
      rcsc::Vector2D interpret( const VectorSubtraction& point_functor) const;
      rcsc::Vector2D interpret( const VectorAddition& point_functor ) const;
      // Vector scalar operations
      rcsc::Vector2D interpret( const VectorScale& point_functor ) const;
      rcsc::Vector2D interpret( const VectorRotation& point_functor ) const;
      rcsc::Vector2D interpret( const VectorDivision& point_functor ) const;

      /// Specific players interpretation
      PlayerAffiliation interpret(const PlayerAffiliation& participant) const;
      PlayerAffiliation interpret(const VarPlayer& participant) const;
      /// Player functors
      PlayerAffiliation interpret(const PlayerSelf& ) const;
      PlayerAffiliation interpret(const NearestPlayer& player_functor) const;
      PlayerAffiliation interpret(const NearestTeammate& player_functor) const;
      PlayerAffiliation interpret(const NearestOpponent& player_functor) const;
      PlayerAffiliation interpret(const Goalie& player_functor) const;


      /// Specific decimals interpretation
      float interpret(const Decimal& number) const;
      float interpret(const Integer& number) const;
      /// Decimal functors
      // Number operations
      float interpret( const NumberAbs& decimal_functor ) const;
      float interpret( const NumberSignChange& decimal_functor ) const;
      float interpret( const NumberMultiplication& decimal_functor ) const;
      float interpret( const NumberSubtraction& decimal_functor ) const;
      float interpret( const NumberDivision& decimal_functor ) const;
      float interpret( const NumberAddition& decimal_functor ) const;
      // Vector to scalar operations
      float interpret( const VectorDirection& decimal_functor ) const;
      float interpret( const VectorLength& decimal_functor ) const;
      // Two vector to scalar operations
      float interpret( const VectorDistance& decimal_functor ) const;
      float interpret( const VectorAngularDiff& decimal_functor ) const;
      float interpret( const VectorInnerProduct& decimal_functor ) const;
      float interpret( const VectorOuterProduct& decimal_functor ) const;
      // In game queries
      float interpret(const DefenseLine& decimal_functor) const;
      float interpret(const OffenseLine& decimal_functor) const;
      float interpret(const OffsideLine& decimal_functor) const;
      float interpret(const MobileObjectSpeed& decimal_functor) const;

      float interpret(const PointX& decimal_functor) const;
      float interpret(const PointY& decimal_functor) const;

      float interpret(const CurrentTime& ) const;
      float interpret(const RemainingTime& ) const;
      float interpret(const GoalsScored& decimal_functor ) const;
      float interpret(const GoalDiff& decimal_functor ) const;
      float interpret(const InterceptionTime& decimal_functor ) const;

      /// Specific conditions interpretation
      bool interpret(const CondAnd& condition) const;
      bool interpret(const CondOr& condition) const;
      bool interpret(const CondNot& condition) const;
      bool interpret(const CondBallOwner& condition) const;
      bool interpret(const CondTeamInPossession& condition) const;
      bool interpret(const CondBool& condition) const;
      bool interpret(const CondGameSituation& condition) const;
      bool interpret(const CondAtLocation& condition) const;
      bool interpret(const CondTeamFormation& ) const;
      bool interpret(const CondOffside& condition) const;
      bool interpret(const CondCompDecimal& condition) const;
      bool interpret(const CondKickable& condition) const;
      bool interpret(const CondCanIntercept& condition) const;
      bool interpret(const CondCanShoot& condition) const;
      bool interpret(const CondCanPass& condition) const;

      /// Specific actions interpretation
      bool interpret(const ActDribble& action) const;
      bool interpret(const ActFollow& action) const;
      bool interpret(const ActHold& action) const;
      bool interpret(const ActIntercept& action) const;
      bool interpret(const ActPass& action) const;
      bool interpret(const ActPos& action) const;
      bool interpret(const ActLookAt& action) const;

      /// Memory management => Should be private
      void push( std::string name, Object* object);
      bool unset( const std::string& name );
      const Object* resolve( const std::string& name ) const;

      static const std::string& executionStateStr(SetplayInterpreter::ExecutionState execution_state);

      /// Observer interface implementation
      void registerObserver( ISetplayExecutionObserver* observer );
      void unregisterObserver( ISetplayExecutionObserver* observer );

      bool canUpdateState() const;
    private:
      void notifyObservers();

      static std::map<SetplayInterpreter::ExecutionState, string> createExecutionStateStringsMap();

      /// Helper methods
      bool participatesInStep(const Step& step, const PlayerAffiliation& player);

      /// Accessors
      bool shouldAbort();
      bool shouldAbortSetplay();
      bool shouldAbortStep();

      void loadActions(const ITransition& transition);

      /// Name resolution functions
      /*!
        \brief Find the memory space that holds a given symbol
        */
      MemorySpace* currentSpace( );
      MemorySpace* getSpaceWithSymbol( const std::string& name );
      void resetSpace();

      void clearActions();

      // Map setplay concepts to memory spaces
      std::vector<MemorySpace*> M_objects_spaces;
      const MemorySpace* M_global_space;
      std::map<Object*, IScope*> M_objects_scopes;

      /// Setplay variables
      ITransitionSelector* M_transition_selector;
      IWorldModel* M_world_model;
      IActionExecutor* M_action_executor;
      IConditionEvaluator* M_condition_evaluator;
      IAbortDecider* M_abort_decider;
      IStateUpdateDecider* M_state_update_decider;

      bool M_inverted;


      /// State machine variables => Program control flow
      ExecutionState M_execution_state;
      const Setplay* M_current_setplay;
      const Step* M_last_step;
      const Step* M_current_step;
      const Step* M_next_step;
      const ITransition* M_current_transition;
      Actions M_current_actions;

      /// For the implementation of observer interface
      std::list<ISetplayExecutionObserver*> M_observers;
    };

}

#endif
