#ifndef SP_VALIDATOR_SEMANTICAL_H
#define SP_VALIDATOR_SEMANTICAL_H

#include <stack>
#include <vector>
#include <string>

#include <setplay/model/ivisitor.h>
#include <setplay/model/scope.h>


namespace setplay
{
    class IAction;
    class ICondition;
    class IDirective;
    class ITransition;
    class IPlayerID;
    class IParticipation;
    class IWorldModel;

    class ValidatorSemantical
            : public IVisitor
    {
    public:
        void visit( Setplay& setplay);
        void visit( Step& step);
        void visit( IParticipation& participation);

        /// Transitions
        void visit( NextStep& transition);
        void visit( Finish& transition);
        void visit( Abort& transition);

        /// Directives
        void visit( Do& directive);
        void visit( DoWhile& directive);
        void visit( Dont& directive);

        /// Actions
        void visit( ActDribble& action);
        void visit( ActFollow& action);
        void visit( ActHold& action);
        void visit( ActIntercept& action);
        void visit( ActPass& action);
        void visit( ActPos& action);
        void visit( ActShoot& action);

        /// Conditions
        void visit( CondAnd& condition);
        void visit( CondOr& condition);
        void visit( CondNot& condition);
        void visit( CondBallOwner& condition);
        void visit( CondTeamInPossession& condition);
        void visit( CondBool& condition);
        void visit( CondGameSituation& condition);
        void visit( CondAtLocation& condition);
        void visit( CondTeamFormation& condition);
        void visit( CondOffside& condition);
        void visit( CondCompDecimal& condition);
        void visit( CondKickable& condition);
        void visit( CondCanIntercept& condition);
        void visit( CondCanShoot& condition);
        void visit( CondCanPass& condition);

        /// Locations
        void visit( Rect2D& location);
        void visit( Triangle2D& location);
        void visit( Sector2D& location);
        void visit( Circle2D& location);
        void visit( Ray2D& location);
        void visit( Line2D& location);

        /// Participant
        void visit( PlayerAffiliation& participant);

        /// Player functors
        void visit( PlayerSelf& player_functor);
        void visit( NearestPlayer& player_functor);
        void visit( NearestTeammate& player_functor);
        void visit( NearestOpponent& player_functor);
        void visit( Goalie& player_functor);

        /// Points
        void visit( Point2D& point);

        /// Variables => Is this really necessary??
        void visit( Variable& variable);
        void visit( VarLocation& variable);
        void visit( VarRegion& variable);
        void visit( VarPoint& variable);
        void visit( VarCondition& variable);
        void visit( VarDecimal& variable);
        void visit( VarInteger& variable);
        void visit( VarPlayer& variable);

        /// Point Functors
        void visit( MobileObjectLocation& point_functor);
        void visit( PlayerPosition& point_functor);
        void visit( BallPosition& point_functor);
        // Vector unary operations
        void visit( VectorNormalize& point_functor);
        void visit( VectorInvertX& point_functor);
        void visit( VectorInvertY& point_functor);
        void visit( VectorInvertXY& point_functor);
        void visit( VectorSubtraction& point_functor);
        void visit( VectorAddition& point_functor);
        // Vector scalar operations
        void visit( VectorScale& point_functor);
        void visit( VectorRotation& point_functor);
        void visit( VectorDivision& point_functor);
        void visit( NearestIntersectionPoint& point_functor);
        void visit( RegionCentroid& point_functor);

        /// Simple Data Types
        void visit( Decimal& number);
        void visit( Integer& number);
        void visit( Bool& number );

        /// Number Functors
        // Number operations
        void visit( NumberAbs& decimal_functor);
        void visit( NumberSignChange& decimal_functor);
        void visit( NumberMultiplication& decimal_functor);
        void visit( NumberSubtraction& decimal_functor);
        void visit( NumberDivision& decimal_functor);
        void visit( NumberAddition& decimal_functor);
        // Vector to scalar operations
        void visit( VectorDirection& decimal_functor);
        void visit( VectorLength& decimal_functor);
        // Two vector to scalar operations
        void visit( VectorDistance& decimal_functor);
        void visit( VectorAngularDiff& decimal_functor);
        void visit( VectorInnerProduct& decimal_functor);
        void visit( VectorOuterProduct& decimal_functor);
        // In game queries
        void visit( DefenseLine& decimal_functor);
        void visit( OffenseLine& decimal_functor);
        void visit( OffsideLine& decimal_functor);
        void visit( MobileObjectSpeed& decimal_functor);

        void visit( PointX& decimal_functor );
        void visit( PointY& decimal_functor );

        /// Integer Functors
        void visit( CurrentTime& integer_functor);
        void visit( RemainingTime& integer_functor);
        void visit( GoalsScored& integer_functor);
        void visit( GoalDiff& integer_functor);
        void visit( InterceptionTime& integer_functor );

        /// Assignment function
        void visit( Assignment& assign);

        void visit( Ball& assign);
    private:
        bool validate(const Setplay& setplay, IWorldModel& context);

        // FA 03/06 Validate the semantics of the setplay, rather than its syntax => To be called in the end of parsing
        bool validate(const Step& step);
        bool validate(const Step& step, const ITransition* transition);
        bool validate(const IDirective& directive);
        bool validate(const IPlayerID& player, bool setplayParticipationRequired = true);
        static bool validate(const IPlayerID& player, const vector<IParticipation*>* participations);
        static bool validate(const IPlayerID& player, const vector<IPlayerID*>* participants);

        bool validate(const ICondition* condition);
        bool validate(const vector<IPlayerID*>* participants, bool setplayParticipationRequired = true);
        bool validate(const std::list<ICondition*>& conditions);

        bool validate(const IAction& action);
        bool validate(const ActPass& action);
        bool validate(const ActDribble& action);
        bool validate(const ActFollow& action);
        bool validate(const ActHold& action);
        bool validate(const ActIntercept& action);

        bool validate(const IRegion2D& region);
        bool validate(const Rect2D& region);
        bool validate(const Triangle2D& region);
        bool validate(const Arc2D& region);
        bool validate(const RegionVar& region);

        bool validate(const IPoint2D& point);
        bool validate(const IPointOperation& point);
        bool validate(const PointRel& point);
        bool validate(const PointStatic2D& point);
        bool validate(const PointPlayer& point);
        bool validate(const PointBall& point);

        void addIssue( IssueType type,
                       Severity severity,
                       std::string message );

        const Setplay* M_setplay;
        size_t M_step_idx;
        size_t M_transition_idx;
        size_t M_directive_idx;
        size_t M_action_idx;

        std::stack<Scope> M_scopes;

        enum Severity {
            WARNING,
            CRITICAL
        };

        enum IssueType
        {
            DuplicateStepId,
            TransitionToInexistentStep,
            UndefinedVariable,
            Other
        };

        struct Issue
        {
            Severity severity;
            IssueType type;
            std::string message;
        };

        std::vector<Issue> M_issues;

    public:
        const std::vector<Issue>& issues() const;
    };

}

#endif // SP_VALIDATOR_SEMANTICAL_H
