#include "xml_event_writer.h"

#include <sstream>

// Mandatory for using any feature of Xerces.
#include <xercesc/util/PlatformUtils.hpp>

// For character transcoding operations
#include <xercesc/util/XMLString.hpp>

#include <xercesc/util/OutOfMemoryException.hpp>

// Required for outputing a Xerces DOMDocument
// to a standard output stream (Also see: XMLFormatTarget)
#include <xercesc/framework/StdOutFormatTarget.hpp>

// Required for outputing a Xerces DOMDocument
// to the file system (Also see: XMLFormatTarget)
#include <xercesc/framework/LocalFileFormatTarget.hpp>

#include <soccer/events/xml_event_common.h>

using namespace std;
using namespace std::tr1;
XERCES_CPP_NAMESPACE_USE

namespace soccer {
namespace event {

XmlEventWriter::XmlEventWriter( )
{
    try
    {
        XMLPlatformUtils::Initialize();
    }
    catch (const XMLException& toCatch)
    {
        // Do your failure processing here
        std::cerr << "Could not initialize platform utils" << std::endl;
        assert( false );
    }

    DOMImplementation* dom_impl = DOMImplementationRegistry::getDOMImplementation( X("Core") );

    if ( dom_impl == NULL)
    {
        XERCES_STD_QUALIFIER cerr << "Requested implementation is not supported" << XERCES_STD_QUALIFIER endl;
        // errorCode = 4;
        assert( false );
    }

    try
    {
        M_dom_doc = dom_impl->createDocument( 0,                    // root element namespace URI.
                                              X("events"),          // root element name
                                              0);                   // document type object (DTD).
        M_root_elem = M_dom_doc->getDocumentElement();
    }
    catch (const OutOfMemoryException&)
    {
        XERCES_STD_QUALIFIER cerr << "OutOfMemoryException" << XERCES_STD_QUALIFIER endl;
        // errorCode = 5;
        assert(false);
    }
    catch (const DOMException& e)
    {
        XERCES_STD_QUALIFIER cerr << "DOMException code is:  " << e.code << XERCES_STD_QUALIFIER endl;
        // errorCode = 2;
        assert(false);
    }
    catch (...)
    {
        XERCES_STD_QUALIFIER cerr << "An error occurred creating the document" << XERCES_STD_QUALIFIER endl;
        // errorCode = 3;
        assert(false);
    }
}

XmlEventWriter::~XmlEventWriter()
{
    // Cleanup
    M_dom_doc->release();
    XMLPlatformUtils::Terminate();
}

void XmlEventWriter::visit(const KickOff& event)
{
    DOMElement* elem = createElementWithBaseAttributes("kick-off", event );
    appendTeamSideAttribute( elem, "teamSide", event.side() );
    appendChildElement( elem );
}

void XmlEventWriter::visit(const KickOffTaken& event)
{
    DOMElement* elem = createElementWithBaseAttributes("kick-off-taken", event );
    appendTeamSideAttribute( elem, "teamSide", event.side() );
    appendChildElement( elem );
}

void XmlEventWriter::visit(const ThrowIn& event)
{
    DOMElement* elem = createElementWithBaseAttributes("throw-in", event );
    appendTeamSideAttribute( elem, "teamSide", event.side() );
    appendChildElement( elem );
}

void XmlEventWriter::visit(const ThrowInTaken& event)
{
    DOMElement* elem = createElementWithBaseAttributes("throw-in-taken", event );
    appendTeamSideAttribute( elem, "teamSide", event.side() );
    appendChildElement( elem );
}

void XmlEventWriter::visit(const GoalKick& event)
{
    DOMElement* elem = createElementWithBaseAttributes("goal-kick", event );
    appendTeamSideAttribute( elem, "teamSide", event.side() );
    appendChildElement( elem );
}

void XmlEventWriter::visit(const GoalKickTaken& event)
{
    DOMElement* elem = createElementWithBaseAttributes("goal-kick-taken", event );
    appendTeamSideAttribute( elem, "teamSide", event.side() );
    appendChildElement( elem );
}

void XmlEventWriter::visit(const CornerKick& event)
{
    DOMElement* elem = createElementWithBaseAttributes("corner-kick", event );
    appendTeamSideAttribute( elem, "teamSide", event.side() );
    appendChildElement( elem );
}

void XmlEventWriter::visit(const CornerKickTaken& event)
{
    DOMElement* elem = createElementWithBaseAttributes("corner-kick-taken", event );
    appendTeamSideAttribute( elem, "teamSide", event.side() );
    appendChildElement( elem );
}

void XmlEventWriter::visit(const FreeKick& event)
{
    DOMElement* elem = createElementWithBaseAttributes("free-kick", event );
    appendTeamSideAttribute( elem, "teamSide", event.side() );
    appendChildElement( elem );
}

void XmlEventWriter::visit(const FreeKickTaken& event)
{
    DOMElement* elem = createElementWithBaseAttributes("free-kick-taken", event );
    appendTeamSideAttribute( elem, "teamSide", event.side() );
    appendChildElement( elem );
}

void XmlEventWriter::visit(const IndirectFreeKick& event)
{
    DOMElement* elem = createElementWithBaseAttributes("indirect-free-kick", event );
    appendTeamSideAttribute( elem, "teamSide", event.side() );
    appendChildElement( elem );
}

void XmlEventWriter::visit(const IndirectFreeKickTaken& event)
{
    DOMElement* elem = createElementWithBaseAttributes("indirect-free-kick-taken", event );
    appendTeamSideAttribute( elem, "teamSide", event.side() );
    appendChildElement( elem );
}

void XmlEventWriter::visit(const SoccerMomentChanged& event)
{
    DOMElement* elem = createElementWithBaseAttributes("soccer-moment-changed", event );
    appendTeamSideAttribute( elem, "teamSide", event.teamMoment().teamSide() );
    elem->setAttribute( X("moment"),
                        X( SoccerMomentChanged::TeamSideMoment::moment( event.teamMoment().moment()).c_str() ) );
    appendChildElement( elem );
}

void XmlEventWriter::visit(const PossessionChanged& event)
{
    DOMElement* elem = createElementWithBaseAttributes("possession-changed", event );
    appendTeamSideAttribute( elem, "teamInPossession", event.teamInPossession() );
    appendChildElement( elem );
}

void XmlEventWriter::visit(const Interception& event)
{
    DOMElement* elem = createElementWithBaseAttributes("interception", event );
    appendTeamSideAttribute( elem, "teamSide", event.interceptor().side() );
    appendPlayerUnumAttribute( elem, "number", event.interceptor().number() );
    appendChildElement( elem );
}

void XmlEventWriter::visit(const Turnover& event)
{
    DOMElement* elem = createElementWithBaseAttributes("turnover", event );
    appendTeamSideAttribute( elem, "teamInPossession", event.teamInPossession() );
    appendChildElement( elem );
}

void XmlEventWriter::visit(const BallDispute& event)
{
    DOMElement* elem = createElementWithBaseAttributes("ball-dispute", event );
    for ( std::vector<PlayerId>::const_iterator itp = event.players().begin();
          itp != event.players().end() ; itp++ )
    {
        DOMElement* player_elem = M_dom_doc->createElement( X("player") );
        appendTeamSideAttribute( player_elem, "teamSide", (*itp).side() );
        appendPlayerUnumAttribute( player_elem, "number", (*itp).number() );
        elem->appendChild( player_elem );
    }
    appendChildElement( elem );
}

void XmlEventWriter::visit(const BallHold& event)
{
    DOMElement* elem = createElementWithBaseAttributes("ball-hold", event );
    appendTeamSideAttribute( elem, "teamSide", event.holder().side() );
    appendPlayerUnumAttribute( elem, "number", event.holder().number() );
    appendChildElement( elem );
}

void XmlEventWriter::visit(const GoalieCatch& event)
{
    DOMElement* elem = createElementWithBaseAttributes("goalie-catch", event );
    appendTeamSideAttribute( elem, "teamSide", event.teamSide() );
    appendChildElement( elem );
}

void XmlEventWriter::visit(const BallKick& event)
{
    DOMElement* elem = createElementWithBaseAttributes("ball-kick", event );
    appendTeamSideAttribute( elem, "teamSide", event.kicker().side() );
    appendPlayerUnumAttribute( elem, "number", event.kicker().number() );
    appendChildElement( elem );
}

void XmlEventWriter::visit(const PassChain& event)
{
    createEventChain( "pass-chain", event);
}

void XmlEventWriter::visit(const GoalPassChain& event )
{
    createEventChain( "goal-pass-chain", event);
}

void XmlEventWriter::visit(const Goal& event)
{
    DOMElement* elem = createElementWithBaseAttributes("goal", event );
    appendTeamSideAttribute( elem, "teamSide", event.side() );
    appendChildElement( elem );
}

void XmlEventWriter::visit(const Shoot& event)
{
    DOMElement* elem = createElementWithBaseAttributes("shoot", event );
    appendTeamSideAttribute( elem, "teamSide", event.shooter().side() );
    appendPlayerUnumAttribute( elem, "number", event.shooter().number() );
    appendChildElement( elem );
}

void XmlEventWriter::visit(const Dribble& event)
{
    DOMElement* elem = createElementWithBaseAttributes("dribble", event );
    appendTeamSideAttribute( elem, "teamSide", event.dribbler().side() );
    appendPlayerUnumAttribute( elem, "number", event.dribbler().number() );
    appendChildElement( elem );
}

void XmlEventWriter::visit(const PlayerMove& event)
{
    DOMElement* elem = createElementWithBaseAttributes("player-move", event );
    appendTeamSideAttribute( elem, "teamSide", event.player().side() );
    appendPlayerUnumAttribute( elem, "number", event.player().number() );
    appendChildElement( elem );
    appendVector( elem, "from", event.from() );
    appendVector( elem, "to", event.to() );
}

void XmlEventWriter::visit(const BallMove& event)
{
    DOMElement* elem = createElementWithBaseAttributes("ball-move", event );
    appendChildElement( elem );
    appendVector( elem, "from", event.from() );
    appendVector( elem, "to", event.to() );
}

void XmlEventWriter::visit(const Pass& event)
{
    DOMElement* elem = createElementWithBaseAttributes("pass", event );
    appendTeamSideAttribute( elem, "teamSide", event.teamSide() );
    appendPlayerUnumAttribute( elem, "sender", event.sender() );
    appendPlayerUnumAttribute( elem, "receiver", event.receiver() );
    appendChildElement( elem );
}

void XmlEventWriter::visit(const Offside& event)
{
    DOMElement* elem = createElementWithBaseAttributes("offside", event );
    appendTeamSideAttribute( elem, "teamSide", event.side() );
    appendChildElement( elem );
}

/// Helper methods

void XmlEventWriter::write( const std::string& file_name,
                            const std::string& source )
{
    DOMImplementation* pImplement  = NULL;
    DOMLSSerializer*   pSerializer = NULL;

    M_root_elem->setAttribute( X("source"), X( source.c_str() ) );

    /*
    Return the first registered implementation that has
    the desired features. In this case, we are after
    a DOM implementation that has the LS feature... or Load/Save.
    */
    pImplement = DOMImplementationRegistry::getDOMImplementation( X("LS") );

    /*
    From the DOMImplementation, create a DOMSerializer.
    DOMSerializer are used to serialize a DOM tree [back] into an XML document.
    */
    pSerializer = pImplement->createLSSerializer();

    /*
    This line is optional. It just sets a feature
    of the Serializer to make the output
    more human-readable by inserting line-feeds and tabs,
    without actually inserting any new nodes
    into the DOM tree. (There are many different features to set.)
    Comment it out and see the difference.
    */
    pSerializer->getDomConfig()->setParameter( XMLUni::fgDOMWRTFormatPrettyPrint, true );

    /*
    Choose a location for the serialized output. The 3 options are:
        1) StdOutFormatTarget     (std output stream -  good for debugging)
        2) MemBufFormatTarget     (to Memory)
        3) LocalFileFormatTarget  (save to file)
        (Note: You'll need a different header file for each one)
    */
    // pTarget = new LocalFileFormatTarget( X(file_name.c_str()) );

    // Write the serialized output to the target.
    pSerializer->writeToURI( M_dom_doc, X(file_name.c_str()) );
}


/// Helper methods
void XmlEventWriter::createEventChain( const std::string& label, const EventChain& event )
{
    DOMElement* elem = createElementWithBaseAttributes( label, event );
    /// TODO Echo all inner elements
    appendChildElement( elem );
    M_stack.push( elem );

    for ( std::vector< shared_ptr<IEvent> >::const_iterator ite = event.begin();
          ite != event.end(); ite++ )
    {
        (*ite)->visit( *this );
    }
    M_stack.pop();
}

void XmlEventWriter::appendChildElement( DOMElement* elem )
{
    if ( !M_stack.empty() )
    {
        DOMElement* parent = M_stack.top();
        parent->appendChild( elem );
        return;
    }

    M_root_elem->appendChild( elem );
}

DOMElement* XmlEventWriter::createElementWithBaseAttributes( std::string type_attr_value, const IEvent &event )
{
    // DOMElement* element = M_dom_doc->createElement( X("event") );
    // element->setAttribute( X("type") , X(type_attr_value.c_str()) );
    DOMElement* element = M_dom_doc->createElement( X(type_attr_value.c_str()) );

    std::stringstream ss_start_time;
    ss_start_time << event.startTime();

    element->setAttribute( X("startTime") , X( ss_start_time.str().c_str()) );
    if ( event.duration() > 0 )
    {
        std::stringstream ss_end_time;
        ss_end_time << event.endTime();

        element->setAttribute( X("endTime") , X( ss_end_time.str().c_str() ) );
    }
    return element;
}

void XmlEventWriter::appendTeamSideAttribute( xercesc::DOMElement* element,
                                              const std::string& label,
                                              const TeamSide& side )
{
    element->setAttribute( X(label.c_str()), X( teamSide( side ).c_str() ));
}

void XmlEventWriter::appendPlayerUnumAttribute( xercesc::DOMElement* element,
                                                const std::string& label,
                                                const unsigned int& unum )
{
    stringstream ss;
    ss << unum;
    element->setAttribute( X(label.c_str()), X( ss.str().c_str() ));
}

void XmlEventWriter::appendVector( xercesc::DOMElement* element,
                                   const std::string& label,
                                   const rcsc::Vector2D& vector )
{
    DOMElement* vector_elem = M_dom_doc->createElement( X(label.c_str()) );
    element->appendChild( vector_elem );

    stringstream ssx;
    ssx << vector.x;
    vector_elem->setAttribute( X("x"), X( ssx.str().c_str() ) );

    stringstream ssy;
    ssy << vector.y;
    vector_elem->setAttribute( X("y"), X( ssy.str().c_str() ) );
}

}
}
