#ifndef IWORLD_MODEL_H
#define IWORLD_MODEL_H

#include <vector>
#include <rcsc/geom/angle_deg.h>
#include <rcsc/geom/rect_2d.h>
#include <rcsc/geom/line_2d.h>

#include <soccer/game_mode.h>
#include <soccer/player_id.h>
#include <soccer/player_affiliation.h>
#include <soccer/types.h>

namespace soccer {
    class Player;
    class IPlayer;

    class IWorldModel
    {
    public:
        virtual ~IWorldModel() { }

        virtual unsigned int currentTime() const = 0;
        virtual unsigned int remainingTimeInPeriod() const = 0;

        virtual unsigned int ballInterceptionTime( const TeamAffiliation& affiliation,
                                                   const Unum& number ) const = 0;
        inline unsigned int ballInterceptionTime( const PlayerAffiliation& player ) const
        {
            return ballInterceptionTime( player.affiliation(), player.number() );
        }

        virtual unsigned int playerInterceptionTime( const TeamAffiliation& player1_affiliation,
                                                     const Unum& player1_number,
                                                     const TeamAffiliation& player2_affiliation,
                                                     const Unum& player2_number ) const = 0;
        inline unsigned int playerInterceptionTime( const PlayerAffiliation& player1,
                                                    const PlayerAffiliation& player2 ) const
        {
            return playerInterceptionTime( player1.affiliation(), player1.number(),
                                           player2.affiliation(), player2.number());
        }

        // Game stopped?
        virtual GameMode currentGameMode() const = 0;

        virtual GameSituation gameSituation() const = 0;

        /*!
          \brief Determine the current soccer moment of the game
          */
        virtual SoccerMomentAffiliation soccerMoment() const = 0;

        /*!
          \brief Determine the current time period
          */
        virtual TimePeriod timePeriod() const = 0;

        /// COMMON REGIONS
        virtual rcsc::Rect2D field( ) const = 0;
        virtual double defenseLine(TeamAffiliation team_affiliation) const = 0;
        virtual double offenseLine(TeamAffiliation team_affiliation) const = 0;
        double offsideLine(TeamAffiliation team_affiliation) const
        {
            return defenseLine( oppositeAffiliation( team_affiliation ) );
        }

        virtual double goalLine(TeamAffiliation team_affiliation) const = 0;
        virtual double endLine(TeamAffiliation team_affiliation) const = 0;

        virtual rcsc::Rect2D goalieArea(TeamAffiliation team_affiliation) const = 0;
        virtual rcsc::Rect2D penaltyArea(TeamAffiliation team_affiliation) const = 0;
        virtual rcsc::Rect2D midfieldArea(TeamAffiliation team_affiliation) const = 0;

        /// BALL INFORMATION
        virtual rcsc::Vector2D ballPosition( ) const = 0;
        virtual double ballSpeed() const = 0;
        virtual rcsc::AngleDeg ballDirection() const = 0;

        /// PLAYER INFORMATION
        virtual Unum self() const = 0;
        virtual Unum goalie( TeamAffiliation team_affiliation ) const = 0;
        virtual rcsc::Vector2D playerPosition( const TeamAffiliation& affiliation, const Unum& number ) const = 0;
        virtual double playerSpeed( const TeamAffiliation& affiliation, const Unum& number ) const = 0;
        virtual rcsc::AngleDeg playerBodyDirection( const TeamAffiliation& affiliation, const Unum& number ) const = 0;
        virtual int playerTimeToReachPosition( const TeamAffiliation& affiliation,
                                               const Unum& number,
                                               const rcsc::Vector2D& position ) const = 0;
        inline int playerTimeToReachPosition( const PlayerAffiliation& player,
                                               const rcsc::Vector2D& position ) const
        {
            return playerTimeToReachPosition( player.affiliation(), player.number(), position );
        }

        virtual double playerMaxSpeed( const TeamAffiliation& affiliation,
                                       const Unum& number ) const = 0;

        inline double playerMaxSpeed( const PlayerAffiliation& player ) const
        {
            return playerMaxSpeed( player.affiliation(), player.number() );
        }

        virtual std::vector<Unum> activePlayers( TeamAffiliation team_affiliation ) const = 0;

        virtual PlayerAffiliation playerFastestTo( const rcsc::Vector2D& position ) = 0;
        virtual PlayerAffiliation playerFastestTo( const rcsc::Region2D& region ) = 0;
        virtual PlayerAffiliation playerFastestTo( const TeamAffiliation& team,
                                                   const Unum& unum ) = 0;
        inline PlayerAffiliation playerFastestTo( const PlayerAffiliation& player )
        {
            return playerFastestTo( player.affiliation(), player.number() );
        }

        virtual PlayerAffiliation playerFastestToBall( ) = 0;

        virtual Unum teammateFastestTo( const rcsc::Vector2D& position ) = 0;
        virtual Unum teammateFastestTo( const rcsc::Region2D& region ) = 0;
        virtual Unum teammateFastestTo( const TeamAffiliation& team,
                                        const Unum& unum ) = 0;
        inline Unum teammateFastestTo( const PlayerAffiliation& player )
        {
            return teammateFastestTo( player.affiliation(), player.number() );
        }
        virtual Unum teammateFastestToBall( ) = 0;

        virtual Unum opponentFastestTo( const rcsc::Vector2D& position ) = 0;
        virtual Unum opponentFastestTo( const rcsc::Region2D& region ) = 0;
        virtual Unum opponentFastestTo( const TeamAffiliation& team,
                                        const Unum& unum ) = 0;
        inline Unum opponentFastestTo( const PlayerAffiliation& player )
        {
            return opponentFastestTo( player.affiliation(), player.number() );
        }

        virtual Unum opponentFastestToBall( ) = 0;

        virtual std::vector<Unum> teammatesFromPlayer( const TeamAffiliation& team_affiliation,
                                                               const Unum& number) const = 0;
        inline std::vector<Unum> teammatesFromPlayer( const PlayerAffiliation& player ) const
        {
            return teammatesFromPlayer( player.affiliation(), player.number() );
        }

        virtual std::vector<Unum> opponentsFromPlayer( const TeamAffiliation& team_affiliation,
                                                               const Unum& number ) const = 0;
        inline std::vector<Unum> opponentsFromPlayer( const PlayerAffiliation& player ) const
        {
            return opponentsFromPlayer( player.affiliation(), player.number() );
        }

        virtual std::vector<Unum> playersFromBall( TeamAffiliation team_affiliation ) const = 0;
        virtual std::vector<Unum> playersFromGoal( TeamAffiliation team_affiliation,
                                                           TeamAffiliation goal_affiliation) const = 0;
        virtual std::vector<Unum> playersBehindBall( TeamAffiliation team_affiliation) const = 0;
        virtual std::vector<Unum> playersInFrontOfBall( TeamAffiliation team_affiliation) const = 0;

        // virtual std::vector<IPlayerId*> teammatesToPlayer( const IPlayerId& player_id ) const = 0;
        // virtual std::vector<IPlayerId*> teammatesToPosition( const rcsc::Vector2D& position ) const = 0;
        virtual PlayerAffiliation lastKicker() const = 0;

        virtual TeamAffiliation teamInPossession() const = 0;

        virtual unsigned int goalsScored(TeamAffiliation affiliation) const = 0;

        int relativeScore() const
        {
            return goalsScored( TA_OUR ) - goalsScored( TA_OPP );
        }

        /*!
          \brief Determine the players belonging to a particular team affiliation that are offside
          \param team_affiliation The team affiliation the players belong to
          \returns Players that are offside
          */
        virtual std::vector<Unum> playersOffside(TeamAffiliation team_affiliation) const = 0;
        /*!
          \brief Determine the players belonging to a particular team affiliation that are not offside
          \param team_affiliation The team affiliation the players belong to
          \returns Players that are not offside
          */
        virtual std::vector<Unum> playersNotOffside(TeamAffiliation team_affiliation) const = 0;

        virtual int gamePace( ) const = 0;

        virtual ScoreAssessment scoreAssessment() const = 0;
    };

}

#endif // IWORLD_MODEL_H
