#include <strategy/dpre.h>

#include <lemon/matching.h>
#include <lemon/concepts/graph.h>
#include <lemon/list_graph.h>
#include <cassert>

using namespace std;

using namespace soccer;

using namespace lemon;
using namespace lemon::concepts;

namespace soccer
{

    namespace strategy
    {

        /*!
          \brief Get the positioning of a particular player by his number
          \return The player number whose role he is currently filling in
          */
        Unum DPRE::positioning(Unum number) const
        {
            return M_positionings.at(number);
        }

        /*!
          \brief Check if it's helpful to exchange positionings
          \abstract Uses a max weighted matching algorithm to calculate positionings
          \return True if al least one of the positionings changed
          */
        void DPRE::update( )
        {
            // Clear the previous positionings
            // M_positionings.clear();

            // Get eligible players for positioning exchanges => Deferred implementation
            const std::vector<Unum>& eligible_players = eligiblePlayers( );

            if ( eligible_players.size() <= 1 )
            {
                return;
            }

            ListGraph players_graph;
            // Initialize player nodes
            std::map<Unum, ListGraph::Node> player_to_node_map;
            std::map<ListGraph::Node, Unum> node_to_player_map;
            for ( std::vector<Unum>::const_iterator itp1 = eligible_players.begin();
                  itp1 != eligible_players.end();
                  itp1++)
            {
                ListGraph::Node node = players_graph.addNode();
                player_to_node_map.insert(std::pair<Unum, ListGraph::Node>(*itp1, node));
                node_to_player_map.insert(std::pair<ListGraph::Node, Unum>(node, *itp1));
            }

            MaxWeightedMatching<ListGraph,
                                ListGraph::EdgeMap<double> >::WeightMap players_exchange_values(players_graph);

            // Calculate players positionings
            for ( std::vector<Unum>::const_iterator itp1 = eligible_players.begin();
                  itp1 != eligible_players.end() - 1;
                  itp1++)
            {
                for ( std::vector<Unum>::const_iterator itp2 = itp1+1;
                      itp2 != eligible_players.end();
                      itp2++)
                {
                    ListGraph::Edge edge = players_graph.addEdge( player_to_node_map.at( *itp1 ),
                                                                  player_to_node_map.at( *itp2 ));
                    players_exchange_values[edge] = exchangeValue( *itp1 , *itp2 );
                }
            }

            MaxWeightedMatching<ListGraph,
                                ListGraph::EdgeMap<double> > matching(players_graph, players_exchange_values);
            matching.run();

            Positionings changed_positionings;
            for ( std::vector<Unum>::const_iterator itp1 = eligible_players.begin();
                  itp1 != eligible_players.end();
                  itp1++)
            {
                ListGraph::Node match_node = matching.mate( player_to_node_map[ *itp1 ] );
                if ( match_node == INVALID )
                {
                    continue;
                }

                Unum matched_player = node_to_player_map.at(match_node);
                if ( matched_player != M_positionings.at(*itp1) )
                {
                    changed_positionings[ *itp1 ] = matched_player;
                    M_positionings[ *itp1 ] = matched_player;
                }
            }

            if ( !changed_positionings.empty() )
            {
                positioningsChanged(changed_positionings);
            }
        }

        void DPRE::reset( )
        {
            for ( Positionings::iterator itp = M_positionings.begin();
                  itp != M_positionings.end(); itp++)
            {
                M_positionings[ itp->first ] = itp->first;
            }
        }

    }
}
