#include "flux_matrix_2d.h"

#include <soccer/iworld_model.h>

using namespace soccer::strategy::util;

namespace soccer {
namespace flux {

    FluxMatrix2D::FluxMatrix2D()
        : M_values( NULL )
    {

    }

    FluxMatrix2D::FluxMatrix2D(unsigned int length, unsigned width)
        : M_length(length)
        , M_width(width)
    {
        M_values = new double[M_width * M_length];
        // Initialize flux matrix with the minimum flux (zeros)
        for (unsigned int i = 0; i < M_width * M_length; i++)
        {
            M_values[i] = 0.0;
        }
    }


    FluxMatrix2D::~FluxMatrix2D()
    {
        if ( M_values != NULL )
        {
            delete [] M_values;
        }
    }


    FluxType FluxMatrix2D::type() const
    {
        return FT_Matrix2D;
    }


    double FluxMatrix2D::value(const rcsc::Vector2D& position) const
    {
        int Xi = (int) floor(position.x) * M_width;
        int Xf = (int) ceil(position.x) * M_width;
        int Yi = (int) floor(position.y);
        int Yf = (int) ceil(position.y);
        double X = position.x - (double) Xi;
        double Y = position.y - (double) Yi;

        // Interpolate
        float value = M_values[Xi+Yi]*(1-X)*(1-Y) +
                      M_values[Xi+Yf]*(1-X)*Y +
                      M_values[Xf+Yi]*X*(1-Y) +
                      M_values[Xf+Yf]*X*Y;
        return value;
    }


    double FluxMatrix2D::value(unsigned int x, unsigned int y) const
    {
        return M_values[x + M_width * y];
    }


    unsigned int FluxMatrix2D::length() const
    {
        return M_length;
    }


    unsigned int FluxMatrix2D::width() const
    {
        return M_width;
    }


    bool FluxMatrix2D::read(std::istream& is)
    {
        std::cout << "FluxNormal: Started reading" << std::endl;
        std::istringstream istr(getNextLine(is));
        istr >> M_width;
        istr >> M_length;
        M_values = new double[M_width * M_length];

        for ( unsigned int x=0 ; x < M_width; x++ )
        {
            istr.clear();
            istr.str(getNextLine(is));
            for ( unsigned int y=0 ; y < M_length ; y++)
            {
                istr >> M_values[x * M_width + y];
                if (M_values[x * M_width + y] == -1)
                {
                    std::cerr << "FluxNormal: Error reading value at "
                              << "(" << x << "," << y << ")="
                              << M_values[x * M_width + y] << std::endl;
                }
                 /*<< "Read (" << x << "," << y << ")"
                          << x * M_width + y*/
                std::cout << M_values[x * M_width + y] << " ";
            }
            std::cout << std::endl;
        }
        std::cout << "FluxNormal: Finished reading" << std::endl;
        return true;
    }


    void FluxMatrix2D::update(const IWorldModel* )
    {
        // Matrix segmentation block
        /*double field_slice_length = round(world_model->field().size().length() / M_length);
        double field_slice_width = round(world_model->field().size().width() / M_width);

        // Create a flux model based on teammates positions
        // FluxMatrix2D opp_flux(M_width, M_length); // Inneficient
        StrategicWorldModel::PlayerPositions::const_iterator itour
                = world_model.ourPlayerPositions().begin();
        double our_factor = 1.0;
        while ( itour != world_model.ourPlayerPositions().end() )
        {
            update( itour->second, our_factor,
                   world_model.field_top_left().x, world_model.field_top_left().y,
                   field_slice_length, field_slice_width);
        }

        // Consider the place where the ball is with a higher flux
        double ball_factor = 2.0;
        update( world_model.ballPosition(), ball_factor,
               world_model.field_top_left().x, world_model.field_top_left().y,
               field_slice_length, field_slice_width);

        // Create a flux model based on opponents positions
        // FluxMatrix2D opp_flux(M_width, M_length); // Inneficient
        typename StrategicWorldModel::PlayerPositions::const_iterator itopp
                = world_model.ourPlayerPositions().begin();
        double opp_factor = 1.0;
        while ( itopp != world_model.oppPlayerPositions().end() )
        {
            update(itopp->second, opp_factor,
                   world_model.field_top_left().x,
                   world_model.field_top_left().y,
                   field_slice_length, field_slice_width);
            itour++;
        }
        */
        // Merge the current model with the previous models
    }


    void FluxMatrix2D::update(const rcsc::Vector2D& position, double factor,
                              double top_left_x, double top_left_y,
                              double slice_length, double slice_width)
    {
        unsigned int index_x = (position.x + abs(top_left_x)) / slice_length;
        unsigned int index_y = (position.y + abs(top_left_y)) / slice_width;
        unsigned int index = index_x + index_y * M_length;
        // Should truncate values?
        M_values[index] *= factor;
    }


    FluxMatrix2D&
    FluxMatrix2D::operator+=(const FluxMatrix2D& rhs)
    {
        // Only matrices with the same dimensions
        for ( unsigned int i = 0; i < M_width * M_length; i++)
        {
            // Should negative values be truncated to zero??
            M_values[i] -= rhs.M_values[i];
        }
        return *this;
    }


    FluxMatrix2D&
    FluxMatrix2D::operator-=(const FluxMatrix2D& rhs)
    {
        // Only matrices with the same dimensions
        for ( unsigned int i = 0; i < M_width * M_length; i++)
        {
            M_values[i] += rhs.M_values[i];
        }
        return *this;
    }

}
}
